/* $Id: AlignDistribute.cc,v 1.3 2004/07/21 06:31:20 pklaus Exp $
 *
 * AlignDistribute.cc
 *
 * Gestalter - The free vector drawing application
 * Copyright (c) 2002 Philipp Klaus <gestalter@linotux.ch>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "config.h"
#include "AlignDistribute.hh"
#include "AlignDistributeAction.hh"
#include "GDApplication.hh"
#include "GDDocument.hh"
#include <algorithm>

struct BboxLeft {
	bool operator()(const AlignDistributeObject& a, const AlignDistributeObject& b) {
		return a.bbox.left < b.bbox.left;
	};
};

struct BboxRight {
	bool operator()(const AlignDistributeObject& a, const AlignDistributeObject& b) {
		return a.bbox.right < b.bbox.right;
	};
};

struct BboxHCenter {
	bool operator()(const AlignDistributeObject& a, const AlignDistributeObject& b) {
		return a.bbox.left + a.bbox.right < b.bbox.left + b.bbox.right;
	};
};

struct BboxTop {
	bool operator()(const AlignDistributeObject& a, const AlignDistributeObject& b) {
		return a.bbox.top < b.bbox.top;
	};
};

struct BboxBottom {
	bool operator()(const AlignDistributeObject& a, const AlignDistributeObject& b) {
		return a.bbox.bottom < b.bbox.bottom;
	};
};

struct BboxVCenter {
	bool operator()(const AlignDistributeObject& a, const AlignDistributeObject& b) {
		return a.bbox.top + a.bbox.bottom < b.bbox.top + b.bbox.bottom;
	};
};

AlignDistribute::AlignDistribute(GDApplication *application) {
	app = application;
}

AlignDistributeObjectVector AlignDistribute::getObjects() {
	AlignDistributeObjectVector objectList;
	if (doc) {
		SelectionItemList *selection = new SelectionItemList();
		doc->getSelection(*selection);
		G_List_Iterator<SelectionItem *> it = selection->begin();
		SelectionItem *item;
		while ((item = *it++)) {
			SelectableItem *sitem = item->orig;
			AlignDistributeObject adObject;
			adObject.object = sitem->getGObject();
			adObject.bbox = adObject.object->getBoundingBox();
			objectList.push_back(adObject);
		}
		delete selection;
	}
	return objectList;
}

void AlignDistribute::alignLeft() {
	AlignDistributeObjectVector objectList = getObjects();
	if (objectList.size() > 1 && doc) {
		AlignDistributeAction *newAction = new AlignDistributeAction(ALIGNDISTRIBUTEACTION_TYPE_ALIGN_LEFT);
		doc->setAction(newAction);
	}
	gdouble leftMost = G_MAXDOUBLE;
	for (AlignDistributeObjectVector::iterator it = objectList.begin(); it != objectList.end(); it++) {
		if ((*it).bbox.left < leftMost) {
			leftMost = (*it).bbox.left;
		}
	}
	for (AlignDistributeObjectVector::iterator it = objectList.begin(); it != objectList.end(); it++) {
		TransformationMatrix m;
		m.setTranslation(leftMost - (*it).bbox.left, 0);
		(*it).object->transformByMatrix(m, true);
	}
	if (doc) {
		doc->redraw();
	}
}

void AlignDistribute::alignHorizontalCenter() {
	AlignDistributeObjectVector objectList = getObjects();
	if (objectList.size() > 1 && doc) {
		AlignDistributeAction *newAction = new AlignDistributeAction(ALIGNDISTRIBUTEACTION_TYPE_ALIGN_HCENTER);
		doc->setAction(newAction);
	}
	gdouble average = 0;
	gint count = 0;
	for (AlignDistributeObjectVector::iterator it = objectList.begin(); it != objectList.end(); it++) {
		average += (*it).bbox.left;
		average += (*it).bbox.right;
		count += 2;
	}
	average /= count;
	for (AlignDistributeObjectVector::iterator it = objectList.begin(); it != objectList.end(); it++) {
		TransformationMatrix m;
		m.setTranslation(average - ((*it).bbox.right + (*it).bbox.left) / 2, 0);
		(*it).object->transformByMatrix(m, true);
	}
	if (doc) {
		doc->redraw();
	}
}

void AlignDistribute::alignRight() {
	AlignDistributeObjectVector objectList = getObjects();
	if (objectList.size() > 1 && doc) {
		AlignDistributeAction *newAction = new AlignDistributeAction(ALIGNDISTRIBUTEACTION_TYPE_ALIGN_RIGHT);
		doc->setAction(newAction);
	}
	gdouble rightMost = G_MINDOUBLE;
	for (AlignDistributeObjectVector::iterator it = objectList.begin(); it != objectList.end(); it++) {
		if ((*it).bbox.right > rightMost) {
			rightMost = (*it).bbox.right;
		}
	}
	for (AlignDistributeObjectVector::iterator it = objectList.begin(); it != objectList.end(); it++) {
		TransformationMatrix m;
		m.setTranslation(rightMost - (*it).bbox.right, 0);
		(*it).object->transformByMatrix(m, true);
	}
	if (doc) {
		doc->redraw();
	}
}

void AlignDistribute::alignTop() {
	AlignDistributeObjectVector objectList = getObjects();
	if (objectList.size() > 1 && doc) {
		AlignDistributeAction *newAction = new AlignDistributeAction(ALIGNDISTRIBUTEACTION_TYPE_ALIGN_TOP);
		doc->setAction(newAction);
	}
	gdouble topMost = G_MAXDOUBLE;
	for (AlignDistributeObjectVector::iterator it = objectList.begin(); it != objectList.end(); it++) {
		if ((*it).bbox.top < topMost) {
			topMost = (*it).bbox.top;
		}
	}
	for (AlignDistributeObjectVector::iterator it = objectList.begin(); it != objectList.end(); it++) {
		TransformationMatrix m;
		m.setTranslation(0, topMost - (*it).bbox.top);
		(*it).object->transformByMatrix(m, true);
	}
	if (doc) {
		doc->redraw();
	}
}

void AlignDistribute::alignVerticalCenter() {
	AlignDistributeObjectVector objectList = getObjects();
	if (objectList.size() > 1 && doc) {
		AlignDistributeAction *newAction = new AlignDistributeAction(ALIGNDISTRIBUTEACTION_TYPE_ALIGN_VCENTER);
		doc->setAction(newAction);
	}
	gdouble average = 0;
	gint count = 0;
	for (AlignDistributeObjectVector::iterator it = objectList.begin(); it != objectList.end(); it++) {
		average += (*it).bbox.top;
		average += (*it).bbox.bottom;
		count += 2;
	}
	average /= count;
	for (AlignDistributeObjectVector::iterator it = objectList.begin(); it != objectList.end(); it++) {
		TransformationMatrix m;
		m.setTranslation(0, average - ((*it).bbox.top + (*it).bbox.bottom) / 2);
		(*it).object->transformByMatrix(m, true);
	}
	if (doc) {
		doc->redraw();
	}
}

void AlignDistribute::alignBottom() {
	AlignDistributeObjectVector objectList = getObjects();
	if (objectList.size() > 1 && doc) {
		AlignDistributeAction *newAction = new AlignDistributeAction(ALIGNDISTRIBUTEACTION_TYPE_ALIGN_BOTTOM);
		doc->setAction(newAction);
	}
	gdouble bottomMost = G_MINDOUBLE;
	for (AlignDistributeObjectVector::iterator it = objectList.begin(); it != objectList.end(); it++) {
		if ((*it).bbox.bottom > bottomMost) {
			bottomMost = (*it).bbox.bottom;
		}
	}
	for (AlignDistributeObjectVector::iterator it = objectList.begin(); it != objectList.end(); it++) {
		TransformationMatrix m;
		m.setTranslation(0, bottomMost - (*it).bbox.bottom);
		(*it).object->transformByMatrix(m, true);
	}
	if (doc) {
		doc->redraw();
	}
}

void AlignDistribute::distributeLeftEdge() {
	AlignDistributeObjectVector objectList = getObjects();
	sort(objectList.begin(), objectList.end(), BboxLeft());
	if (objectList.size() > 2) {
		AlignDistributeAction *newAction = new AlignDistributeAction(ALIGNDISTRIBUTEACTION_TYPE_DISTRIBUTE_LEFT_EDGE);
		doc->setAction(newAction);
		gdouble step = (objectList.back().bbox.left - objectList.front().bbox.left) / (objectList.size() - 1);
		gdouble x = objectList.front().bbox.left + step;
		AlignDistributeObjectVector::iterator it = objectList.begin();
		it++;
		for (uint i = 1; i < objectList.size() - 1; i++, it++) {
			TransformationMatrix m;
			m.setTranslation(x - (*it).bbox.left, 0);
			(*it).object->transformByMatrix(m, true);
			x += step;
		}
	}
	if (doc) {
		doc->redraw();
	}
}

void AlignDistribute::distributeHorizontalCenter() {
	AlignDistributeObjectVector objectList = getObjects();
	sort(objectList.begin(), objectList.end(), BboxHCenter());
	if (objectList.size() > 2) {
		AlignDistributeAction *newAction = new AlignDistributeAction(ALIGNDISTRIBUTEACTION_TYPE_DISTRIBUTE_HCENTER);
		doc->setAction(newAction);
		gdouble step = ((objectList.back().bbox.left + objectList.back().bbox.right) - (objectList.front().bbox.left + objectList.front().bbox.right)) / 2 / (objectList.size() - 1);
		gdouble x = (objectList.front().bbox.left + objectList.front().bbox.right) / 2 + step;
		AlignDistributeObjectVector::iterator it = objectList.begin();
		it++;
		for (uint i = 1; i < objectList.size() - 1; i++, it++) {
			TransformationMatrix m;
			m.setTranslation(x - ((*it).bbox.left + (*it).bbox.right) / 2, 0);
			(*it).object->transformByMatrix(m, true);
			x += step;
		}
	}
	if (doc) {
		doc->redraw();
	}
}

void AlignDistribute::distributeHorizontalGap() {
	AlignDistributeObjectVector objectList = getObjects();
	sort(objectList.begin(), objectList.end(), BboxHCenter());
	if (objectList.size() > 2) {
		AlignDistributeAction *newAction = new AlignDistributeAction(ALIGNDISTRIBUTEACTION_TYPE_DISTRIBUTE_HGAP);
		doc->setAction(newAction);
		AlignDistributeObjectVector::iterator fit = objectList.begin();
		AlignDistributeObjectVector::iterator last = fit;
		fit++;
		gdouble totgap = 0;
		for (uint i = 1; i < objectList.size(); i++, fit++) {
			totgap += (*fit).bbox.left - (*last).bbox.right;
			last = fit;
		}
		gdouble step = totgap / (objectList.size() - 1);
		AlignDistributeObjectVector::iterator it = objectList.begin();
		gdouble x = (*it).bbox.right;
		it++;
		for (uint i = 1; i < objectList.size() - 1; i++, it++) {
			TransformationMatrix m;
			x += step;
			m.setTranslation(x - (*it).bbox.left, 0);
			(*it).object->transformByMatrix(m, true);
			x += (*it).bbox.right - (*it).bbox.left;
		}
	}
	if (doc) {
		doc->redraw();
	}
}

void AlignDistribute::distributeRightEdge() {
	AlignDistributeObjectVector objectList = getObjects();
	sort(objectList.begin(), objectList.end(), BboxRight());
	if (objectList.size() > 2) {
		AlignDistributeAction *newAction = new AlignDistributeAction(ALIGNDISTRIBUTEACTION_TYPE_DISTRIBUTE_RIGHT_EDGE);
		doc->setAction(newAction);
		gdouble step = (objectList.back().bbox.right - objectList.front().bbox.right) / (objectList.size() - 1);
		gdouble x = objectList.front().bbox.right + step;
		AlignDistributeObjectVector::iterator it = objectList.begin();
		it++;
		for (uint i = 1; i < objectList.size() - 1; i++, it++) {
			TransformationMatrix m;
			m.setTranslation(x - (*it).bbox.right, 0);
			(*it).object->transformByMatrix(m, true);
			x += step;
		}
	}
	if (doc) {
		doc->redraw();
	}
}

void AlignDistribute::distributeTopEdge() {
	AlignDistributeObjectVector objectList = getObjects();
	sort(objectList.begin(), objectList.end(), BboxTop());
	if (objectList.size() > 2) {
		AlignDistributeAction *newAction = new AlignDistributeAction(ALIGNDISTRIBUTEACTION_TYPE_DISTRIBUTE_TOP_EDGE);
		doc->setAction(newAction);
		gdouble step = (objectList.back().bbox.top - objectList.front().bbox.top) / (objectList.size() - 1);
		gdouble y = objectList.front().bbox.top + step;
		AlignDistributeObjectVector::iterator it = objectList.begin();
		it++;
		for (uint i = 1; i < objectList.size() - 1; i++, it++) {
			TransformationMatrix m;
			m.setTranslation(0, y - (*it).bbox.top);
			(*it).object->transformByMatrix(m, true);
			y += step;
		}
	}
	if (doc) {
		doc->redraw();
	}
}

void AlignDistribute::distributeVerticalCenter() {
	AlignDistributeObjectVector objectList = getObjects();
	sort(objectList.begin(), objectList.end(), BboxVCenter());
	if (objectList.size() > 2) {
		AlignDistributeAction *newAction = new AlignDistributeAction(ALIGNDISTRIBUTEACTION_TYPE_DISTRIBUTE_VCENTER);
		doc->setAction(newAction);
		gdouble step = ((objectList.back().bbox.top + objectList.back().bbox.bottom) - (objectList.front().bbox.top + objectList.front().bbox.bottom)) / 2 / (objectList.size() - 1);
		gdouble y = (objectList.front().bbox.top + objectList.front().bbox.bottom) / 2 + step;
		AlignDistributeObjectVector::iterator it = objectList.begin();
		it++;
		for (uint i = 1; i < objectList.size() - 1; i++, it++) {
			TransformationMatrix m;
			m.setTranslation(0, y - ((*it).bbox.top + (*it).bbox.bottom) / 2);
			(*it).object->transformByMatrix(m, true);
			y += step;
		}
	}
	if (doc) {
		doc->redraw();
	}
}

void AlignDistribute::distributeVerticalGap() {
	AlignDistributeObjectVector objectList = getObjects();
	sort(objectList.begin(), objectList.end(), BboxVCenter());
	if (objectList.size() > 2) {
		AlignDistributeAction *newAction = new AlignDistributeAction(ALIGNDISTRIBUTEACTION_TYPE_DISTRIBUTE_VGAP);
		doc->setAction(newAction);
		AlignDistributeObjectVector::iterator fit = objectList.begin();
		AlignDistributeObjectVector::iterator last = fit;
		fit++;
		gdouble totgap = 0;
		for (uint i = 1; i < objectList.size(); i++, fit++) {
			totgap += (*fit).bbox.top - (*last).bbox.bottom;
			last = fit;
		}
		gdouble step = totgap / (objectList.size() - 1);
		AlignDistributeObjectVector::iterator it = objectList.begin();
		gdouble y = (*it).bbox.bottom;
		it++;
		for (uint i = 1; i < objectList.size() - 1; i++, it++) {
			TransformationMatrix m;
			y += step;
			m.setTranslation(0, y - (*it).bbox.top);
			(*it).object->transformByMatrix(m, true);
			y += (*it).bbox.bottom - (*it).bbox.top;
		}
	}
	if (doc) {
		doc->redraw();
	}
}

void AlignDistribute::distributeBottomEdge() {
	AlignDistributeObjectVector objectList = getObjects();
	sort(objectList.begin(), objectList.end(), BboxBottom());
	if (objectList.size() > 2) {
		AlignDistributeAction *newAction = new AlignDistributeAction(ALIGNDISTRIBUTEACTION_TYPE_DISTRIBUTE_BOTTOM_EDGE);
		doc->setAction(newAction);
		gdouble step = (objectList.back().bbox.bottom - objectList.front().bbox.bottom) / (objectList.size() - 1);
		gdouble y = objectList.front().bbox.bottom + step;
		AlignDistributeObjectVector::iterator it = objectList.begin();
		it++;
		for (uint i = 1; i < objectList.size() - 1; i++, it++) {
			TransformationMatrix m;
			m.setTranslation(0, y - (*it).bbox.bottom);
			(*it).object->transformByMatrix(m, true);
			y += step;
		}
	}
	if (doc) {
		doc->redraw();
	}
}

void AlignDistribute::closeWindow() {
	app->hideAlignDistribute();
}

void AlignDistribute::setDocument(GDDocument *document) {
	doc = document;
}

gint AlignDistribute::deleteWindow(GdkEventAny *ev) {
	closeWindow();
	return 1;
}
