/* $Id: BezierPath.cc,v 1.30 2005/05/07 17:39:32 pklaus Exp $
 *
 * BezierPath.cc
 *
 * Gestalter - The free vector drawing application
 * Copyright (c) 2002 Philipp Klaus <gestalter@linotux.ch>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "BezierPath.hh"
#include "BezierNode.hh"
#include "Layer.hh"
#include "TransformationMatrix.hh"
#include "SelectionItem.hh"
#include "BoundingBox.hh"
#include "DocWindow.hh"
#include "GDDocument.hh"
#include "GDStyle.hh"
#include "GDApplication.hh"
#include <stdio.h>
#include <sys/timeb.h>
#include <math.h>
#include <libart_lgpl/art_misc.h>
#include <libart_lgpl/art_svp.h>
#include <libart_lgpl/art_svp_vpath.h>
#include <libart_lgpl/art_rgb_svp.h>
#include <libart_lgpl/art_vpath_bpath.h>
#include <libart_lgpl/art_vpath_dash.h>
#include <libart_lgpl/art_svp_vpath_stroke.h>
#include <libart_lgpl/art_svp_point.h>

extern GdkPoint gdkpts[];
position	segpoints[SUBDIVISIONS+2];


BezierPath::BezierPath(GDDocument *document) : GObject(document) {
	nodes = NULL;
	drawStyle = new GDStyle(document->currentStrokeStyle);
	fillStyle = new GDStyle(document->currentFillStyle);
}

void BezierPath::addNode(BezierNode *node) {
	node->path = this;
	if (nodes == NULL) {
		nodes = node;
	} else {
		nodes->insert(0, node); 
	}
}

void BezierPath::removeNode(BezierNode *node) {
	nodes->remove(node);
	if (node == nodes) {
		nodes = node->next;
	}
}

void BezierPath::deleteNode(BezierNode *node) {
	//g_printerr("BezierPath::deleteNode\n");
	if (nodes->prev) {
		// path is closed
		//g_printerr("path was closed\n");
		nodes = node->next;
		nodes->prev = NULL;
		node->prev->next = NULL;
	} else {
		// path is open
		if (node == nodes) {
			// node is first
			nodes = node->next;
			nodes->prev = NULL;
		} else if (node->next) {
			// split path in two parts
			BezierPath *path = new BezierPath(doc);
			*(path->drawStyle) = *drawStyle;
			*(path->fillStyle) = *fillStyle;
			path->nodes = node->next;
			BezierNode *n = path->nodes;
			n->prev = NULL;
			while (n) {
				n->path = path;
				n = n->next;
			}
			node->prev->next = NULL;
			doc->findObjectLayer(this)->addObject(path);
		} else {
			// node is last
			node->prev->next = NULL;
		}
	}
}

void BezierPath::deleteCurve(BezierNode *node) {
	if (nodes->prev) {
		// path is closed
		nodes = node->next;
		nodes->prev = NULL,
		node->next = NULL;
	} else {
		// path is open, split in two
		BezierPath *path = new BezierPath(doc);
		*(path->drawStyle) = *drawStyle;
		*(path->fillStyle) = *fillStyle;
		path->nodes = node->next;
		BezierNode *n = path->nodes;
		n->prev = NULL;
		while (n) {
			n->path = path;
			n = n->next;
		}
		node->next = NULL;
		doc->findObjectLayer(this)->addObject(path);
		path->selectAll();
	}
	selectAll();
}

void BezierPath::calcMorphValues(BezierPath *source, BezierPath *target, double n) {
	fillStyle->calcMorphValues(source->fillStyle, target->fillStyle, n);
	drawStyle->calcMorphValues(source->drawStyle, target->drawStyle, n);
}

guint BezierPath::getNodeIndex(BezierNode *node) {
	if (node == nodes) {
		return 0;
	}
	guint i = 1;
	BezierNode *cn = nodes->next;
	while (cn != node && cn && cn != nodes) {
		cn = cn->next;
		i++;
	}
	return i;
}

BezierNode *BezierPath::getIndexNode(guint index) {
	guint i = 0;
	BezierNode *cn = nodes;
	while (cn && i < index) {
		cn = cn->next;
		i++;
	}
	return cn;
}


IntersectionVector *BezierPath::getIntersections(BezierPath *with) {
	IntersectionVector *result = NULL;
	BezierNode *n1 = nodes;
	while (n1) {
		BezierNode *n2 = with->nodes;
		while (n2) {
			if (n1->next && n2->next) {
				// ensure we have a complete curve segment
				CurveSegment a,b;
				a.origNode = n1;
				a.segmentSize = 1.0;
				a.splitbase = 0.0;
				a.oc1 = n1->onCurve;
				a.ha1 = n1->nextHandle;
				a.ha2 = n1->next->prevHandle;
				a.oc2 = n1->next->onCurve;
				b.origNode = n2;
				b.segmentSize = 1.0;
				b.splitbase = 0.0;
				b.oc1 = n2->onCurve;
				b.ha1 = n2->nextHandle;
				b.ha2 = n2->next->prevHandle;
				b.oc2 = n2->next->onCurve;
				IntersectionVector *ints = n1->getIntersections(a, b);
				//g_printerr("BezierPath::getIntersections nA=%lX + nB=%lX = %lX\n", (long) n1, (long) n2, (long) ints);
				
				//g_printerr("BezierPath::getIntersections %f/%f,%f/%f,%f/%f,%f/%f  <-> %f/%f,%f/%f,%f/%f,%f/%f => %d\n", a.oc1.x, a.oc1.y, a.ha1.x, a.ha1.y, a.ha2.x, a.ha2.y, a.oc2.x, a.oc2.y, b.oc1.x, b.oc1.y, b.ha1.x, b.ha1.y, b.ha2.x, b.ha2.y, b.oc2.x, b.oc2.y, ints ? ints->size() : 0);
				if (ints) {
					if (result) {
						for (IntersectionVector::iterator it = ints->begin(); it != ints->end(); it++) {
							result->push_back(*it);
						}
						delete ints;
					} else {
						result = ints;
					}
				}
			}
			n2 = n2->next;
			if (n2 == with->nodes) {
				n2 = NULL;
			}
		}
		n1 = n1->next;
		if (n1 == nodes) {
			n1 = NULL;
		}
	}
	return result;
}

void BezierPath::drawWireFrame(DocWindow *docwin, Gdk_GC& context, RectangleArea& drawArea) {
	BezierNode *node = nodes;
	gint32		numpts;
	//gint		isSelected = 0;
	
	BoundingBox bbox = getBoundingBox();
	if (drawArea.intersect(bbox)) {
		gboolean contained = bbox.contained(drawArea);
		if (node && node->next == NULL) {
			// single node
			node->drawSingleNode(docwin, context);
		} else {
			Gdk_GC tmpcontext(context);
			if (drawStyle && drawStyle->isGuide) {
				tmpcontext.set_dashes(1, 4);
				tmpcontext.set_line_style(GDK_LINE_ON_OFF_DASH);
			}
			while (node && node->next) {
	
				//g_printerr("BezierPath::drawWireFrame node %f,%f\n", node->onCurve.x, node->onCurve.y);
				BoundingBox bbox = node->getBoundingBox();
				if (node->prev) {
					BoundingBox bb = node->prev->getBoundingBox();
					bbox.combine(bb);
				}
				if (node->next) {
					BoundingBox bb = node->next->getBoundingBox();
					bbox.combine(bb);
				}
				if (contained || drawArea.intersect(bbox)) {				
					node->drawCurve(docwin, tmpcontext);
				}
				//isSelected += node->selection;
				node = node->next;
				if (node == nodes) break; // special case closed path
			}
			if (drawStyle && drawStyle->isGuide) {
				tmpcontext.set_line_style(GDK_LINE_SOLID);
			}
		}
	}
	/*if (isSelected || (node && node->selection)) {
		drawPoints(docwin, context);
	}*/
}

guint BezierPath::fillBpath(ArtBpath *artpath, guint size, gdouble sF, 
								gdouble ox, gdouble oy, gboolean filled, RectangleArea& drawArea) {
	guint i = 0;
	BezierNode *node = nodes;
	if (node) {
		artpath[i].code = ART_MOVETO;
		artpath[i].x1 = 0;
		artpath[i].y1 = 0;
		artpath[i].x2 = 0;
		artpath[i].y2 = 0;
		artpath[i].x3 = (node->onCurve.x - ox) * sF;
		artpath[i].y3 = (node->onCurve.y - oy) * sF;
		i++;
		node = node->next;
	}
	while (node && i < size) {
		artpath[i].code = ART_CURVETO;
		artpath[i].x1 = (node->prev->nextHandle.x - ox) * sF;
		artpath[i].y1 = (node->prev->nextHandle.y - oy) * sF;
		artpath[i].x2 = (node->prevHandle.x - ox) * sF;
		artpath[i].y2 = (node->prevHandle.y - oy) * sF;
		artpath[i].x3 = (node->onCurve.x - ox) * sF;
		artpath[i].y3 = (node->onCurve.y - oy) * sF;
		i++;
		node = node->next;
		if (node == nodes) {
			artpath[i].code = ART_CURVETO;
			artpath[i].x1 = (node->prev->nextHandle.x - ox) * sF;
			artpath[i].y1 = (node->prev->nextHandle.y - oy) * sF;
			artpath[i].x2 = (node->prevHandle.x - ox) * sF;
			artpath[i].y2 = (node->prevHandle.y - oy) * sF;
			artpath[i].x3 = (node->onCurve.x - ox) * sF;
			artpath[i].y3 = (node->onCurve.y - oy) * sF;
			i++;
			filled = false; // because we already closed the path
			break;
		}
	}
	if (filled && nodes) {
		artpath[i].code = ART_LINETO;
		artpath[i].x1 = 0;
		artpath[i].y1 = 0;
		artpath[i].x2 = 0;
		artpath[i].y2 = 0;
		artpath[i].x3 = (nodes->onCurve.x - ox) * sF;
		artpath[i].y3 = (nodes->onCurve.y - oy) * sF;
		i++;
	}
	return i;
}

guint BezierPath::countNodes() {
	guint size = 0;
	BezierNode *n = nodes;
	while (n) {
		size++;
		n = n->next;
		if (n == nodes) {
			break;
		}
	}
	return size;
}

ArtVpath *BezierPath::getVpath(gdouble sF, gdouble ox, gdouble oy, gboolean filled, RectangleArea& drawArea) {
	//g_printerr("BezierPath::getVpath\n");
	guint size = countNodes();
	ArtBpath *artpath = art_new(ArtBpath, size+2);
	guint len = fillBpath(artpath, size, sF, ox, oy, filled, drawArea);
	artpath[len].code = ART_END;
	artpath[len].x1 = 0;
	artpath[len].y1 = 0;
	artpath[len].x2 = 0;
	artpath[len].y2 = 0;
	artpath[len].x3 = 0;
	artpath[len].y3 = 0;
	ArtVpath *ap = art_bez_path_to_vec(artpath, 0.5); // 0.5 is the flatness
	art_free(artpath);
	//g_printerr("BezierPath::getVpath size=%d\n", size);
	return ap;
	
}

void BezierPath::draw(OffscreenBuffer& osBuffer, RectangleArea& drawArea) {
	//g_printerr("BezierPath::draw fillStyle:%lX drawType:%d\n", (long) fillStyle, fillStyle ? fillStyle->drawType : 0);
	BoundingBox bbox = getBoundingBox();
	if (drawArea.intersect(bbox)) {
		if (fillStyle && fillStyle->drawType != GDSTYLE_DRAWTYPE_TRANSPARENT) {
			//g_printerr("BezierPath::draw: fill type=%d\n", fillStyle->drawType);
			ArtVpath *avp = getVpath(osBuffer.scaleFactor, osBuffer.origin.x,
										osBuffer.origin.y, true, drawArea);
			ArtSVP *asvp = art_svp_from_vpath(avp);
			art_free(avp);
			fillStyle->fillSVP(asvp, osBuffer.pixbuffer, osBuffer.pxbRowstride,
							osBuffer.pxbWidth, osBuffer.pxbHeight,
							osBuffer.origin.x, osBuffer.origin.y, osBuffer.scaleFactor	);
			art_svp_free(asvp);
		}
		if (drawStyle && drawStyle->drawType != GDSTYLE_DRAWTYPE_TRANSPARENT) {
			//g_printerr("BezierPath::draw: stroke width=%d\n", drawStyle->strokeWidth);
			ArtVpath *avp = getVpath(osBuffer.scaleFactor, osBuffer.origin.x,
										osBuffer.origin.y, false, drawArea);
			drawStyle->strokeAVP(avp, osBuffer.pixbuffer, osBuffer.pxbRowstride,
								osBuffer.pxbWidth, osBuffer.pxbHeight,
								osBuffer.origin.x, osBuffer.origin.y,
								osBuffer.scaleFactor);
		}
	}
	//g_printerr("BezierPath::draw finished (%lX, %lX)\n", prev, next);
}

void BezierPath::drawBlack(OffscreenBuffer& osBuffer, RectangleArea& drawArea) {
	//g_printerr("BezierPath::drawBlack\n");
	BoundingBox bbox = getBoundingBox();
	if (drawArea.intersect(bbox)) {
		guint32 color = 0x000000FF;
		if (fillStyle && fillStyle->drawType != GDSTYLE_DRAWTYPE_TRANSPARENT) {
			//g_printerr("BezierPath::draw: fill\n");
			ArtVpath *avp = getVpath(osBuffer.scaleFactor, osBuffer.origin.x,
										osBuffer.origin.y, true, drawArea);
			ArtSVP *asvp = art_svp_from_vpath(avp);
			art_free(avp);
			art_rgb_svp_alpha(asvp, 0, 0,
						osBuffer.pxbWidth,
						osBuffer.pxbHeight,
						color,
						osBuffer.pixbuffer,
						osBuffer.pxbRowstride,
						NULL);
			art_svp_free(asvp);
		}
		if (drawStyle && drawStyle->drawType != GDSTYLE_DRAWTYPE_TRANSPARENT) {
			ArtVpath *avp = getVpath(osBuffer.scaleFactor, osBuffer.origin.x,
										osBuffer.origin.y, false, drawArea);
			if (drawStyle->lineType == GDSTYLE_LINETYPE_DASHED) {
				ArtVpathDash dash;
				dash.dash = drawStyle->dashes;
				for (int i = 0; i < 6; i++) {
					if (dash.dash[i] < 0) {
						dash.n_dash = i;
						i = 6;
					}
				}
				dash.offset = 0;
				if (dash.n_dash) {
					avp = art_vpath_dash(avp, &dash);
				}
			}
			ArtPathStrokeJoinType joins;
			ArtPathStrokeCapType caps;
			if (drawStyle->lineCaps == GDSTYLE_CAPS_ROUND) {
				caps = ART_PATH_STROKE_CAP_ROUND;
			} else if (drawStyle->lineCaps == GDSTYLE_CAPS_SQUARE) {
				caps = ART_PATH_STROKE_CAP_SQUARE;
			} else {
				caps = ART_PATH_STROKE_CAP_BUTT;
			}
			if (drawStyle->lineJoins == GDSTYLE_JOINS_BEVEL) {
				joins = ART_PATH_STROKE_JOIN_BEVEL;
			} else if (drawStyle->lineJoins == GDSTYLE_JOINS_ROUND) {
				joins = ART_PATH_STROKE_JOIN_ROUND;
			} else {
				joins = ART_PATH_STROKE_JOIN_MITER;
			}
			gdouble swidth = drawStyle->strokeWidth * osBuffer.scaleFactor;
			if (swidth < 0.1) swidth = 0.1;
			ArtSVP *asvp = art_svp_vpath_stroke(avp, 
									joins,
									caps,
									swidth, 
									drawStyle->lineMiterLimit, 0.5);
			art_free(avp);
			art_rgb_svp_alpha(asvp, 0, 0,
						osBuffer.pxbWidth,
						osBuffer.pxbHeight,
						color,
						osBuffer.pixbuffer,
						osBuffer.pxbRowstride,
						NULL);
			art_svp_free(asvp);
		}
	}
}

void BezierPath::fillBlack(OffscreenBuffer& osBuffer, RectangleArea& drawArea) {
	BoundingBox bbox = getBoundingBox();
	if (drawArea.intersect(bbox)) {
		ArtVpath *avp = getVpath(osBuffer.scaleFactor, osBuffer.origin.x,
									osBuffer.origin.y, true, drawArea);
		ArtSVP *asvp = art_svp_from_vpath(avp);
		art_free(avp);
		guint32 color = 0x000000FF;
		art_rgb_svp_alpha(asvp, 0, 0,
					osBuffer.pxbWidth,
					osBuffer.pxbHeight,
					color,
					osBuffer.pixbuffer,
					osBuffer.pxbRowstride,
					NULL);
		art_svp_free(asvp);
	}
}

void BezierPath::drawWireFrameSelection(DocWindow *docwin, RectangleArea& drawArea) {
	BoundingBox bbox = getBoundingBox();
	if (drawArea.intersect(bbox)) {
		Layer *myLayer = docwin->doc->findObjectLayer(this);
		guint color = 0xFFFFFFFF;
		if (myLayer) {
			color = myLayer->getSelectionColor();
		} else {
			g_printerr("BezierPath::drawWireFrameSelection: no layer found for path %lX\n", (long) this);
		}
		//g_printerr("BezierPath::drawWireFrameSelection got color\n");
		Gdk_Colormap cmap = Gdk_Colormap::get_system();
		Gdk_Color gcolor;
		gcolor.set_rgb(((color >> 24) & 0xFF) << 8, 
						((color >> 16) & 0xFF) << 8,
						((color >> 8) & 0xFF) << 8);
		cmap.alloc(gcolor);
		Gdk_Window drawable = docwin->DocWinDrawingArea->get_window();
		Gdk_GC context = Gdk_GC(drawable);
		context.set_foreground(gcolor);
		context.set_line_width(1);
		context.set_function(GDK_COPY);
		drawWireFrame(docwin, context, drawArea);
		drawPoints(docwin, context, drawArea);
	}
}

GDStyle *BezierPath::getStrokeStyle() {
	return drawStyle;
}

GDStyle *BezierPath::getFillStyle() {
	return fillStyle;
}

void BezierPath::gatherResource(GDResourceList& collection, gint type) {
	if (drawStyle) {
		drawStyle->gatherResource(collection, type);
	}
	if (fillStyle) {
		fillStyle->gatherResource(collection, type);
	}
}

void BezierPath::drawWithSelection(OffscreenBuffer& osBuffer, RectangleArea& drawArea) {
	//g_printerr("BezierPath::drawWithSelection\n");
	BoundingBox bbox = getBoundingBox();
	if (drawArea.intersect(bbox)) {
		Layer *myLayer = doc->findObjectLayer(this);
		guint color = 0xFFFFFFFF;
		if (myLayer) {
			color = myLayer->getSelectionColor();
		} else {
			g_printerr("BezierPath::drawWithSelection: no layer found for path %lX\n", (long) this);
		}
		//g_printerr("BezierPath::drawWithSelection got color\n");
		ArtVpath *avp = getVpath(osBuffer.scaleFactor, osBuffer.origin.x,
								osBuffer.origin.y, false, drawArea);
		ArtSVP *asvp = art_svp_vpath_stroke(avp,
									ART_PATH_STROKE_JOIN_MITER,
									ART_PATH_STROKE_CAP_BUTT,
									1, 4, 0.5);
		art_free(avp);
		art_rgb_svp_alpha(asvp, 0, 0,
						osBuffer.pxbWidth,
						osBuffer.pxbHeight,
						color,
						osBuffer.pixbuffer,
						osBuffer.pxbRowstride,
						NULL);
		art_svp_free(asvp);
		
		BezierNode	*node = nodes;
		
		while (node) {
			if (node->selected() && (! node->prev || ! node->prev->selected())
							&& (! node->next || ! node->next->selected())) {
				// single node selected => show both handles
				node->bufferDrawPreviousHandle(osBuffer, color);
				node->bufferDrawNextHandle(osBuffer, color);
			} else if (! node->selected()) {
				// node itself is not selected
				if (node->prev && node->prev->selection) {
					node->bufferDrawPreviousHandle(osBuffer, color);
				}
				if (node->selection == 2 || (node->next && node->next->selected())) {
					node->bufferDrawNextHandle(osBuffer, color);
				}
			}
			/* draw the anchor */
			node->bufferDrawPoint(osBuffer, color);
			node = node->next;
			if (node == nodes) break; // special case closed path
		}
	}
}

void BezierPath::drawPoints(DocWindow *docwin, Gdk_GC& context, RectangleArea& drawArea) {
	BezierNode	*node = nodes;
	gint32		numpts;
	
	while (node) {
		/* draw the anchor */
		node->drawPoint(docwin, context);
		if (node->selected() && (! node->prev || ! node->prev->selected())
						&& (! node->next || ! node->next->selected())) {
			// single node selected => show both handles
			node->drawPreviousHandle(docwin, context);
			node->drawNextHandle(docwin, context);
		} else if (! node->selected()) {
			// node itself is not selected
			if (node->prev && node->prev->selection) {
				node->drawPreviousHandle(docwin, context);
			}
			if (node->selection == 2 || (node->next && node->next->selected())) {
				node->drawNextHandle(docwin, context);
			}
		}
		node = node->next;
		if (node == nodes) break; // special case closed path
	}
}

BezierPathVector *BezierPath::getBezierPaths() {
	BezierPathVector *pvec = new BezierPathVector();
	pvec->push_back(this);
	return pvec;
}

BoundingBox BezierPath::getBoundingBox() {
	BezierNode	*n = nodes;
	BoundingBox bb;
		
	while (n) {
		n->boundingBox(bb.left, bb.top, bb.right, bb.bottom);
		n = n->next;
		if (n == nodes) break; // special case closed path
	}
	return bb;
}

gdouble BezierPath::attractPoint(gdouble x, gdouble y, gdouble& ax, gdouble& ay, gdouble tolerance, int attractType) {
	gdouble nearest = G_MAXDOUBLE;
	BezierNode *list = nodes;
	while (list) {
		if (list->checkPoint(x, y, tolerance)) {
			gdouble dx = fabs(list->onCurve.x - x);
			gdouble dy = fabs(list->onCurve.y - y);
			gdouble sqrdist = dx * dx + dy * dy;
			if (sqrdist < nearest) {
				nearest = sqrdist;
				ax = list->onCurve.x;
				ay = list->onCurve.y;
			}
		}
		list = list->next;
		if (list == nodes) break; // special case closed path
	}
	if (drawStyle && drawStyle->isGuide) {
		// in the case of a guide check the lines between the nodes too
		gdouble sqrdist = 0, tax = 0, tay = 0;
		if (doCheckCurves(x, y, tolerance, tax, tay, sqrdist)) {
			if (sqrdist < tolerance * tolerance && sqrdist < nearest) {
				nearest = sqrdist;
				ax = tax;
				ay = tay;
			}
		}
	}
	return nearest;
}

SelectableItem *BezierPath::checkPoints(gdouble x, gdouble y, gdouble tolerance, gboolean onlyCurvePoints) {
	if (drawStyle && drawStyle->isGuide && doc->guidesLocked) return NULL;
	BezierNode *found = NULL;
	BezierNode *list = nodes;
	while (list) {
		if (list->checkPoint(x, y, tolerance)) {
			found = list;
			list->selectedItem = 0;
		} else if (! onlyCurvePoints){
			if (list->testPreviousHandleVisible() && (list->selectedItem = list->checkHandles(x, y, tolerance)) == -1) {
				found = list;
			} else if (list->testNextHandleVisible() && (list->selectedItem = list->checkHandles(x, y, tolerance)) == 1) {
				found = list;
			}
		}
		list = list->next;
		if (list == nodes) break; // special case closed path
	}
	return found;
}

GObject *BezierPath::checkArea(gdouble x, gdouble y) {
	if (fillStyle && (fillStyle->drawType == GDSTYLE_DRAWTYPE_TRANSPARENT || fillStyle->isGuide)) {
		// don't select by area if not filled
		return NULL;
	}
	RectangleArea dummy;
	ArtVpath *vpath = getVpath(1, 0, 0, true, dummy);
	ArtSVP *asvp = art_svp_from_vpath(vpath);
	art_free(vpath);
	int winding = art_svp_point_wind(asvp, x, y);
	art_svp_free(asvp);
	//g_printerr("BezierPath::checkArea %lX: winding=%d\n", this, winding);
	if (winding) {
		return this;
	}
	return NULL;
}

SelectableItem *BezierPath::checkCurves(gdouble x, gdouble y, gdouble tolerance) {
	if (drawStyle && drawStyle->isGuide && doc->guidesLocked) return NULL;
	gdouble ax = x, ay = y;
	gdouble sqrdist = 0;
	SelectableItem *si = doCheckCurves(x, y, tolerance, ax, ay, sqrdist);
	if (sqrdist < tolerance * tolerance) {
		return si;
	}
	return NULL;
}

SelectableItem *BezierPath::doCheckCurves(gdouble x, gdouble y, gdouble tolerance, gdouble& ax, gdouble& ay, gdouble& sqrdist) {
	BezierNode *found = NULL;
	BezierNode *list = nodes;
	sqrdist = G_MAXDOUBLE;
	while (list && list->next) {
		// check curve bounding box
		gdouble left, right, top, bottom;
		/* initialize the values */
		left = G_MAXDOUBLE;
		top = G_MAXDOUBLE;
		right = G_MINDOUBLE;
		bottom = G_MINDOUBLE;
		list->boundingBox(left, top, right, bottom);
		list->next->boundingBox(left, top, right, bottom);
		//g_printerr("BezierPath::checkCurves %e, %e, %e, %e (%e, %e), %e\n", left, top, right, bottom, x, y, tolerance);
		if (x > left - tolerance && x < right + tolerance && y > top - tolerance && y < bottom + tolerance) {
			//g_printerr("checking curve %e, %e, %e, %e (%e, %e)\n", left, top, right, bottom, x, y);
			list->calcSegment(list->next, SUBDIVISIONS, segpoints);
			for (gint32 i = 0; i < SUBDIVISIONS - 1; i++) {
				// check line bounding box
				gdouble l, r, t, b;
				if (segpoints[i].x <= segpoints[i+1].x) {
					l = segpoints[i].x;
					r = segpoints[i+1].x;
				} else {
					l = segpoints[i+1].x;
					r = segpoints[i].x;
				}
				if (segpoints[i].y <= segpoints[i+1].y) {
					t = segpoints[i].y;
					b = segpoints[i+1].y;
				} else {
					t = segpoints[i+1].y;
					b = segpoints[i].y;
				}
				//g_printerr("check line %d, %d, %d, %d (%e, %e)\n", l, t, r, b, x, y);
				if (x > l - tolerance && x < r + tolerance && y > t - tolerance && y < b + tolerance) {
					// check distance to line
					gdouble dx = segpoints[i+1].x - segpoints[i].x;
					gdouble dy = segpoints[i+1].y - segpoints[i].y;
					gdouble m = dy / dx;
					gdouble xs, ys; // coordinates of crossing point
					gdouble n1, n2;
					// allow extrema
					if (fabs(m) < 0.0001) {
						// almost horizontal
						xs = x;
						ys = segpoints[i].y;
					} else if (fabs(m) > 10000.0) {
						// almost vertical
						xs = segpoints[i].x;
						ys = y;
					} else {
						// normal case
						// n1: cross point of the segpoint line with the x-axis
						// n2: cross point of the right-angled line with the x-axis
						n1 = segpoints[i].x - segpoints[i].y / m;
						n2 = x + y * m;
						ys = (m * (n2 - n1)) / (1 + m * m);
						xs = n1 + (ys / m);
					}
					// if the crossing point is outside the direct line connecting
					// our segpoints move it to the nearest segpoint
					if (! (xs >= l && xs <= r && ys >= t && ys <= b)) {
						//g_printerr("Crosspoint outside rectangle (%f,%f,%f,%f) (%f,%f)\n", l, t, r, b, xs, ys);
						if (fabs(xs - segpoints[i].x) < fabs(xs - segpoints[i+1].x) &&
							fabs(ys - segpoints[i].y) < fabs(ys - segpoints[i+1].y)) {
							xs = segpoints[i].x;
							ys = segpoints[i].y;
						} else {
							xs = segpoints[i+1].x;
							ys = segpoints[i+1].y;
						}
					}
					gdouble lineportion;
					if (fabs(dx) > fabs(dy)) {
						lineportion = (xs - segpoints[i].x) / dx;
					} else {
						lineportion = (ys - segpoints[i].y) / dy;
					}
					gdouble deltax = x - xs;
					gdouble deltay = y - ys;
					gdouble distsquare = deltax * deltax + deltay * deltay;
					//g_printerr("segpoints %f,%f  %f,%f \n", segpoints[i].x, segpoints[i].y, segpoints[i+1].x, segpoints[i+1].y);
					//g_printerr("check cross point m=%f, n1=%f, n2=%f, S(%f, %f) Mouse(%f, %f) i=%d\n", m, n1, n2, xs, ys, x, y, i);
					//g_printerr("mindistsquare: %f distsquare: %f, tolerancesquare: %f, lineportion: %f\n", mindistsquare, distsquare, tolerance * tolerance, lineportion);
					if (distsquare < sqrdist) {
						sqrdist = distsquare;
						ax = xs;
						ay = ys;
						found = list;
						found->curveDivision = (i + lineportion) / SUBDIVISIONS;
					}
				}
			}
		}
		list = list->next;
		if (list == nodes) break; // special case closed path
	}
	//g_printerr("check curves result: %lX\n", found);
	return found;
}

BezierNode *BezierPath::getStartNode() {
	return nodes;
}

BezierNode *BezierPath::getEndNode() {
	BezierNode *node = nodes;
	while (node && node->next) {
		if (node->next == nodes) break; //closed path
		node = node->next;
	}
	return node;
}

void BezierPath::selectRectangle(gdouble left, gdouble top, gdouble right, gdouble bottom, gint what) {
	if (drawStyle && drawStyle->isGuide && doc->guidesLocked) return;
	//g_printerr("BezierPath::selectRectangle what:%d shiftSelect:%d\n", what, what & 128);
	BezierNode *node = nodes;
	while (node) {
		if (node->checkRectangle(left, top, right, bottom)) {
			if ((what & 3) == 1) {
				select((what & 128) != 0);
				break;
			} else if ((what & 3) == 2) {
				selectObject((what & 128) != 0);
				break;
			} else {
				node->select((what & 128) != 0);
			}
		}
		node = node->next;
		if (node == nodes) break; // special case closed path
	}
}

void BezierPath::deSelectRectangle(gdouble left, gdouble top, gdouble right, gdouble bottom, gint what) {
	//g_printerr("BezierPath::deSelectRectangle what:%d\n", what);
	BezierNode *node = nodes;
	while (node) {
		if (node->checkRectangle(left, top, right, bottom)) {
			if (what & 3 == 1) {
				deSelect();
				break;
			} else if (what & 3 == 2) {
				deSelectObject();
				break;
			} else {
				node->deSelect();
			}
		}
		node = node->next;
		if (node == nodes) break; // special case closed path
	}
}

void BezierPath::doSelectAll() {
	BezierNode *node = nodes;
	while (node) {
		node->selection = 1;
		node = node->next;
		if (node == nodes) break; // special case closed path
	}
}

void BezierPath::selectAll() {
	if (drawStyle && drawStyle->isGuide && doc->guidesLocked) return;
	//g_printerr("BezierPath::selectAll() %lX\n", this);
	doSelectAll();
	doc->selection->insertObject(this);
}

void BezierPath::doDeSelectAll() {
	BezierNode *node = nodes;
	while (node) {
		node->selection = 0;
		node = node->next;
		if (node == nodes) break; // special case closed path
	}
}

void BezierPath::deSelectAll() {
	//g_printerr("BezierPath::deSelectAll() %lX\n", this);
	doDeSelectAll();
	doc->selection->removeObject(this);
}

GObject *BezierPath::findObject(GObject *object) {
	if (object == this) {
		return this;
	} else {
		return NULL;
	}
}

GObject *BezierPath::removeObject(GObject *object) {
	if (object == this) {
		if (next) {
			next->prev = prev;
		}
		if (prev) {
			prev->next = next;
		}
		return this;
	} else {
		return NULL;
	}
}

void BezierPath::reverseOrder() {
	BezierNode *node = nodes;
	BezierNode *last = NULL;
	while (node) {
		BezierNode *tmp = node->prev;
		node->prev = node->next;
		node->next = tmp;
		last = node;
		node->reverseOrder();
		node = node->prev;
		if (node == nodes) break; // special case closed path
	}
	nodes = last;
}

void BezierPath::setDocument(GDDocument *document) {
	if (doc != document) {
		// this really is a change of the document
		if (fillStyle) {
			fillStyle->changeDocument(document);
		}
	}
	GObject::setDocument(document);
}

void BezierPath::drawSelection(DocWindow *docwin, Gdk_GC& context, RectangleArea& drawArea) {
	//g_printerr("BezierPath::drawSelection this: %lX docwin: %lX\n", this, docwin);
	BezierNode *node = nodes;
	while (node) {
		node->drawSelection(docwin, context, drawArea);
		node = node->next;
		if (node == nodes) break;
	}
}

gboolean BezierPath::selected() {
	gboolean selected = true;
	BezierNode *node = nodes;
	while (node) {
		if (! node->selected()) {
			selected = false;
			break;
		}
		node = node->next;
		if (node == nodes) break;
	}
	return selected;
}

gboolean BezierPath::partiallySelected() {
	gboolean selected = false;
	BezierNode *node = nodes;
	while (node) {
		if (node->selection) {
			selected = true;
			break;
		}
		node = node->next;
		if (node == nodes) break;
	}
	return selected;
}

gint BezierPath::getSelection(SelectionItemList& selectionList, gint mode) {
	if (selected()) {
		selectionList.insert(selectionList.end(), new SelectionItem(this, mode));
		return 1;
		//g_printerr("Whole path selected\n");
	} else {
		SelectionItem *last = NULL;
		SelectionItem *first = NULL;
		BezierNode *node = nodes;
		gint selectedNodes = 0;
		while (node) {
			if (node->selection) {
				//g_printerr("Selecting node (%f, %f)\n", node->onCurve.x, node->onCurve.y);
				SelectionItem *curr = new SelectionItem(node, mode);
				selectionList.insert(selectionList.end(), curr);
				selectedNodes++;
				if (node != nodes && node->prev && node->prev->selected() && last) {
					//g_printerr("Connecting selection node to previous curr=%lX last=%lX last->moving=%lX curr->moving=%lX\n", curr, last, last->moving, curr->moving);
					if (last->moving) {
						((BezierNode *)last->moving)->next = ((BezierNode *)curr->moving);
					}
					if (curr->moving) {
						((BezierNode *)curr->moving)->prev = ((BezierNode *)last->moving);
					}
				}
				last = curr;
				if (node == nodes) first = curr;
			}
			node = node->next;
			if (node == nodes) {
				if (node->selected() && node->prev && 
						node->prev->selected() && last && first) {
					//g_printerr("Connection over path begin/end\n");
					if (last->moving) {
						((BezierNode *)last->moving)->next = ((BezierNode *)first->moving);
					}
					if (first->moving) {
						((BezierNode *)first->moving)->prev = ((BezierNode *)last->moving);
					}
					
				}
				break;
			}
		}
		return selectedNodes;
	}
	//g_printerr("Selection finished\n");
}

void BezierPath::writePSPathDescription(ostream& psout) {
	gdouble dx = doc->nullPoint.x;
	gdouble dy = doc->nullPoint.y;
	gdouble pfx = doc->printFactor.x;
	gdouble pfy = doc->printFactor.y;
	gdouble ph = doc->pageHeight;
	BezierNode *node = nodes;
	if (node) {
		psout << (node->onCurve.x - dx) * pfx << ' ';
		psout << (ph - node->onCurve.y + dy) * pfy << " m\n";
		node = node->next;
	}
	while (node) {
		psout << (node->prev->nextHandle.x - dx) * pfx << ' ';
		psout << (ph - node->prev->nextHandle.y + dy) * pfy << ' ';
		psout << (node->prevHandle.x - dx) * pfx << ' ';
		psout << (ph - node->prevHandle.y + dy) * pfy << ' ';
		psout << (node->onCurve.x - dx) * pfx << ' ';
		psout << (ph - node->onCurve.y + dy) * pfy << " c\n";
		if (node == nodes) {
			// closed path
			psout << "cp\n";
			break; 
		}
		node = node->next;
	}	
}

gint BezierPath::writePS(ostream& psout) {
	if (! nodes) return 0;
	psout << "np\n";
	writePSPathDescription(psout);
	fillStyle->writeFillPS(psout, doc);
	drawStyle->writeStrokePS(psout, doc);
	return 0;
}

void BezierPath::writeXMLPathDescription(ostream& xmlout) {
	gdouble dx = doc->nullPoint.x;
	gdouble dy = doc->nullPoint.y;
	BezierNode *node = nodes;
	if (node) {
		xmlout << "M " << node->onCurve.x - dx << ' ' << node->onCurve.y - dy;
		node = node->next;
	} 
	while (node) {
		xmlout << " C " << node->prev->nextHandle.x -dx << ' ' << node->prev->nextHandle.y - dy;
		xmlout << ' ' << node->prevHandle.x - dx << ' ' << node->prevHandle.y - dy;
		xmlout << ' ' << node->onCurve.x - dx << ' ' << node->onCurve.y - dy;
		node = node->next;
		if (node == nodes) {
			// closed path
			xmlout << " C " << node->prev->nextHandle.x - dx << ' ' << node->prev->nextHandle.y - dy;
			xmlout << ' ' << node->prevHandle.x - dx << ' ' << node->prevHandle.y - dy;
			xmlout << ' ' << node->onCurve.x - dx << ' ' << node->onCurve.y - dy;
			xmlout << " Z";
			break;
		}
	}
}

gint BezierPath::writeXML(ostream& xmlout) {
	fillStyle->writeFillPreparationXML(xmlout, doc);
	xmlout << "<path d=\"";
	writeXMLPathDescription(xmlout);
	xmlout << "\"";
	fillStyle->writeFillXML(xmlout, doc);
	drawStyle->writeStrokeXML(xmlout, doc);
	xmlout << "/>\n";
	return 0;
}

const char *nextAlpha(const char *str) {
	while (*str && ! (*str >= 'A' && *str <= 'Z' || *str >= 'a' && *str <= 'z')) 
		str++;
	return str;
}

const char *BezierPath::constructPathFromSVG(const char *description) {
	//g_printerr("BezierPath::constructPathFromSVG(%s)\n", description);
	gdouble dx = doc->nullPoint.x;
	gdouble dy = doc->nullPoint.y;
	BezierNode *node = new BezierNode(this);
	addNode(node);
	gdouble x,y;
	sscanf(description++, "M %lf %lf", &x, &y);
	//g_printerr("Start point: %f, %f\n", x, y);
	x += dx; y += dy;
	node->setPreviousHandle(x, y);
	node->setOnCurvePoint(x, y);
	node->setNextHandle(x, y);
	description = nextAlpha(description);
	while (*description == 'C' || *description == 'L') {
		gdouble lnx, lny, px, py, x, y;
		if (*description == 'C') {
			sscanf(description++, "C %lf %lf %lf %lf %lf %lf", &lnx, &lny, &px, &py, &x, &y);
		} else if (*description == 'L') {
			sscanf(description++, "L %lf %lf", &x, &y);
			lnx = node->onCurve.x - dx;
			lny = node->onCurve.y - dy;
			px = x;
			py = y;
		}
		//g_printerr("Curve element: %f, %f   %f, %f   %f, %f\n", lnx, lny, px, py, x, y);
		lnx += dx; lny += dy; x += dx; y += dy; px += dx; py += dy;
		node->setNextHandle(lnx, lny);
		node = new BezierNode(this);
		addNode(node);
		node->setOnCurvePoint(x, y);
		//g_printerr("BezierPath::constructPathFromSVG onCurve= %f,%f\n", x, y);
		node->setPreviousHandle(px, py);
		description = nextAlpha(description);
	}
	if (*description == 'Z' || *description == 'z') {
		// path is closed
		removeNode(node);
		nodes->prev = node->prev;
		nodes->setPreviousHandle(node->prevHandle.x, node->prevHandle.y);
		if (node->prev) {
			node->prev->next = nodes;
		}
		description++;
		description = nextAlpha(description);
	} else if (description && *description) {
		g_printerr("BezierPath::constructPathFromSVG: got unknown character in Path description: %c\n", *description);
	}
	return description;
}

void BezierPath::transformByMatrix(TransformationMatrix& m, gboolean incFillPattern) {
	BezierNode *node = nodes;
	while (node) {
		node->transformByMatrix(m);
		node = node->next;
		if (node == nodes) break;
	}
	// transform the styles
	if (drawStyle && incFillPattern) drawStyle->transformByMatrix(m);
	if (fillStyle && incFillPattern) fillStyle->transformByMatrix(m);
}

void BezierPath::setFillMatrix(TransformationMatrix& m) {
	if (fillStyle) fillStyle->tMatrix = m;
}

SelectableItem *BezierPath::clone() {
	BezierPath *clone = new BezierPath(doc);
	*(clone->drawStyle) = *drawStyle;
	*(clone->fillStyle) = *fillStyle;
	BezierNode *node = nodes;
	while (node) {
		BezierNode *nodeClone = (BezierNode *)node->clone();
		nodeClone->path = clone;
		nodeClone->prev = NULL;
		nodeClone->next = NULL;
		clone->addNode(nodeClone);
		node = node->next;
		if (node == nodes) {
			// special case closed path, close clone too
			nodeClone->next = clone->nodes;
			clone->nodes->prev = nodeClone;
			break;
		}
	}
	return clone;
}

void BezierPath::setCoords(SelectableItem* from) {
	BezierPath *f = (BezierPath *)from;
	BezierNode *node = nodes;
	BezierNode *fromnode = f->nodes;
	while (node) {
		node->setCoords(fromnode);
		fromnode = fromnode->next;
		node = node->next;
		if (node == nodes) break; // special case closed path
	}
	// set the changed matrix of the styles too
	if (drawStyle && f->drawStyle) {
		drawStyle->tMatrix = f->drawStyle->tMatrix;
	}
	if (fillStyle && f->fillStyle) {
		fillStyle->tMatrix = f->fillStyle->tMatrix;
	}
}

void BezierPath::clear() {
	deSelectAll();
	doc->removeObject(this);
	delete this;
}

gchar *BezierPath::identify() {
	return "BezierPath";
}

BezierPath::~BezierPath() {
	//g_printerr("Removing BezierPath %lX\n", this);
	/* deSelectAll(); Removed from Destructor for clean-up */
	BezierNode *node = nodes;
	while (node) {
		BezierNode *next = node->next;
		delete node;
		node = next;
		if (node == nodes) break; // special case closed path
	}
	// delete the styles
	delete drawStyle;
	delete fillStyle;
}
