/* $Id: CircleTool.cc,v 1.6 2004/03/24 19:31:42 pklaus Exp $
 *
 * CircleTool.cc
 *
 * Gestalter - The free vector drawing application
 * Copyright (c) 2002 Philipp Klaus <gestalter@linotux.ch>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include <math.h>
#include "../pixmaps/circle.xpm"
#include "../cursors/crosshair.xbm"
#include "../cursors/crosshair_mask.xbm"
#include "../cursors/move.xbm"
#include "../cursors/move_mask.xbm"
#include "../cursors/move_magnet.xbm"
#include "../cursors/move_magnet_mask.xbm"
#include "CircleTool.hh"
#include "GDDocument.hh"
#include "BezierPath.hh"
#include "BezierNode.hh"
#include "DocWindow.hh"
#include "CursorFactory.hh"
#include "AddAction.hh"
#include "DimensionDialog.hh"

#define doubleMAX(a,b)	(fabs(a) > fabs(b)) ? fabs(a) : fabs(b)

/* handle to radius factor
   hr = (4 * sqrt(2) - 4) / 3
*/
#define hr		0.5522847498307933
#define sqrt2	1.4142135623730950
#define hsqrt2	1.2071067811865475

CircleTool::CircleTool() {
	pixmapData = circle_xpm;
	toolName = "CircleTool";
	moving = false;
	nodeCaught = false;
	valueDialog = NULL;
	dialogDoc = NULL;
	CursorFactory::createCursorFromData(crosshairCursor,
					crosshair_bits, crosshair_mask_bits,
					crosshair_width, crosshair_height,
					crosshair_x_hot, crosshair_y_hot);
	CursorFactory::createCursorFromData(moveCursor,
					move_bits, move_mask_bits,
					move_width, move_height,
					move_x_hot, move_y_hot);
	CursorFactory::createCursorFromData(moveMagnetCursor,
					move_magnet_bits, move_magnet_mask_bits,
					move_magnet_width, move_magnet_height,
					move_magnet_x_hot, move_magnet_y_hot);
}

void CircleTool::activate() {
	Tool::activate();
}

void CircleTool::deactivate() {
	Tool::deactivate();
}

gint CircleTool::buttonDown(GdkEventButton *ev, DocWindow *docwin) {
	GDDocument	*doc;
	downButton = ev->button;

	if (downButton == 1) {
		doc = docwin->doc;
		doc->deSelectAll();
		if (ev->state & GDK_MOD1_MASK) {
			sx = ev->x / docwin->osBuffer.scaleFactor + docwin->osBuffer.origin.x;
			sy = ev->y / docwin->osBuffer.scaleFactor + docwin->osBuffer.origin.y;
			magnet(sx, sy, docwin);
			dialogDoc = doc;
			showValueDialog();
		} else {			
			sx = ev->x / docwin->osBuffer.scaleFactor + docwin->osBuffer.origin.x;
			sy = ev->y / docwin->osBuffer.scaleFactor + docwin->osBuffer.origin.y;
			gboolean caught = magnet(sx, sy, docwin);
			cx = sx;
			cy = sy;
			startx = (gint) ev->x;
			starty = (gint) ev->y;
			oldx = startx;
			oldy = starty;
			moving = true;
		}
	} else {
		Tool::buttonDown(ev, docwin);
	}
	return 0;
}

gint CircleTool::motion(GdkEventMotion *ev, DocWindow *docwin) {
	if (ev->state & GDK_BUTTON1_MASK) {
		GDDocument *doc = docwin->doc;
		Gdk_Window drawable = docwin->DocWinDrawingArea->get_window();
		doc->context.set_function(GDK_INVERT);
		gdouble dx = ev->x / docwin->osBuffer.scaleFactor + docwin->osBuffer.origin.x;
		gdouble dy = ev->y / docwin->osBuffer.scaleFactor + docwin->osBuffer.origin.y;
		nodeCaught = magnet(dx, dy, docwin);
		setCursor(ev->state, docwin);
		if ((ev->state & GDK_SHIFT_MASK) && ! (ev->state & GDK_CONTROL_MASK)) {
			// circle (cursor on curve special case is handled below)
			gdouble delta = doubleMAX(dx-cx, dy-cy);
			dx = cx + copysign(delta, dx-cx);
			dy = cy + copysign(delta, dy-cy);
		}
		if (ev->state & GDK_MOD1_MASK) {
			// grow from center
			if (ev->state & GDK_CONTROL_MASK) {
				// cursor on curve
				if (ev->state & GDK_SHIFT_MASK) {
					// circle
					gdouble r = sqrt((dx - cx) * (dx - cx) + (dy - cy) * (dy - cy));
					dx = cx + r;
					dy = cy + r;
				} else {
					dx = cx + (dx - cx) * sqrt2;
					dy = cy + (dy - cy) * sqrt2;
				}
			}
			sx = cx - (dx - cx);
			sy = cy - (dy - cy);
		} else {
			if (ev->state & GDK_CONTROL_MASK) {
				// cursor on curve
				if (ev->state & GDK_SHIFT_MASK) {
					// circle
					gdouble r = sqrt((dx - cx) * (dx - cx) + (dy - cy) * (dy - cy)) / 2;
					gdouble centerx = cx + (dx - cx) / 2;
					gdouble centery = cy + (dy - cy) / 2;
					sx = centerx - r;
					sy = centery - r;
					dx = centerx + r;
					dy = centery + r;
				} else {
					gdouble odx = dx;
					gdouble ody = dy;
					dx = cx + (dx - cx) * hsqrt2;
					dy = cy + (dy - cy) * hsqrt2;
					sx = cx - dx + odx;
					sy = cy - dy + ody;
				}
			} else {
				sx = cx;
				sy = cy;
			}
		}
		gint x = (gint) ((dx - docwin->osBuffer.origin.x) * docwin->osBuffer.scaleFactor);
		gint y = (gint) ((dy - docwin->osBuffer.origin.y) * docwin->osBuffer.scaleFactor);
		drawCircle(drawable, doc->context, startx, starty, oldx, oldy);
		startx = (gint) ((sx - docwin->osBuffer.origin.x) * docwin->osBuffer.scaleFactor);
		starty = (gint) ((sy - docwin->osBuffer.origin.y) * docwin->osBuffer.scaleFactor);
		drawCircle(drawable, doc->context, startx, starty, x, y);
		oldx = x;
		oldy = y;
		
	} else if (ev->state & (GDK_BUTTON2_MASK | GDK_BUTTON3_MASK)) {
		Tool::motion(ev, docwin);
	} else {
		setCursor(ev->state, docwin);
	}
	
  return 0;
}

gint CircleTool::buttonUp(GdkEventButton *ev, DocWindow *docwin) {
	if (downButton == 1) {
		GDDocument *doc = docwin->doc;
		Gdk_Window drawable = docwin->DocWinDrawingArea->get_window();
		gdouble dx = ev->x / docwin->osBuffer.scaleFactor + docwin->osBuffer.origin.x;
		gdouble dy = ev->y / docwin->osBuffer.scaleFactor + docwin->osBuffer.origin.y;
		gboolean caught = magnet(dx, dy, docwin);
		if ((ev->state & GDK_SHIFT_MASK) && ! (ev->state & GDK_CONTROL_MASK)) {
			// circle (cursor on curve special case is handled below)
			gdouble delta = doubleMAX(dx-cx, dy-cy);
			dx = cx + copysign(delta, dx-cx);
			dy = cy + copysign(delta, dy-cy);
		}
		if (ev->state & GDK_MOD1_MASK) {
			// grow from center
			if (ev->state & GDK_CONTROL_MASK) {
				// cursor on curve
				if (ev->state & GDK_SHIFT_MASK) {
					// circle
					gdouble r = sqrt((dx - cx) * (dx - cx) + (dy - cy) * (dy - cy));
					dx = cx + r;
					dy = cy + r;
				} else {
					dx = cx + (dx - cx) * sqrt2;
					dy = cy + (dy - cy) * sqrt2;
				}
			}
			sx = cx - (dx - cx);
			sy = cy - (dy - cy);
		} else {
			if (ev->state & GDK_CONTROL_MASK) {
				// cursor on curve
				if (ev->state & GDK_SHIFT_MASK) {
					// circle
					gdouble r = sqrt((dx - cx) * (dx - cx) + (dy - cy) * (dy - cy)) / 2;
					gdouble centerx = cx + (dx - cx) / 2;
					gdouble centery = cy + (dy - cy) / 2;
					sx = centerx - r;
					sy = centery - r;
					dx = centerx + r;
					dy = centery + r;
				} else {
					gdouble odx = dx;
					gdouble ody = dy;
					dx = cx + (dx - cx) * hsqrt2;
					dy = cy + (dy - cy) * hsqrt2;
					sx = cx - dx + odx;
					sy = cy - dy + ody;
				}
			} else {
				sx = cx;
				sy = cy;
			}
		}
		addCircle(doc, sx, sy, dx, dy);
		// redraw the documentList
		doc->context.set_function(GDK_SET);
		doc->redraw();
		moving = false;
	} else {
		Tool::buttonUp(ev, docwin);
	}
	return 0;
}

void CircleTool::addCircle(GDDocument *doc, gdouble tlx, gdouble tly, gdouble brx, gdouble bry) {
		gdouble mx = (tlx + brx) / 2;
		gdouble my = (tly + bry) / 2;
		gdouble xh = (mx - tlx) * hr;
		gdouble yh = (my - tly) * hr;
		BezierPath *path = new BezierPath(doc);
		BezierNode *n1 = new BezierNode(path);
		n1->setOnCurvePoint(mx, tly);
		n1->setPreviousHandle(mx - xh, tly);
		n1->setNextHandle(mx + xh, tly);
		path->addNode(n1);
		BezierNode *n2 = new BezierNode(path);
		n2->setOnCurvePoint(brx, my);
		n2->setPreviousHandle(brx, my - yh);
		n2->setNextHandle(brx, my + yh);
		path->addNode(n2);
		BezierNode *n3 = new BezierNode(path);
		n3->setOnCurvePoint(mx, bry);
		n3->setPreviousHandle(mx + xh, bry);
		n3->setNextHandle(mx - xh, bry);
		path->addNode(n3);
		BezierNode *n4 = new BezierNode(path);
		n4->setOnCurvePoint(tlx, my);
		n4->setPreviousHandle(tlx, my + yh);
		n4->setNextHandle(tlx, my - yh);
		path->addNode(n4);
		n4->next = n1;
		n1->prev = n4; // close the path
		// create an action event
		AddAction *newAction = new AddAction(path, doc->activeLayer);
		doc->setAction(newAction);
		path->selectAll();
		doc->addPath(path);
		doc->currPath = NULL;
}


void CircleTool::drawCircle(Gdk_Drawable& drawable, Gdk_GC& context,
									gint x1, gint y1,
									gint x2, gint y2) {
	gint x,y,w,h;
	
	// there seems to be a bug in XDrawArc not to draw circles with
	// width or height greater than 800 pixels.
	// Perhaps I should later rewrite this routine to draw the circle with
	// bezier curves myself.
	
	if (x1 > x2) {
		x = x2;
		w = x1 - x2;
	} else {
		x = x1;
		w = x2 - x1;
	}
	if (y1 > y2) {
		y = y2;
		h = y1 - y2;
	} else {
		y = y1;
		h = y2 - y1;
	}
	//g_printerr("CircleTool::drawCircle (%d , %d , %d , %d)\n", x, y, w, h);
	drawable.draw_arc(context, 0, x, y, w, h, 0, 64 * 360);
}

void CircleTool::setCursor(guint state, DocWindow *docwin) {
	if (moving) {
		if (nodeCaught) {
			docwin->DocWinDrawingArea->get_window().set_cursor(moveMagnetCursor);
		} else {
			docwin->DocWinDrawingArea->get_window().set_cursor(moveCursor);
		}
	} else {
		if (state & GDK_MOD1_MASK) {
			docwin->DocWinDrawingArea->get_window().set_cursor(crosshairCursor);
		} else {
			docwin->DocWinDrawingArea->get_window().set_cursor(crosshairCursor);
		}
	}
}

gint CircleTool::deleteValueDialog(GdkEventAny *ev) {
	cancelValueDialog();
	return 1;
}

void CircleTool::cancelValueDialog() {
	valueDialog->hide();
}

void CircleTool::createObject() {
	gdouble w = fabs(valueDialog->getWidth());
	gdouble h = fabs(valueDialog->getHeight());
	valueDialog->hide();
	addCircle(dialogDoc, sx, sy, sx + w, sy + h);
	dialogDoc->redraw();
}

void CircleTool::showValueDialog() {
	if (! valueDialog) {
		valueDialog = new DimensionDialog(dialogDoc->application);
		valueDialog->getOKButton()->clicked.connect(SigC::slot(this, &CircleTool::createObject));
		valueDialog->getCancelButton()->clicked.connect(SigC::slot(this, &CircleTool::cancelValueDialog));
		valueDialog->delete_event.connect(SigC::slot(this, &CircleTool::deleteValueDialog));
		valueDialog->set_modal(true);
	}
	valueDialog->show();
}
