/* $Id: ClipGroup.cc,v 1.16 2004/07/21 06:31:20 pklaus Exp $
 *
 * ClipGroup.cc
 *
 * Gestalter - The free vector drawing application
 * Copyright (c) 2002 Philipp Klaus <gestalter@linotux.ch>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include "ClipGroup.hh"
#include "GDDocument.hh"
#include "SelectionItemList.hh"
#include "SelectableItem.hh"
#include "DocWindow.hh"
#include "BoundingBox.hh"
#include <libart_lgpl/art_rgb_rgba_affine.h>

guint ClipGroupNumber = 0;

ClipGroup::ClipGroup(GDDocument *document) : ObjectGroup(document) {
	clipparts = NULL;
	ClipGroupNumber++;
}


void ClipGroup::addClipPart(GObject *newObject) {
	newObject->parent = this;
	if (clipparts) {
		clipparts->insert(NULL, newObject);
	} else {
		clipparts = newObject;
	}
}

void ClipGroup::removeClipPart(GObject *object) {
	object->parent = NULL;
	clipparts->remove(object);
	if (clipparts == object) {
		clipparts = clipparts->next;
	}
}

// DrawObject methods
void ClipGroup::drawWithSelection(OffscreenBuffer& osBuffer, RectangleArea& drawArea) {
	ObjectGroup::drawWithSelection(osBuffer, drawArea);
	GObject *part = clipparts;
	while (part) {
		part->drawWithSelection(osBuffer, drawArea);
		part = part->next;
	}
}

void ClipGroup::drawWireFrameSelection(DocWindow *docwin, RectangleArea& drawArea) {
	ObjectGroup::drawWireFrameSelection(docwin, drawArea);
	GObject *part = clipparts;
	while (part) {
		part->drawWireFrameSelection(docwin, drawArea);
		part = part->next;
	}
}

void ClipGroup::gatherResource(GDResourceList& collection, gint type) {
	ObjectGroup::gatherResource(collection, type);
	GObject *part = clipparts;
	while (part) {
		part->gatherResource(collection, type);
		part = part->next;
	}
}


// GObject methods
void ClipGroup::draw(OffscreenBuffer& osBuffer, RectangleArea& drawArea) {
	//g_printerr("ClipGroup::draw: docwin=%lX\n", docwin);
	BoundingBox bbox = getBoundingBox();
	if (drawArea.intersect(bbox)) {
		OffscreenBuffer groupBuffer;
		groupBuffer.origin = osBuffer.origin;
		groupBuffer.scaleFactor = osBuffer.scaleFactor;
		groupBuffer.allocBuffer(osBuffer.pxbRowstride, osBuffer.pxbWidth, osBuffer.pxbHeight);
		groupBuffer.whiteBackground();
		GObject *part = parts;
		while (part) {
			part->draw(groupBuffer, drawArea);
			part = part->next;
		}
		
		
		OffscreenBuffer backBuffer;
		backBuffer.origin = osBuffer.origin;
		backBuffer.scaleFactor = osBuffer.scaleFactor;
		backBuffer.allocBuffer(osBuffer.pxbRowstride, osBuffer.pxbWidth, osBuffer.pxbHeight);
		backBuffer.whiteBackground();
		part = parts;
		while (part) {
			part->drawBlack(backBuffer, drawArea);
			part = part->next;
		}
		
		OffscreenBuffer clipBuffer;
		clipBuffer.origin = osBuffer.origin;
		clipBuffer.scaleFactor = osBuffer.scaleFactor;
		clipBuffer.allocBuffer(osBuffer.pxbRowstride, osBuffer.pxbWidth, osBuffer.pxbHeight);
		clipBuffer.whiteBackground();
		part = clipparts;
		while (part) {
			part->fillBlack(clipBuffer, drawArea);
			part = part->next;
		}
		guint pixSize = osBuffer.pxbSize / 3;
		guchar *apixbuffer = (guchar *) g_malloc(pixSize * 4);
		guchar *apb = apixbuffer;
		guchar *gb = groupBuffer.pixbuffer;
		for (guint i = 0; i < pixSize; i++) {
			// copy the rgb bytes
			*apb++ = *gb++;
			*apb++ = *gb++;
			*apb++ = *gb++;
			// and the alpha value
			*apb++ = ((256 - clipBuffer.pixbuffer[i*3]) * (256 - backBuffer.pixbuffer[i*3]) - 1) >> 8;
		}
		double affine[6] = { 1, 0, 0, 1, 0, 0 };
		gint width = osBuffer.pxbWidth;
		gint height = osBuffer.pxbHeight;
		art_rgb_rgba_affine(osBuffer.pixbuffer, 0, 0, 
							width, height,
							osBuffer.pxbRowstride, (art_u8 *) apixbuffer,
							width, height, width * 4, affine, 
							ART_FILTER_NEAREST, NULL);
	}
}

void ClipGroup::drawBlack(OffscreenBuffer& osBuffer, RectangleArea& drawArea) {
	BoundingBox bbox = getBoundingBox();
	if (drawArea.intersect(bbox)) {
		OffscreenBuffer groupBuffer;
		groupBuffer.origin = osBuffer.origin;
		groupBuffer.scaleFactor = osBuffer.scaleFactor;
		groupBuffer.allocBuffer(osBuffer.pxbRowstride, osBuffer.pxbWidth, osBuffer.pxbHeight);
		groupBuffer.whiteBackground();
		GObject *part = parts;
		while (part) {
			part->drawBlack(groupBuffer, drawArea);
			part = part->next;
		}
		OffscreenBuffer clipBuffer;
		clipBuffer.origin = osBuffer.origin;
		clipBuffer.scaleFactor = osBuffer.scaleFactor;
		clipBuffer.allocBuffer(osBuffer.pxbRowstride, osBuffer.pxbWidth, osBuffer.pxbHeight);
		clipBuffer.whiteBackground();
		part = clipparts;
		while (part) {
			part->fillBlack(clipBuffer, drawArea);
			part = part->next;
		}
		guint pixSize = osBuffer.pxbSize / 3;
		guchar *apixbuffer = (guchar *) g_malloc(pixSize * 4);
		guchar *apb = apixbuffer;
		guchar *gb = groupBuffer.pixbuffer;
		for (guint i = 0; i < pixSize; i++) {
			// copy the rgb bytes
			*apb++ = *gb++;
			*apb++ = *gb++;
			*apb++ = *gb++;
			// and the alpha value
			*apb++ = ((0xFF - clipBuffer.pixbuffer[i*3]) * (0xFF - groupBuffer.pixbuffer[i*3])) >> 8; 
		}
		double affine[6] = { 1, 0, 0, 1, 0, 0 };
		gint width = osBuffer.pxbWidth;
		gint height = osBuffer.pxbHeight;
		art_rgb_rgba_affine(osBuffer.pixbuffer, 0, 0, 
							width, height,
							osBuffer.pxbRowstride, (art_u8 *) apixbuffer,
							width, height, width * 4, affine, 
							ART_FILTER_NEAREST, NULL);
	}
}

void ClipGroup::fillBlack(OffscreenBuffer& osBuffer, RectangleArea& drawArea) {
	BoundingBox bbox = getBoundingBox();
	if (drawArea.intersect(bbox)) {
		OffscreenBuffer groupBuffer;
		groupBuffer.origin = osBuffer.origin;
		groupBuffer.scaleFactor = osBuffer.scaleFactor;
		groupBuffer.allocBuffer(osBuffer.pxbRowstride, osBuffer.pxbWidth, osBuffer.pxbHeight);
		groupBuffer.whiteBackground();
		GObject *part = parts;
		while (part) {
			part->fillBlack(groupBuffer, drawArea);
			part = part->next;
		}
		OffscreenBuffer clipBuffer;
		clipBuffer.origin = osBuffer.origin;
		clipBuffer.scaleFactor = osBuffer.scaleFactor;
		clipBuffer.allocBuffer(osBuffer.pxbRowstride, osBuffer.pxbWidth, osBuffer.pxbHeight);
		clipBuffer.whiteBackground();
		part = clipparts;
		while (part) {
			part->fillBlack(clipBuffer, drawArea);
			part = part->next;
		}
		guint pixSize = osBuffer.pxbSize / 3;
		guchar *apixbuffer = (guchar *) g_malloc(pixSize * 4);
		guchar *apb = apixbuffer;
		guchar *gb = groupBuffer.pixbuffer;
		for (guint i = 0; i < pixSize; i++) {
			// copy the rgb bytes
			*apb++ = *gb++;
			*apb++ = *gb++;
			*apb++ = *gb++;
			// and the alpha value
			*apb++ = ((0xFF - clipBuffer.pixbuffer[i*3]) * (0xFF - groupBuffer.pixbuffer[i*3])) >> 8; 
		}
		double affine[6] = { 1, 0, 0, 1, 0, 0 };
		gint width = osBuffer.pxbWidth;
		gint height = osBuffer.pxbHeight;
		art_rgb_rgba_affine(osBuffer.pixbuffer, 0, 0, 
							width, height,
							osBuffer.pxbRowstride, (art_u8 *) apixbuffer,
							width, height, osBuffer.pxbRowstride, affine, ART_FILTER_NEAREST, NULL);
	}
						
}

void ClipGroup::drawWireFrame(DocWindow *docwin, Gdk_GC& context, RectangleArea& drawArea) {
	GObject *part = parts;
	while (part) {
		part->drawWireFrame(docwin, context, drawArea);
		part = part->next;
	}
	part = clipparts;
	while (part) {
		part->drawWireFrame(docwin, context, drawArea);
		part = part->next;
	}
	
}

void ClipGroup::drawPoints(DocWindow *docwin, Gdk_GC& context, RectangleArea& drawArea) {
	ObjectGroup::drawPoints(docwin, context, drawArea);
	GObject *part = clipparts;
	while (part) {
		part->drawPoints(docwin, context, drawArea);
		part = part->next;
	}
}

BoundingBox ClipGroup::getBoundingBox() {
	GObject *part = parts;
	BoundingBox bbox;
	while (part) {
		BoundingBox partbb = part->getBoundingBox();
		bbox.combine(partbb);
		part = part->next;
	}
	part = clipparts;
	while (part) {
		BoundingBox partbb = part->getBoundingBox();
		bbox.combine(partbb);
		part = part->next;
	}
	return bbox;
}
gdouble ClipGroup::attractPoint(gdouble x, gdouble y, gdouble& ax, gdouble& ay, gdouble tolerance, int attractType) {
	gdouble nearest = ObjectGroup::attractPoint(x, y, ax, ay, tolerance, attractType);
	gdouble temp_ax = ax, temp_ay = ay;
	GObject *part = clipparts;
	while (part) {
		gdouble res;
		if ((res = part->attractPoint(x, y, temp_ax, temp_ay, tolerance, attractType)) < nearest) {
			nearest = res;
			ax = temp_ax;
			ay = temp_ay;
		}
		part = part->next;
	}
	return nearest;
}

SelectableItem *ClipGroup::checkPoints(gdouble x, gdouble y, gdouble tolerance, gboolean onlyCurvePoints) {
	SelectableItem *found = ObjectGroup::checkPoints(x, y, tolerance, onlyCurvePoints);
	GObject *part = clipparts;
	while (part) {
		SelectableItem *check = part->checkPoints(x, y, tolerance, onlyCurvePoints);
		if (check) {
			found = check;
		}
		part = part->next;
	}
	return found;
}

GObject *ClipGroup::checkArea(gdouble x, gdouble y) {
	GObject *part = clipparts;
	GObject *found = NULL;
	while (part) {
		if ((found = part->checkArea(x, y))) {
			break;
		}
		part = part->next;
	}
	if (found) {
		part = parts;
		while (part) {
			if (part->checkArea(x, y)) {
				return this;
			}
			part = part->next;
		}
	}
	return NULL;
}

SelectableItem *ClipGroup::checkCurves(gdouble x, gdouble y, gdouble tolerance) {
	SelectableItem *found = ObjectGroup::checkCurves(x, y, tolerance);
	GObject *part = clipparts;
	while (part) {
		SelectableItem *check = part->checkCurves(x, y, tolerance);
		if (check) {
			found = check;
		}
		part = part->next;
	}
	return found;
}

void ClipGroup::selectRectangle(gdouble left, gdouble top, gdouble right, gdouble bottom, gint what) {
	ObjectGroup::selectRectangle(left, top, right, bottom, what);
	GObject *part = clipparts;
	while (part) {
		part->selectRectangle(left, top, right, bottom, what);
		part = part->next;
	}
}

void ClipGroup::deSelectRectangle(gdouble left, gdouble top, gdouble right, gdouble bottom, gint what) {
	ObjectGroup::deSelectRectangle(left, top, right, bottom, what);
	GObject *part = clipparts;
	while (part) {
		part->deSelectRectangle(left, top, right, bottom, what);
		part = part->next;
	}
}

void ClipGroup::selectAll() {
	ObjectGroup::selectAll();
	GObject *part = clipparts;
	while (part) {
		part->selectAll();
		part = part->next;
	}
}

void ClipGroup::deSelectAll() {
	ObjectGroup::deSelectAll();
	GObject *part = clipparts;
	while (part) {
		part->deSelectAll();
		part = part->next;
	}
}

GObject *ClipGroup::findObject(GObject *object) {
	GObject *part = parts;
	while (part) {
		GObject *found;
		if ((found = part->findObject(object))) {
			return found;
		}
		part = part->next;
	}
	part = clipparts;
	while (part) {
		GObject *found;
		if ((found = part->findObject(object))) {
			return found;
		}
		part = part->next;
	}
	return NULL;
}

void ClipGroup::reverseOrder() {
	ObjectGroup::reverseOrder();
	GObject *part = clipparts;
	while (part) {
		part->reverseOrder();
		part = part->next;
	}
}

GObject *ClipGroup::removeObject(GObject *object) {
	if (object == this) {
		if (next) {
			next->prev = prev;
		}
		if (prev) {
			prev->next = next;
		}
		return this;
	}
	// try subparts
	GObject *part = parts;
	while (part) {
		if (part == object) {
			if (parts == object) {
				parts = parts->next;
			} else {
				parts->remove(object);
			}
			return part;
		} else {
			GObject *found;
			if ((found = part->removeObject(object))) {
				return found;
			}
		}
		part = part->next;
	}
	part = clipparts;
	while (part) {
		if (part == object) {
			if (clipparts == object) {
				clipparts = clipparts->next;
			} else {
				clipparts->remove(object);
			}
			return part;
		} else {
			GObject *found;
			if ((found = part->removeObject(object))) {
				return found;
			}
		}
		part = part->next;
	}
	return NULL;
}

gint ClipGroup::getSelection(SelectionItemList& selectionList, gint mode) {
	if (selected()) {
		selectionList.insert(selectionList.end(), new SelectionItem(this, mode));
		return 1;
	} else {
		GObject *part = parts;
		gint selectedParts = 0;
		while (part) {
			selectedParts += part->getSelection(selectionList, mode);
			part = part->next;
		}
		part = clipparts;
		while (part) {
			selectedParts += part->getSelection(selectionList, mode);
			part = part->next;
		}
		return selectedParts;
	}
}

gint ClipGroup::writeXML(ostream& xmlout) {
	xmlout << "<g clip-rule=\"evenodd\">\n";
	xmlout << "<clipPath id=\"ClipPath" << ClipGroupNumber << "\">\n";
	GObject *part = clipparts;
	while (part) {
		part->writeXML(xmlout);
		part = part->next;
	}
	xmlout << "</clipPath>\n";
	xmlout << "<g clip-path=\"ClipPath" << ClipGroupNumber << "\">\n";
	part = parts;
	while (part) {
		part->writeXML(xmlout);
		part = part->next;
	}
	xmlout << "</g>\n</g>\n";
}

void ClipGroup::writePSPathDescription(ostream& psout) {
	GObject *part = parts;
	while (part) {
		part->writePSPathDescription(psout);
		part = part->next;
	}
}

gint ClipGroup::writePS(ostream& psout) {
	psout << "gs\n"; // gsave
	psout << "np\n"; // newpath
	GObject *part = clipparts;
	while (part) {
		part->writePSPathDescription(psout);
		part = part->next;
	}
	psout << "clip\n";
	part = parts;
	while (part) {
		part->writePS(psout);
		part = part->next;
	}
	psout << "gr\n"; // grestore
	return 0;
}

// SelectableItem methods
void ClipGroup::drawSelection(DocWindow *docwin, Gdk_GC& context, RectangleArea& drawArea) {
	GObject *part = parts;
	while (part) {
		part->drawSelection(docwin, context, drawArea);
		part = part->next;
	}
	part = clipparts;
	while (part) {
		part->drawSelection(docwin, context, drawArea);
		part = part->next;
	}
}

void ClipGroup::transformByMatrix(TransformationMatrix& m, gboolean incFillPattern) {
	GObject *part = parts;
	while (part) {
		part->transformByMatrix(m, incFillPattern);
		part = part->next;
	}
	part = clipparts;
	while (part) {
		part->transformByMatrix(m, incFillPattern);
		part = part->next;
	}
}

SelectableItem *ClipGroup::clone() {
	ClipGroup *newGroup = new ClipGroup(doc);
	GObject *part = parts;
	while (part) {
		newGroup->addPart((GObject *) part->clone());
		part = part->next;
	}
	part = clipparts;
	while (part) {
		newGroup->addClipPart((GObject *) part->clone());
		part = part->next;
	}
	return newGroup;
}

void ClipGroup::setCoords(SelectableItem* from) {
	//g_printerr("ClipGroup::setCoords\n");
	GObject *part = parts;
	GObject *frompart = ((ClipGroup *)from)->parts;
	while (part) {
		part->setCoords(frompart);
		part = part->next;
		frompart = frompart->next;
	}
	part = clipparts;
	frompart = ((ClipGroup *)from)->clipparts;
	while (part) {
		part->setCoords(frompart);
		part = part->next;
		frompart = frompart->next;
	}
}

void ClipGroup::setDocument(GDDocument *document) {
	GObject *part = parts;
	while (part) {
		part->setDocument(document);
		part = part->next;
	}
	part = clipparts;
	while (part) {
		part->setDocument(document);
		part = part->next;
	}
}
	
gboolean ClipGroup::selected() {
	GObject *part = clipparts;
	while (part) {
		if (! part->selected()) {
			return false;
		}
		part = part->next;
	}
	return ObjectGroup::selected();
}

void ClipGroup::clear() {
	deSelectAll();
	doc->removeObject(this);
	delete this;
}

gchar *ClipGroup::identify() {
	return "ClipGroup";
}


ClipGroup::~ClipGroup() {
	GObject *part = clipparts;
	while (part) {
		//g_printerr("ClipGroup::~ClipGroup remove clippart %lX\n", part);
		GObject *nextObject = part->next;
		delete part;
		part = nextObject;
	}
}
