/* $Id: CompoundPath.cc,v 1.26 2004/07/21 06:31:20 pklaus Exp $
 *
 * CompoundPath.cc
 *
 * Gestalter - The free vector drawing application
 * Copyright (c) 2002 Philipp Klaus <gestalter@linotux.ch>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include "CompoundPath.hh"
#include "GDDocument.hh"
#include "BezierPath.hh"
#include "DocWindow.hh"
#include "GDStyle.hh"
#include "BoundingBox.hh"
#include "SelectionList.hh"
#include <libart_lgpl/art_misc.h>
#include <libart_lgpl/art_svp.h>
#include <libart_lgpl/art_svp_vpath.h>
#include <libart_lgpl/art_rgb_svp.h>
#include <libart_lgpl/art_vpath_bpath.h>
#include <libart_lgpl/art_vpath_dash.h>
#include <libart_lgpl/art_svp_vpath_stroke.h>
#include <libart_lgpl/art_svp_point.h>

CompoundPath::CompoundPath(GDDocument *document) : GObject(document) {
	parts = NULL;
	drawStyle = new GDStyle();
	fillStyle = new GDStyle();
}


void CompoundPath::addPart(BezierPath *newObject) {
	//g_printerr("CompoundPath::addPart object: %lX\n", newObject);
	newObject->parent = this;
	//newObject->deSelectAll();
	if (parts) {
		parts->insert(NULL, newObject);
	} else {
		parts = newObject;
		*drawStyle = *(parts->getStrokeStyle());
		*fillStyle = *(parts->getFillStyle());
	}
}

void CompoundPath::removePart(BezierPath *object) {
	object->parent = NULL;
	parts->remove(object);
	if (parts == object) {
		parts = (BezierPath *) parts->next;
	}
}


// DrawObject methods
void CompoundPath::drawWithSelection(OffscreenBuffer& osBuffer, RectangleArea& drawArea) {
	//g_printerr("CompoundPath::drawWithSelection\n");
	GObject *part = parts;
	while (part) {
		part->drawWithSelection(osBuffer, drawArea);
		part = part->next;
	}
}

void CompoundPath::drawWireFrameSelection(DocWindow *docwin, RectangleArea& drawArea) {
	GObject *part = parts;
	while (part) {
		part->drawWireFrameSelection(docwin, drawArea);
		part = part->next;
	}
}

GDStyle *CompoundPath::getStrokeStyle() {
	return drawStyle;
}

GDStyle *CompoundPath::getFillStyle() {
	return fillStyle;
}

void CompoundPath::gatherResource(GDResourceList& collection, gint type) {
	if (drawStyle) {
		drawStyle->gatherResource(collection, type);
	}
	if (fillStyle) {
		fillStyle->gatherResource(collection, type);
	}
	GObject *part = parts;
	while (part) {
		part->gatherResource(collection, type);
		part = part->next;
	}
}


ArtVpath *CompoundPath::getVpath(gdouble sF, gdouble ox, gdouble oy, gboolean filled, RectangleArea& drawArea) {
	//g_printerr("CompoundPath::getVpath\n");
	BezierPath *part = parts;
	guint size = 0;
	while (part) {
		//g_printerr("CompoundPath::getVpath get size of part %lX\n", part);
		size += part->countNodes()+1;
		part = (BezierPath *) part->next;
	}
	ArtBpath *artpath = art_new(ArtBpath, size + 1);
	guint len = 0;
	part = parts;
	while (part) {
		//g_printerr("CompoundPath::getVpath get Bpath of part %lX\n", part);
		len += part->fillBpath(&artpath[len], size - len, sF, ox, oy, filled, drawArea);
		part = (BezierPath *) part->next;
	}
	artpath[len].code = ART_END;
	artpath[len].x1 = 0;
	artpath[len].y1 = 0;
	artpath[len].x2 = 0;
	artpath[len].y2 = 0;
	artpath[len].x3 = 0;
	artpath[len].y3 = 0;
	ArtVpath *ap = art_bez_path_to_vec(artpath, 0.5); // 0.5 is the flatness
	art_free(artpath);
	//g_printerr("CompoundPath::getVpath finished\n");
	return ap;
}


// GObject methods
void CompoundPath::draw(OffscreenBuffer& osBuffer, RectangleArea& drawArea) {
	BoundingBox bbox = getBoundingBox();
	if (drawArea.intersect(bbox)) {
		if (fillStyle && fillStyle->drawType != GDSTYLE_DRAWTYPE_TRANSPARENT) {
			ArtVpath *avp = getVpath(osBuffer.scaleFactor, osBuffer.origin.x,
										osBuffer.origin.y, true, drawArea);
			ArtSVP *asvp = art_svp_from_vpath(avp);
			art_free(avp);
			fillStyle->fillSVP(asvp, osBuffer.pixbuffer, osBuffer.pxbRowstride,
							osBuffer.pxbWidth, osBuffer.pxbHeight,
							osBuffer.origin.x, osBuffer.origin.y, osBuffer.scaleFactor	);
			art_svp_free(asvp);
		}
		if (drawStyle && drawStyle->drawType != GDSTYLE_DRAWTYPE_TRANSPARENT) {
			ArtVpath *avp = getVpath(osBuffer.scaleFactor, osBuffer.origin.x,
										osBuffer.origin.y, false, drawArea);
			drawStyle->strokeAVP(avp, osBuffer.pixbuffer, osBuffer.pxbRowstride,
								osBuffer.pxbWidth, osBuffer.pxbHeight,
								osBuffer.origin.x, osBuffer.origin.y,
								osBuffer.scaleFactor);
		}
	}
}

void CompoundPath::drawBlack(OffscreenBuffer& osBuffer, RectangleArea& drawArea) {
	BoundingBox bbox = getBoundingBox();
	if (drawArea.intersect(bbox)) {
		guint32 color = 0x000000FF;
		if (fillStyle && fillStyle->drawType != GDSTYLE_DRAWTYPE_TRANSPARENT) {
			ArtVpath *avp = getVpath(osBuffer.scaleFactor, osBuffer.origin.x,
										osBuffer.origin.y, true, drawArea);
			ArtSVP *asvp = art_svp_from_vpath(avp);
			art_free(avp);
			art_rgb_svp_alpha(asvp, 0, 0,
						osBuffer.pxbWidth,
						osBuffer.pxbHeight,
						color,
						osBuffer.pixbuffer,
						osBuffer.pxbRowstride,
						NULL);
			art_svp_free(asvp);
		}
		if (drawStyle && drawStyle->drawType != GDSTYLE_DRAWTYPE_TRANSPARENT) {
			ArtVpath *avp = getVpath(osBuffer.scaleFactor, osBuffer.origin.x,
										osBuffer.origin.y, false, drawArea);
			if (drawStyle->lineType == GDSTYLE_LINETYPE_DASHED) {
				ArtVpathDash dash;
				dash.dash = drawStyle->dashes;
				for (int i = 0; i < 6; i++) {
					if (dash.dash[i] < 0) {
						dash.n_dash = i;
						i = 6;
					}
				}
				dash.offset = 0;
				if (dash.n_dash) {
					avp = art_vpath_dash(avp, &dash);
				}
			}
			ArtPathStrokeJoinType joins;
			ArtPathStrokeCapType caps;
			if (drawStyle->lineCaps == GDSTYLE_CAPS_ROUND) {
				caps = ART_PATH_STROKE_CAP_ROUND;
			} else if (drawStyle->lineCaps == GDSTYLE_CAPS_SQUARE) {
				caps = ART_PATH_STROKE_CAP_SQUARE;
			} else {
				caps = ART_PATH_STROKE_CAP_BUTT;
			}
			if (drawStyle->lineJoins == GDSTYLE_JOINS_BEVEL) {
				joins = ART_PATH_STROKE_JOIN_BEVEL;
			} else if (drawStyle->lineJoins == GDSTYLE_JOINS_ROUND) {
				joins = ART_PATH_STROKE_JOIN_ROUND;
			} else {
				joins = ART_PATH_STROKE_JOIN_MITER;
			}
			gdouble swidth = drawStyle->strokeWidth * osBuffer.scaleFactor;
			if (swidth < 0.1) swidth = 0.1;
			ArtSVP *asvp = art_svp_vpath_stroke(avp, 
									joins,
									caps,
									swidth, 
									drawStyle->lineMiterLimit, 0.5);
			art_free(avp);
			art_rgb_svp_alpha(asvp, 0, 0,
						osBuffer.pxbWidth,
						osBuffer.pxbHeight,
						color,
						osBuffer.pixbuffer,
						osBuffer.pxbRowstride,
						NULL);
			art_svp_free(asvp);
		}
	}
}

void CompoundPath::fillBlack(OffscreenBuffer& osBuffer, RectangleArea& drawArea) {
	BoundingBox bbox = getBoundingBox();
	if (drawArea.intersect(bbox)) {
		ArtVpath *avp = getVpath(osBuffer.scaleFactor, osBuffer.origin.x,
									osBuffer.origin.y, true, drawArea);
		ArtSVP *asvp = art_svp_from_vpath(avp);
		art_free(avp);
		guint32 color = 0x000000FF;
		art_rgb_svp_alpha(asvp, 0, 0,
					osBuffer.pxbWidth,
					osBuffer.pxbHeight,
					color,
					osBuffer.pixbuffer,
					osBuffer.pxbRowstride,
					NULL);
		art_svp_free(asvp);
	}
}

void CompoundPath::drawWireFrame(DocWindow *docwin, Gdk_GC& context, RectangleArea& drawArea) {
	GObject *part = parts;
	while (part) {
		part->drawWireFrame(docwin, context, drawArea);
		part = part->next;
	}
}

void CompoundPath::drawPoints(DocWindow *docwin, Gdk_GC& context, RectangleArea& drawArea) {
	GObject *part = parts;
	while (part) {
		part->drawPoints(docwin, context, drawArea);
		part = part->next;
	}
}

BezierPathVector *CompoundPath::getBezierPaths() {
	BezierPathVector *result = NULL;
	GObject *part = parts;
	while (part) {
		part = part->next;
		BezierPathVector *paths = part->getBezierPaths();
		if (paths) {
			if (result) {
				for (BezierPathVector::iterator it = paths->begin(); it != paths->end(); it++) {
					result->push_back(*it);
				}
				delete paths;
			} else {
				result = paths;
			}
		}
	}
	return result;
}


BoundingBox CompoundPath::getBoundingBox() {
	GObject *part = parts;
	BoundingBox bbox;
	while (part) {
		BoundingBox partbb = part->getBoundingBox();
		bbox.combine(partbb);
		part = part->next;
	}
	return bbox;
}

gdouble CompoundPath::attractPoint(gdouble x, gdouble y, gdouble& ax, gdouble& ay, gdouble tolerance, int attractType) {
	gdouble nearest = G_MAXDOUBLE;
	gdouble temp_ax = ax, temp_ay = ay;
	GObject *part = parts;
	while (part) {
		gdouble res;
		if ((res = part->attractPoint(x, y, temp_ax, temp_ay,tolerance, attractType)) < nearest) {
			nearest = res;
			ax = temp_ax;
			ay = temp_ay;
		}
		part = part->next;
	}
	return nearest;
}

SelectableItem *CompoundPath::checkPoints(gdouble x, gdouble y, gdouble tolerance, gboolean onlyCurvePoints) {
	if (drawStyle && drawStyle->isGuide && doc->guidesLocked) return NULL;
	SelectableItem *found = NULL;
	GObject *part = parts;
	while (part) {
		SelectableItem *check = part->checkPoints(x, y, tolerance, onlyCurvePoints);
		if (check) {
			found = check;
		}
		part = part->next;
	}
	return found;
}

GObject *CompoundPath::checkArea(gdouble x, gdouble y) {
	if (fillStyle && (fillStyle->drawType == GDSTYLE_DRAWTYPE_TRANSPARENT || fillStyle->isGuide)) {
		// don't select by area if not filled
		return NULL;
	}
	RectangleArea dummy;
	ArtVpath *vpath = getVpath(1, 0, 0, true, dummy);
	ArtSVP *asvp = art_svp_from_vpath(vpath);
	art_free(vpath);
	int winding = art_svp_point_wind(asvp, x, y);
	art_svp_free(asvp);
	if (winding) {
		return this;
	}
	return NULL;
}

SelectableItem *CompoundPath::checkCurves(gdouble x, gdouble y, gdouble tolerance) {
	if (drawStyle && drawStyle->isGuide && doc->guidesLocked) return NULL;
	SelectableItem *found = NULL;
	GObject *part = parts;
	while (part) {
		SelectableItem *check = part->checkCurves(x, y, tolerance);
		if (check) {
			found = check;
		}
		part = part->next;
	}
	return found;
}

void CompoundPath::selectRectangle(gdouble left, gdouble top, gdouble right, gdouble bottom, gint what) {
	if (drawStyle && drawStyle->isGuide && doc->guidesLocked) return;
	GObject *part = parts;
	while (part) {
		part->selectRectangle(left, top, right, bottom, what);
		part = part->next;
	}
}

void CompoundPath::deSelectRectangle(gdouble left, gdouble top, gdouble right, gdouble bottom, gint what) {
	GObject *part = parts;
	while (part) {
		part->deSelectRectangle(left, top, right, bottom, what);
		part = part->next;
	}
}

void CompoundPath::selectAll() {
	if (drawStyle && drawStyle->isGuide && doc->guidesLocked) return;
	BezierPath *part = parts;
	while (part) {
		part->doSelectAll();
		part = (BezierPath *) part->next;
	}
	doc->selection->insertObject(this);
}

void CompoundPath::deSelectAll() {
	BezierPath *part = parts;
	while (part) {
		part->doDeSelectAll();
		part = (BezierPath *) part->next;
	}
	doc->selection->removeObject(this);
}

GObject *CompoundPath::findObject(GObject *object) {
	GObject *part = parts;
	while (part) {
		GObject *found;
		if ((found = part->findObject(object))) {
			return found;
		}
		part = part->next;
	}
	return NULL;
}

void CompoundPath::reverseOrder() {
	GObject *part = parts;
	while (part) {
		part->reverseOrder();
		part = part->next;
	}
}

GObject *CompoundPath::removeObject(GObject *object) {
	if (object == this) {
		if (next) {
			next->prev = prev;
		}
		if (prev) {
			prev->next = next;
		}
		return this;
	}
	// try subparts
	GObject *part = parts;
	while (part) {
		if (part == object) {
			if (parts == object) {
				parts = (BezierPath *)parts->next;
			} else {
				parts->remove(object);
			}
			return part;
		} else {
			GObject *found;
			if ((found = part->removeObject(object))) {
				return found;
			}
		}
		part = part->next;
	}
	return NULL;
}

gint CompoundPath::getSelection(SelectionItemList& selectionList, gint mode) {
	if (selected()) {
		//g_printerr("CompoundPath::getSelection cpath %lX fully selected\n", this);
		selectionList.insert(selectionList.end(), new SelectionItem(this, mode));
		return 1;
	} else {
		GObject *part = parts;
		gint selectedParts = 0;
		while (part) {
			//g_printerr("CompoundPath::getSelection cpath %lX select path %lX\n", this, part);
			selectedParts += part->getSelection(selectionList, mode);
			part = part->next;
		}
		return selectedParts;
	}
}

gint CompoundPath::writeXML(ostream& xmlout) {
	fillStyle->writeFillPreparationXML(xmlout, doc);
	xmlout << "<path d=\"";
	xmlout.setf(ios::fixed, ios::floatfield);
	xmlout.precision(12);
	BezierPath *part = parts;
	while (part) {
		part->writeXMLPathDescription(xmlout);
		part = (BezierPath *)part->next;
		if (part) {
			xmlout << " ";
		}
	}
	xmlout << "\"";
	fillStyle->writeFillXML(xmlout, doc);
	drawStyle->writeStrokeXML(xmlout, doc);
	xmlout << "/>\n";
	return 0;
}

void CompoundPath::writePSPathDescription(ostream& psout) {
	GObject *part = parts;
	while (part) {
		part->writePSPathDescription(psout);
		part = part->next;
	}
}

gint CompoundPath::writePS(ostream& psout) {
	gdouble pfx = doc->printFactor.x;
	gdouble pfy = doc->printFactor.y;
	if (! parts) return 0;
	psout << "np\n";
	writePSPathDescription(psout);
	fillStyle->writeFillPS(psout, doc);
	drawStyle->writeStrokePS(psout, doc);
	
	return 0;
}


// SelectableItem methods
void CompoundPath::drawSelection(DocWindow *docwin, Gdk_GC& context, RectangleArea& drawArea) {
	//g_printerr("CompoundPath::drawSelection this=%lX\n", this);
	GObject *part = parts;
	while (part) {
		//g_printerr("CompoundPath::drawSelection part=%lX\n", part);
		part->drawSelection(docwin, context, drawArea);
		part = part->next;
	}
}

void CompoundPath::transformByMatrix(TransformationMatrix& m, gboolean incFillPattern) {
	//g_printerr("CompoundPath::transformByMatrix\n");
	//m.printMatrix();
	GObject *part = parts;
	while (part) {
		part->transformByMatrix(m, incFillPattern);
		part = part->next;
	}
	// transform styles
	if (drawStyle && incFillPattern) drawStyle->transformByMatrix(m);
	if (fillStyle && incFillPattern) fillStyle->transformByMatrix(m);
}

void CompoundPath::setFillMatrix(TransformationMatrix& m) {
	GObject *part = parts;
	while (part) {
		part->setFillMatrix(m);
		part = part->next;
	}
	if (fillStyle) fillStyle->tMatrix = m;
}

SelectableItem *CompoundPath::clone() {
	//g_printerr("CompoundPath::clone\n");
	CompoundPath *newCPath = new CompoundPath(doc);
	BezierPath *part = parts;
	while (part) {
		BezierPath *clonePart = (BezierPath *) part->clone();
		//doc->selection->removeObject(part);
		newCPath->addPart(clonePart);
		part = (BezierPath *) part->next;
	}
	//newCPath->selectAll();
	return newCPath;
}

void CompoundPath::setCoords(SelectableItem* from) {
	GObject *part = parts;
	GObject *frompart = ((CompoundPath *)from)->parts;
	while (part) {
		part->setCoords(frompart);
		part = part->next;
		frompart = frompart->next;
	}
	// set the changed matrix of the styles too
	CompoundPath *f = (CompoundPath *) from;
	if (drawStyle && f->drawStyle) {
		drawStyle->tMatrix = f->drawStyle->tMatrix;
	}
	if (fillStyle && f->fillStyle) {
		fillStyle->tMatrix = f->fillStyle->tMatrix;
	}
}

void CompoundPath::setDocument(GDDocument *document) {
	if (doc != document) {
		// this really is a change of the document
		if (fillStyle) {
			fillStyle->changeDocument(document);
		}
	}
	GObject::setDocument(document);
	GObject *part = parts;
	while (part) {
		part->setDocument(document);
		part = part->next;
	}
}

gboolean CompoundPath::selected() {
	gboolean isSelected = true;
	GObject *part = parts;
	while (part) {
		if (! part->selected()) {
			isSelected = false;
		}
		part = part->next;
	}
	return isSelected;
}

void CompoundPath::clear() {
	deSelectAll();
	doc->removeObject(this);
	delete this;
}

gchar *CompoundPath::identify() {
	return "CompoundPath";
}


CompoundPath::~CompoundPath() {
	// deSelectAll(); // not used anymore (after code cleanup)
	GObject *part = parts;
	while (part) {
		GObject *nextObject = part->next;
		delete part;
		part = nextObject;
	}
	// delete the styles
	delete drawStyle;
	delete fillStyle;
}
