/* $Id: GDCharacter.cc,v 1.15 2005/05/07 17:39:32 pklaus Exp $
 *
 * GDCharacter.cc
 *
 * Gestalter - The free vector drawing application
 * Copyright (c) 2002 Philipp Klaus <gestalter@linotux.ch>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include "GDCharacter.hh"
#include "TextStyle.hh"
#include "GObject.hh"
#include "BezierPath.hh"
#include "BezierNode.hh"
#include "CompoundPath.hh"
#include "GDFont.hh"

extern FT_Library freetype_library;

GDCharacter::GDCharacter() {
	//g_printerr("GDCharacter::GDCharacter\n");
	unicode = 0;
	startX = -1; // initial nonsense value
	strokeStyle.drawType = GDSTYLE_DRAWTYPE_TRANSPARENT;
}

GDCharacter::GDCharacter(const GDCharacter& from) {
	//g_printerr("GDCharacter::GDCharacter(const &) %c\n", from.unicode);
	unicode = from.unicode;
	style = from.style;
	strokeStyle = from.strokeStyle;
	fillStyle = from.fillStyle;
	//g_printerr("GDCharacter::GDCharacter(const &) finished\n");
}

GDCharacter::GDCharacter(guint charcode) {
	//g_printerr("GDCharacter::GDCharacter(guint)\n");
	unicode = charcode;
	strokeStyle.drawType = GDSTYLE_DRAWTYPE_TRANSPARENT;
}

string GDCharacter::convert2UTF8() {
	string utf;
	if ((unicode & 0x0000007F) == unicode) {
		utf.append(1, (char) (unicode & 0x007F));
	} else if ((unicode & 0x000007FF) == unicode) {
		utf.append(1, (char) ((unicode >> 6 & 0x001F) | 0xC0));
		utf.append(1, (char) ((unicode & 0x003F) | 0x80));
	} else if ((unicode & 0x00FFFF) == unicode) {
		utf.append(1, (char) (((unicode & 0x00F000) >> 12) | 0xE0));
		utf.append(1, (char) (((unicode & 0x000FC0) >> 6) | 0x80));
		utf.append(1, (char) ((unicode & 0x003F) | 0x80));
	} else {
		utf.append(1, (char) (((unicode & 0x001C0000) >> 18) | 0xF0));
		utf.append(1, (char) (((unicode & 0x0003F000) >> 12) | 0x80));
		utf.append(1, (char) (((unicode & 0x000FC0) >> 6) | 0x80));
		utf.append(1, (char) ((unicode & 0x003F) | 0x80));
	}
	return utf;
}

void GDCharacter::setUnicode(guint code) {
	unicode = code;
}

guint GDCharacter::getUnicode() {
	return unicode;
}

void GDCharacter::setStrokeStyle(GDStyle &style) {
	strokeStyle = style;
}

GDStyle *GDCharacter::getStrokeStyle() {
	return &strokeStyle;
}

void GDCharacter::setFillStyle(GDStyle &style) {
	fillStyle = style;
}

GDStyle *GDCharacter::getFillStyle() {
	return &fillStyle;
}

void GDCharacter::setTextStyle(TextStyle &style) {
	//g_printerr("GDCharacter::setTextStyle char=%c just: %d -> %d\n", unicode, this->style.getJustification(), style.getJustification());
	this->style = style;
}

BezierPath *GDCharacter::convertContour(FT_Vector *vectors, char *tags, gint start, gint end, gboolean reverse, GDDocument *doc) {
	BezierPath *path = new BezierPath(doc);
	int phase = 0;
	BezierNode *currnode = NULL;
	BezierNode *prevnode = NULL;
	for (int i = start; i <= end; i++) {
		gdouble vx = ((gdouble)vectors[i].x) / 64.0;
		gdouble vy = ((gdouble)-vectors[i].y) / 64.0;
		if (tags[i] == FT_Curve_Tag_On) {
			if (phase < 1) {
				// previous vector was on curve or first cubic point
				if (currnode) {
					path->addNode(currnode);
					prevnode = currnode;
				}
				currnode = new BezierNode(path);
				currnode->setOnCurvePoint(vx, vy);
				currnode->setPreviousHandle(currnode->onCurve.x, currnode->onCurve.y);
				currnode->setNextHandle(currnode->onCurve.x, currnode->onCurve.y);
			} else if (phase == 1) {
				// previous vector was second cubic off curve point
				currnode->setOnCurvePoint(vx, vy);
				currnode->setNextHandle(currnode->onCurve.x, currnode->onCurve.y);
			} else if (phase == 2) {
				currnode->setOnCurvePoint(vx, vy);
				currnode->setPreviousHandle((2 * currnode->prevHandle.x + currnode->onCurve.x) / 3,
											(2 * currnode->prevHandle.y + currnode->onCurve.y) / 3);
				currnode->setNextHandle(currnode->onCurve.x, currnode->onCurve.y);
			}
			phase = 0;
		} else if (tags[i] == FT_Curve_Tag_Cubic) {
			if (! currnode) {
				g_printerr("Found FT_Curve_Tag_Cubic at the beginning of the contour (%d)\n", i);
				delete path;
				return NULL;
			}
			if (phase == 0) {
				// previous vector was on curve
				currnode->setNextHandle(vx, vy);
				phase = -1;
			} else if (phase == -1) {
				// previous vector was first cubic off curve point
				if (currnode) {
					path->addNode(currnode);
					prevnode = currnode;
				}
				currnode = new BezierNode(path);
				currnode->setOnCurvePoint(vx, vy);
				currnode->setPreviousHandle(currnode->onCurve.x, currnode->onCurve.y);
				currnode->setNextHandle(currnode->onCurve.x, currnode->onCurve.y);
				phase = 1;
			}
		} else {
			// tags[i] must be FT_Curve_Tag_Conic
			if (! currnode) {
				g_printerr("Found FT_Curve_Tag_Conic at the beginning of the contour (%d)\n", i);
				delete path;
				return NULL;
			}
			if (phase == 0) {
				// previous vector was on curve
				currnode->setNextHandle((2 * vx + currnode->onCurve.x) / 3,
										(2 * vy + currnode->onCurve.y) / 3);
				path->addNode(currnode);
				prevnode = currnode;
				currnode = new BezierNode(path);
				currnode->setPreviousHandle(vx, vy);
			} else if (phase == 2) {
				// previous vector was conic off curve too
				// calculate middle point (must be on curve)
				currnode->setOnCurvePoint((vx + currnode->prevHandle.x) / 2,
								(vy + currnode->prevHandle.y) / 2);
				currnode->setPreviousHandle((2 * currnode->prevHandle.x + currnode->onCurve.x) / 3,
											(2 * currnode->prevHandle.y + currnode->onCurve.y) / 3);
				currnode->setNextHandle((2 * vx + currnode->onCurve.x) / 3,
										(2 * vy + currnode->onCurve.y) / 3);
				path->addNode(currnode);
				prevnode = currnode;
				currnode = new BezierNode(path);
				currnode->setPreviousHandle(vx, vy);
			} else {
				g_printerr("Found FT_Curve_Tag_Conic after FT_Curve_Tag_Cubic (%d)\n", i);
				delete path;
				delete currnode;
				return NULL;
			}
			phase = 2;
				
		}
	}
	// close the path
	if (phase < 1) {
		// last vector was on curve or first cubic off curve point
		path->addNode(currnode);
		path->nodes->prev = currnode;
		currnode->next = path->nodes;
	} else if (phase == 1) {
		// last vector was second cubic off curve point
		path->nodes->setPreviousHandle(currnode->prevHandle.x, currnode->prevHandle.y);
		if (prevnode) {
			prevnode->next = path->nodes;
		}
		path->nodes->prev = prevnode;
		delete currnode;
	} else if (phase == 2) {
		// last vector was conic off curve point
		path->nodes->setPreviousHandle((2 * currnode->prevHandle.x + path->nodes->onCurve.x) / 3,
										(2 * currnode->prevHandle.y + path->nodes->onCurve.y) / 3);
		if (prevnode) {
			prevnode->next = path->nodes;
		}
		path->nodes->prev = prevnode;
		delete currnode;
	}
	if (reverse) {
		path->reverseOrder();
	}
	return path;
}

GObject *GDCharacter::convert2Outline(GDDocument *doc) {
	FT_Face face = getTextStyle()->getFont()->getFace();
	int gindex = FT_Get_Char_Index(face, getUnicode());
	int error = FT_Load_Glyph(face, gindex, FT_LOAD_NO_HINTING | 
							FT_LOAD_NO_BITMAP | FT_LOAD_IGNORE_TRANSFORM);
	if (error) {
		g_printerr("GDCharacter::convert2Outline: ERROR cannot load glyph %d (error: %d)\n", gindex, error);
		return NULL;
	}
	FT_Glyph glyph;
	error = FT_Get_Glyph(face->glyph, &glyph);
	if (error) {
		g_printerr("GDCharacter::convert2Outline: ERROR cannot get glyph %d (error: %d)\n", gindex, error);
		return NULL;
	}
	if (glyph->format != ft_glyph_format_outline) {
		g_printerr("GDCharacter::convert2Outline: ERROR glyph %d returned in wrong format: %d\n", gindex, glyph->format);
		return NULL;
	}
	FT_OutlineGlyph olGlyph = (FT_OutlineGlyph) glyph;
	if (olGlyph->outline.n_contours > 1) {
		// build CompoundPath
		CompoundPath *compound = new CompoundPath(doc);
		for (int i = 0; i < olGlyph->outline.n_contours; i++) {
			BezierPath *contour = convertContour(olGlyph->outline.points, olGlyph->outline.tags, 
										i ? olGlyph->outline.contours[i-1]+1 : 0, olGlyph->outline.contours[i],
										(olGlyph->outline.flags & ft_outline_reverse_fill == 0), doc);
			compound->addPart(contour);
		}
		compound->setStrokeStyle(strokeStyle);
		compound->setFillStyle(fillStyle);
		FT_Done_Glyph(glyph);
		return compound;
	} else if (olGlyph->outline.n_contours == 1) {
		BezierPath *path = convertContour(olGlyph->outline.points, olGlyph->outline.tags, 0, 
										olGlyph->outline.contours[0],
										(olGlyph->outline.flags & ft_outline_reverse_fill == 0), doc);
		path->setStrokeStyle(strokeStyle);
		path->setFillStyle(fillStyle);
		FT_Done_Glyph(glyph);
		return path;
	} else return NULL;
}

TextStyle *GDCharacter::getTextStyle() {
	return &style;
}

void GDCharacter::transformByMatrix(TransformationMatrix& m, gboolean incFillPattern) {
	if (incFillPattern) {
		strokeStyle.transformByMatrix(m);
		fillStyle.transformByMatrix(m);
	}
}

void GDCharacter::setFillMatrix(TransformationMatrix& m) {
	fillStyle.tMatrix = m;
}

void GDCharacter::changeDocument(GDDocument *document) {
	fillStyle.changeDocument(document);
}

GDCharacter::~GDCharacter() {
	//g_printerr("GDCharacter::~GDCharacter\n");
}
