/* $Id: ObjectGroup.cc,v 1.17 2004/07/21 06:31:20 pklaus Exp $
 *
 * ObjectGroup.cc
 *
 * Gestalter - The free vector drawing application
 * Copyright (c) 2002 Philipp Klaus <gestalter@linotux.ch>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include "ObjectGroup.hh"
#include "GDDocument.hh"
#include "SelectionItemList.hh"
#include "SelectableItem.hh"
#include "DocWindow.hh"
#include "BoundingBox.hh"
#include "GDStyle.hh"

ObjectGroup::ObjectGroup(GDDocument *document) : GObject(document) {
	parts = NULL;
	drawStyle = NULL;
	fillStyle = NULL;
}


void ObjectGroup::addPart(GObject *newObject) {
	//g_printerr("ObjectGroup::addPart object=%lX\n", newObject);
	newObject->parent = this;
	if (parts) {
		parts->insert(NULL, newObject);
	} else {
		parts = newObject;
	}
}

void ObjectGroup::removePart(GObject *object) {
	object->parent = NULL;
	parts->remove(object);
	if (parts == object) {
		parts = parts->next;
	}
}


// DrawObject methods
void ObjectGroup::drawWithSelection(OffscreenBuffer& osBuffer, RectangleArea& drawArea) {
	GObject *part = parts;
	while (part) {
		part->drawWithSelection(osBuffer, drawArea);
		part = part->next;
	}
}

void ObjectGroup::drawWireFrameSelection(DocWindow *docwin, RectangleArea& drawArea) {
	GObject *part = parts;
	while (part) {
		part->drawWireFrameSelection(docwin, drawArea);
		part = part->next;
	}
}

GDStyle *ObjectGroup::getStrokeStyle() {
	if (drawStyle) delete drawStyle;
	drawStyle = NULL;
	GObject *part = parts;
	while (part) {
		GDStyle *style = part->getStrokeStyle();
		if (style) {
			if (drawStyle) {
				*drawStyle &= *style;
			} else {
				drawStyle = new GDStyle(*style);
			}
		}
		part = part->next;
	}
	return drawStyle;
}

GDStyle *ObjectGroup::getFillStyle() {
	if (fillStyle) delete fillStyle;
	fillStyle = NULL;
	GObject *part = parts;
	while (part) {
		GDStyle *style = part->getStrokeStyle();
		if (style) {
			if (fillStyle) {
				*fillStyle &= *style;
			} else {
				fillStyle = new GDStyle(*style);
			}
		}
		part = part->next;
	}
	return fillStyle;
}

void ObjectGroup::setStrokeStyle(const GDStyle &from) {
	GObject *part = parts;
	while (part) {
		part->setStrokeStyle(from);
		part = part->next;
	}
}

void ObjectGroup::setFillStyle(const GDStyle &from) {
	GObject *part = parts;
	while (part) {
		part->setFillStyle(from);
		part = part->next;
	}
}

void ObjectGroup::gatherResource(GDResourceList& collection, gint type) {
	GObject *part = parts;
	while (part) {
		part->gatherResource(collection, type);
		part = part->next;
	}
}

// GObject methods
void ObjectGroup::draw(OffscreenBuffer& osBuffer, RectangleArea& drawArea) {
	GObject *part = parts;
	while (part) {
		part->draw(osBuffer, drawArea);
		part = part->next;
	}
}

void ObjectGroup::drawBlack(OffscreenBuffer& osBuffer, RectangleArea& drawArea) {
	GObject *part = parts;
	while (part) {
		part->drawBlack(osBuffer, drawArea);
		part = part->next;
	}
}

void ObjectGroup::fillBlack(OffscreenBuffer& osBuffer, RectangleArea& drawArea) {
	GObject *part = parts;
	while (part) {
		part->fillBlack(osBuffer, drawArea);
		part = part->next;
	}
}

void ObjectGroup::drawWireFrame(DocWindow *docwin, Gdk_GC& context, RectangleArea& drawArea) {
	GObject *part = parts;
	while (part) {
		part->drawWireFrame(docwin, context, drawArea);
		part = part->next;
	}
}

void ObjectGroup::drawPoints(DocWindow *docwin, Gdk_GC& context, RectangleArea& drawArea) {
	GObject *part = parts;
	while (part) {
		part->drawPoints(docwin, context, drawArea);
		part = part->next;
	}
}

BezierPathVector *ObjectGroup::getBezierPaths() {
	BezierPathVector *result = NULL;
	GObject *part = parts;
	while (part) {
		BezierPathVector *paths = part->getBezierPaths();
		if (paths) {
			if (result) {
				for (BezierPathVector::iterator it = paths->begin(); it != paths->end(); it++) {
					result->push_back(*it);
				}
				delete paths;
			} else {
				result = paths;
			}
		}
		part = part->next;
	}
	return result;
}

BoundingBox ObjectGroup::getBoundingBox() {
	GObject *part = parts;
	BoundingBox bbox;
	while (part) {
		BoundingBox partbb = part->getBoundingBox();
		bbox.combine(partbb);
		part = part->next;
	}
	return bbox;
}

gdouble ObjectGroup::attractPoint(gdouble x, gdouble y, gdouble& ax, gdouble& ay, gdouble tolerance, int attractType) {
	gdouble nearest = G_MAXDOUBLE;
	gdouble temp_ax = ax, temp_ay = ay;
	GObject *part = parts;
	while (part) {
		gdouble res;
		if ((res = part->attractPoint(x, y, temp_ax, temp_ay, tolerance, attractType)) < nearest) {
			nearest = res;
			ax = temp_ax;
			ay = temp_ay;
		}
		part = part->next;
	}
	return nearest;
}

SelectableItem *ObjectGroup::checkPoints(gdouble x, gdouble y, gdouble tolerance, gboolean onlyCurvePoints) {
	SelectableItem *found = NULL;
	GObject *part = parts;
	while (part) {
		SelectableItem *check = part->checkPoints(x, y, tolerance, onlyCurvePoints);
		if (check) {
			found = check;
		}
		part = part->next;
	}
	return found;
}

GObject *ObjectGroup::checkArea(gdouble x, gdouble y) {
	GObject *part = parts;
	while (part) {
		if (part->checkArea(x, y)) {
			return this;
		}
		part = part->next;
	}
	return NULL;
}

SelectableItem *ObjectGroup::checkCurves(gdouble x, gdouble y, gdouble tolerance) {
	SelectableItem *found = NULL;
	GObject *part = parts;
	while (part) {
		SelectableItem *check = part->checkCurves(x, y, tolerance);
		if (check) {
			found = check;
		}
		part = part->next;
	}
	return found;
}

void ObjectGroup::selectRectangle(gdouble left, gdouble top, gdouble right, gdouble bottom, gint what) {
	GObject *part = parts;
	while (part) {
		part->selectRectangle(left, top, right, bottom, what);
		part = part->next;
	}
}

void ObjectGroup::deSelectRectangle(gdouble left, gdouble top, gdouble right, gdouble bottom, gint what) {
	GObject *part = parts;
	while (part) {
		part->deSelectRectangle(left, top, right, bottom, what);
		part = part->next;
	}
}

void ObjectGroup::selectAll() {
	GObject *part = parts;
	while (part) {
		part->selectAll();
		part = part->next;
	}
}

void ObjectGroup::deSelectAll() {
	GObject *part = parts;
	while (part) {
		part->deSelectAll();
		part = part->next;
	}
}

GObject *ObjectGroup::findObject(GObject *object) {
	GObject *part = parts;
	while (part) {
		GObject *found;
		if ((found = part->findObject(object))) {
			return found;
		}
		part = part->next;
	}
	return NULL;
}

void ObjectGroup::reverseOrder() {
	GObject *part = parts;
	while (part) {
		part->reverseOrder();
		part = part->next;
	}
}

GObject *ObjectGroup::removeObject(GObject *object) {
	if (object == this) {
		if (next) {
			next->prev = prev;
		}
		if (prev) {
			prev->next = next;
		}
		return this;
	}
	// try subparts
	GObject *part = parts;
	while (part) {
		if (part == object) {
			if (parts == object) {
				parts = parts->next;
			} else {
				parts->remove(object);
			}
			return part;
		} else {
			GObject *found;
			if ((found = part->removeObject(object))) {
				return found;
			}
		}
		part = part->next;
	}
	return NULL;
}

gint ObjectGroup::getSelection(SelectionItemList& selectionList, gint mode) {
	if (selected()) {
		selectionList.insert(selectionList.end(), new SelectionItem(this, mode));
		return 1;
	} else {
		GObject *part = parts;
		gint selectedParts = 0;
		while (part) {
			selectedParts += part->getSelection(selectionList, mode);
			part = part->next;
		}
		return selectedParts;
	}
}

gint ObjectGroup::writeXML(ostream& xmlout) {
	xmlout << "<g>\n";
	GObject *part = parts;
	while (part) {
		part->writeXML(xmlout);
		part = part->next;
	}
	xmlout << "</g>\n";
}

void ObjectGroup::writePSPathDescription(ostream& psout) {
	GObject *part = parts;
	while (part) {
		part->writePSPathDescription(psout);
		part = part->next;
	}
}

gint ObjectGroup::writePS(ostream& psout) {
	psout << "gs\n";
	GObject *part = parts;
	while (part) {
		part->writePS(psout);
		part = part->next;
	}
	psout << "gr\n";
	return 0;
}

// SelectableItem methods
void ObjectGroup::drawSelection(DocWindow *docwin, Gdk_GC& context, RectangleArea& drawArea) {
	GObject *part = parts;
	while (part) {
		part->drawSelection(docwin, context, drawArea);
		part = part->next;
	}
}

void ObjectGroup::transformByMatrix(TransformationMatrix& m, gboolean incFillPattern) {
	GObject *part = parts;
	while (part) {
		part->transformByMatrix(m, incFillPattern);
		part = part->next;
	}
}

void ObjectGroup::setFillMatrix(TransformationMatrix& m) {
	GObject *part = parts;
	while (part) {
		part->setFillMatrix(m);
		part = part->next;
	}
}


SelectableItem *ObjectGroup::clone() {
	ObjectGroup *newGroup = new ObjectGroup(doc);
	GObject *part = parts;
	while (part) {
		newGroup->addPart((GObject *) part->clone());
		part = part->next;
	}
	return newGroup;
}

void ObjectGroup::setCoords(SelectableItem* from) {
	GObject *part = parts;
	GObject *frompart = ((ObjectGroup *)from)->parts;
	while (part) {
		part->setCoords(frompart);
		part = part->next;
		frompart = frompart->next;
	}
}

void ObjectGroup::setDocument(GDDocument *document) {
	GObject *part = parts;
	while (part) {
		part->setDocument(document);
		part = part->next;
	}
}

gboolean ObjectGroup::selected() {
	GObject *part = parts;
	while (part) {
		if (! part->selected()) {
			return false;
		}
		part = part->next;
	}
	return true;
}

void ObjectGroup::clear() {
	deSelectAll();
	doc->removeObject(this);
	delete this;
}

gchar *ObjectGroup::identify() {
	return "ObjectGroup";
}


ObjectGroup::~ObjectGroup() {
	GObject *part = parts;
	while (part) {
		//g_printerr("ObjectGroup::~ObjectGroup remove Part %lX\n", part);
		GObject *nextObject = part->next;
		delete part;
		part = nextObject;
	}
	if (drawStyle) delete drawStyle;
	if (fillStyle) delete fillStyle;
}
