/* $Id: SVGInput.cc,v 1.20 2004/04/19 19:57:56 pklaus Exp $
 *
 * SVGInput.cc
 *
 * Gestalter - The free vector drawing application
 * Copyright (c) 2002 Philipp Klaus <gestalter@linotux.ch>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include "SVGInput.hh"

#include "GDApplication.hh"
#include "GDDocument.hh"
#include "Layer.hh"
#include "BezierPath.hh"
#include "BezierNode.hh"
#include "CompoundPath.hh"
#include "GDStyle.hh"
#include "PixelImage.hh"
#include "PixelImageData.hh"
#include "ObjectGroup.hh"
#include "ClipGroup.hh"
#include "StandardTextObject.hh"
#include "TextObject.hh"
#include "TextRow.hh"
#include "GDCharacter.hh"
#include "FontPool.hh"
#include "Gradient.hh"
#include "Pattern.hh"
#include "GradientListOrgDialog.hh"
#include "PatternListOrgDialog.hh"
#include <stdlib.h>

extern GDApplication *theApplication;

class SVGImage : public XMLOBJECT {
	PixelImageData *imageData;
	GDDocument *doc;
	guint pos;
	int size;
	TransformationMatrix tm;
	char hd[2];
public:
	SVGImage(XMLOBJECT *p, const char *tag, PARSERDATA parser, const char **attrs, GDDocument *doc) :
			XMLOBJECT(p, tag, parser, attrs) {
		//g_printerr("SVGImage::SVGImage\n");
		this->doc = doc;
		imageData = new PixelImageData();
		imageData->incRefCount();
		pos = 0;
		size = 0;
		gchar *width = getAttributeValue(attrs, "width");
		if (width) {
			imageData->width = atoi(width);
		}
		gchar *height = getAttributeValue(attrs, "height");
		if (height) {
			imageData->height = atoi(height);
		}
		gchar *transform = getAttributeValue(attrs, "transform");
		if (transform) {
			sscanf(transform, "matrix(%lf,%lf,%lf,%lf,%lf,%lf)", &tm.m[0][0], &tm.m[1][0], &tm.m[0][1], &tm.m[1][1], &tm.m[0][2], &tm.m[1][2]);
		}
		imageData->rowstride = imageData->width * 4;
		imageData->data = (guchar *) g_malloc(imageData->width * imageData->height * 4);
		//g_printerr("SVGImage::SVGImage finished\n");
	}
	void handleEndTag(const char *name) {
		//g_printerr("SVGImage::handleEndTag\n");
		PixelImage *image = new PixelImage(doc, imageData);
		TransformationMatrix tmp;
		tmp.setTranslation(doc->nullPoint.x, doc->nullPoint.y);
		tm *= tmp;
		image->transformByMatrix(tm);
		addObject(image);
		// call out superclass method
		XMLOBJECT::handleEndTag(name);
	}
	void handleText(const char *data, int len) {
		//g_printerr("SVGImage::handleText len:%d pos:%d\n", len, pos);
		while (len && pos < imageData->width * imageData->height * 4) {
			if ((*data <= 'F' && *data >= 'A') || (*data <= '9' && *data >= '0')) {
				hd[size++] = *data;
				if (size >= 2) {
					setDataByte(hd);
					size = 0;
				}
			} else if (*data & 0x80) {
				// utf-8, next character is special character
				data++;
				len--;
			}
			len--;
			data++;
		}
	}
private:
	inline void setDataByte(const char *hexData) {
		imageData->data[pos++] = (getNibble(hexData[0]) << 4) | (getNibble(hexData[1]) & 0x0F);
	}
	inline guchar getNibble(const char c) {
		return (c >= 'A') ? (c - 'A' + 10) : (c - '0');
	}
};

class SVGPath : public XMLOBJECT {
	GObject *gobject;
public:
	SVGPath(XMLOBJECT *p, const char *tag, PARSERDATA parser, const char **attrs, GDDocument *doc) :
			XMLOBJECT(p, tag, parser, attrs) {
		BezierPath *path = new BezierPath(doc);
		//g_printerr("SVGPath::SVGPath pathdesc=%s\n", getAttributeValue(attrs, "d"));
		const char *pathdesc = path->constructPathFromSVG(getAttributeValue(attrs, "d"));
		gobject = path;
		if (*pathdesc) {
			CompoundPath *cpath = new CompoundPath(doc);
			cpath->addPart(path);
			while (*pathdesc) {
				path = new BezierPath(doc);
				//g_printerr("SVGPath::SVGPath again pathdesc=%s\n", pathdesc);
				pathdesc = path->constructPathFromSVG(pathdesc);
				cpath->addPart(path);
			}
			gobject = cpath;
			//cpath->selectAll();
		}
		gobject->getFillStyle()->setFillStyleFromXML(this, attrs, doc);
		gobject->getStrokeStyle()->setStrokeStyleFromXML(this, attrs, doc);
	}
	void handleBeginTag(const char *name, const char **attrs) {
		// no sub-elements possible
	}
	void handleEndTag(const char *name) {
		addObject(gobject);
		// call our superclass method
		XMLOBJECT::handleEndTag(name);
	}
};


class SVGTextRow : public XMLOBJECT {
	TextRow *textrow;
	StandardTextObject *stdtext;
	GDCharacter *nC;
	GDDocument *doc;
public:
	SigC::Signal1<void, TextRow *> addTextRow;
	
	SVGTextRow(XMLOBJECT *p, const char *tag, PARSERDATA parser, const char **attrs, GDDocument *doc, StandardTextObject *textobj) :
			XMLOBJECT(p, tag, parser, attrs) {
		//g_printerr("SVGInput: got tspan element\n");
		this->doc = doc;
		stdtext = textobj;
		textrow = new TextRow();
		setDrawStyles(attrs);
		setTextStyles(attrs);
	}
	SVGTextRow(XMLOBJECT *p, const char *tag, PARSERDATA parser, const char **attrs, SVGTextRow& from) : XMLOBJECT(p, tag, parser, attrs) {
		doc = from.doc;
		stdtext = from.stdtext;
		textrow = from.textrow;
		setDrawStyles(attrs);
		setTextStyles(attrs);
	}
	
	void setDrawStyles(const char **attrs) {
		textrow->fillStyle.setFillStyleFromXML(this, attrs, doc);
		textrow->drawStyle.setStrokeStyleFromXML(this, attrs, doc);
	}
	void setTextStyles(const char **attrs) {
		gchar *fontname = getAttributeValue(attrs, "gestalter:font");
		if (fontname) {
			textrow->defaultStyle.setFont(doc->application->fontPool->getFont(fontname));
		} else {
			textrow->defaultStyle.setFont(doc->currentTextStyle.getFont());
		}
		gchar *fontsize = getAttributeValue(attrs, "font-size");
		if (fontsize) {
			double size = g_strtod(fontsize, NULL);
			if (size < 0) {
				size = 12;
			}
			textrow->defaultStyle.setPointSize(size);
		} else {
			textrow->defaultStyle.setPointSize(doc->currentTextStyle.getPointSize());
		}
		gchar *linespace = getAttributeValue(attrs, "gestalter:linespace");
		if (linespace) {
			if (! strncasecmp(linespace, "auto", 4)) {
				textrow->defaultStyle.setAutoLineSpace(true);
			} else {
				textrow->defaultStyle.setAutoLineSpace(false);
				double lspace = g_strtod(linespace, NULL);
				textrow->defaultStyle.setLineSpace(lspace);
			}
		}
		gchar *baselineshift = getAttributeValue(attrs, "gestalter:baselineshift");
		if (baselineshift) {
			double blshift = g_strtod(baselineshift, NULL);
			textrow->defaultStyle.setBaseLineShift(blshift);
		}
		gchar *tracking = getAttributeValue(attrs, "gestalter:tracking");
		if (tracking) {
			double track = g_strtod(tracking, NULL);
			textrow->defaultStyle.setTracking(track);
		}
		gchar *horizontalscale = getAttributeValue(attrs, "gestalter:horizontalscale");
		if (horizontalscale) {
			double hscale = g_strtod(horizontalscale, NULL);
			textrow->defaultStyle.setHorizontalScale(hscale);
		}
		gchar *justification = getAttributeValue(attrs, "text-anchor");
		if (justification) {
			if (! strncasecmp(justification, "middle", 6)) {
				textrow->defaultStyle.setJustification(TEXT_STYLE_JUSTIFICATION_CENTER);
			} else if (! strncasecmp(justification, "end", 3)) {
				textrow->defaultStyle.setJustification(TEXT_STYLE_JUSTIFICATION_RIGHT);
			} else if (! strncasecmp(justification, "start", 5)) {
				textrow->defaultStyle.setJustification(TEXT_STYLE_JUSTIFICATION_LEFT);
			}
		}
	}
	void handleBeginTag(const char *name, const char **attrs) {
		// no sub-elements possible
	}
	void handleEndTag(const char *name) {
		//g_printerr("SVGTextRow::handleEndTag: %s\n", name);
		addTextRow(textrow);
		//g_printerr("SVGTextRow::handleEndTag: textrow %lX added\n", textrow);
		// call our superclass method
		XMLOBJECT::handleEndTag(name);
	}
	void handleText(const char *data, int len) {
		//g_printerr("SVGTextRow::handleText len:%d\n", len);
		// decode UTF-8 and insert into row
		while (len) {
			guint val;
			val = 0;
			if (!(*data & 0x80)) {
				// ascii charcter
				val = (*data & 0x000000FF);
			} else if ((*data & 0xE0) == 0xC0) {
				// two byte character
				val = (*data & 0x0000001F) << 6;
				len--;
				data++;
				val |= (*data & 0x0000003F);
			} else if ((*data & 0xF0) == 0xE0) {
				// three byte character
				val = (*data & 0x0000000F) << 12;
				len--;
				data++;
				val |= (*data & 0x0000003F) << 6;
				len--;
				data++;
				val |= (*data & 0x0000003F);
			} else if ((*data & 0xF8) == 0xF0) {
				// four byte character
				val = (*data & 0x00000007) << 18;
				len--;
				data++;
				val |= (*data & 0x0000003F) << 12;
				len--;
				data++;
				val |= (*data & 0x0000003F) << 6;
				len--;
				data++;
				val |= (*data & 0x0000003F);
			}
			//GDCharacter *nC;
			nC = new GDCharacter(val);
			//g_printerr("SVGInput inserting character: %c, %lX\n", val, nC);
			nC->setTextStyle(textrow->defaultStyle);
			nC->setStrokeStyle(textrow->drawStyle);
			nC->setFillStyle(textrow->fillStyle);
			textrow->insertChar(-1, nC);
			len--;
			data++;
		}
	}
};


class SVGGradientStop : public XMLOBJECT {
public:
	SVGGradientStop(XMLOBJECT *p, const char *tag, PARSERDATA parser, const char **attrs, Gradient *gradient) :
			XMLOBJECT(p, tag, parser, attrs) {
		ColorPoint newPoint;
		gdouble pos = 0;
		getAttributeValue(attrs, "offset", pos);
		newPoint.pos = pos;
		newPoint.color.setFromString(getAttributeValue(attrs, "stop-color"));
		gradient->colorPoints.insert(gradient->colorPoints.end(), newPoint);
	}
};

class SVGGradient : public XMLOBJECT {
	SVGGradientStop *gStop;
	Gradient *gradient;
	GDDocument *doc;
public:
	SVGGradient(XMLOBJECT *p, const char *tag, PARSERDATA parser, const char **attrs, GDDocument *document, guint type) :
			XMLOBJECT(p, tag, parser, attrs) {
		gchar *id = getAttributeValue(attrs, "id");
		string idString = id;
		gchar *link = getAttributeValue(attrs, "xlink:href");
		if (link && *link == '#') {
			ResourceTransform gradTrans;
			string parent = (link + 1);
			gradTrans.resource = document->tmpGradients[parent];
			if (gradTrans.resource) {
				gchar *transform = getAttributeValue(attrs, "gradientTransform");
				if (transform) {
					sscanf(transform, "matrix(%lf,%lf,%lf,%lf,%lf,%lf)", 
						&gradTrans.matrix.m[0][0], 
						&gradTrans.matrix.m[1][0], 
						&gradTrans.matrix.m[0][1], 
						&gradTrans.matrix.m[1][1], 
						&gradTrans.matrix.m[0][2], 
						&gradTrans.matrix.m[1][2]);
					document->tmpResourceMap[idString] = gradTrans;
				}
			}
				
		} else {
			gradient = new Gradient();
			gradient->type = type;
			char *name = getAttributeValue(attrs, "gestalter:name");
			if (! name) {
				// backward compatibility
				name = getAttributeValue(attrs, "name");
			}
			if (name) {
				gradient->setName(name);
				string nameString = name;
				document->gradients[name] = gradient;
			}
			document->tmpGradients[id] = gradient;
		}
	}
	void handleBeginTag(const char *name, const char **attrs) {
		if (! strcmp(name, "stop")) {
			gStop = new SVGGradientStop(this, name, parser, attrs, gradient);
			gStop->init();
		}
	}
};

class SVGText;
class SVGGroup;
class SVGDefines;

class SVGPattern : public XMLOBJECT {
	ClipGroup *pattContent;
	SVGPath *path;
	SVGText *text;
	SVGImage *image;
	SVGGroup *group;
	SVGDefines *defines;
	Pattern *pattern;
	GDDocument *doc;
	string patternName;
	gboolean createPattern;
	string idString;
	gdouble width;
	gdouble height;
	gdouble origFactor;
	position docOrigin;
	
public:
	SVGPattern(XMLOBJECT *p, const char *tag, PARSERDATA parser, const char **attrs, GDDocument *document);
	void handleBeginTag(const char *name, const char **attrs);
	void handleEndTag(const char *name);
};

class SVGDefines : public XMLOBJECT {
	SVGGradient *gradient;
	SVGPattern *pattern;
	GDDocument *doc;
public:
	SVGDefines(XMLOBJECT *p, const char *tag, PARSERDATA parser, const char **attrs, GDDocument *document) :
			XMLOBJECT(p, tag, parser, attrs) {
		doc = document;
	}
	void handleBeginTag(const char *name, const char **attrs) {
		if (! strcmp(name, "linearGradient")) {
			gradient = new SVGGradient(this, name, parser, attrs, doc, GRADIENT_TYPE_LINEAR);
			gradient->init();
		}
		if (! strcmp(name, "radialGradient")) {
			gradient = new SVGGradient(this, name, parser, attrs, doc, GRADIENT_TYPE_RADIAL);
			gradient->init();
		}
		if (! strcmp(name, "pattern")) {
			pattern = new SVGPattern(this, name, parser, attrs, doc);
			pattern->init();
		}
	}
};


class SVGText : public XMLOBJECT {
	StandardTextObject *stdtext;
	SVGTextRow *textrow, *addObj;
	TransformationMatrix tm;
	GDDocument *doc;
	SVGDefines *defines;
public:
	SVGText(XMLOBJECT *p, const char *tag, PARSERDATA parser, const char **attrs, GDDocument *doc) :
			XMLOBJECT(p, tag, parser, attrs) {
		
		this->doc = doc;
		stdtext = new StandardTextObject(doc, 1);
		textrow = NULL;
		//g_printerr("SVGInput: got text element\n");
		gchar *transform = getAttributeValue(attrs, "transform");
		if (transform) {
			sscanf(transform, "matrix(%lf,%lf,%lf,%lf,%lf,%lf)", &tm.m[0][0], &tm.m[1][0], &tm.m[0][1], &tm.m[1][1], &tm.m[0][2], &tm.m[1][2]);
		}
	}
	void handleBeginTag(const char *name, const char **attrs) {
		//g_printerr("SVGInput: inside text element, got tag %s\n", name);
		if (! strcmp(name, "tspan")) {
			//g_printerr("SVGInput: inside text, found tspan\n");
			gchar *xcoord = getAttributeValue(attrs, "x");
			//g_printerr("SVGInput: tspan tag, attribute x: %s\n", xcoord);
			if (xcoord && g_strtod(xcoord, NULL) == 0.0) {
				textrow = new SVGTextRow(this, name, parser, attrs, doc, stdtext);
				textrow->addTextRow.connect(SigC::slot(stdtext, &StandardTextObject::appendRow));
				textrow->init();
			} else if (textrow) {
				addObj = new SVGTextRow(this, name, parser, attrs, *textrow);
				addObj->addTextRow.connect(SigC::slot(this, &SVGText::nullAddTextRow));
				addObj->init();
			}
		}
		if (! strcmp(name, "defs")) {
			defines = new SVGDefines(this, name, parser, attrs, doc);
			defines->init();
		}
	}
	void nullAddTextRow(TextRow *dummy) {
		// do nothing
	}
	void handleEndTag(const char *name) {
		//g_printerr("SVGInput: found tag end %s\n", name);
		TransformationMatrix tmp;
		tmp.setTranslation(doc->nullPoint.x, doc->nullPoint.y);
		tm *= tmp;
		stdtext->transformByMatrix(tm, FALSE);
		addObject(stdtext);
		// call our superclass method
		XMLOBJECT::handleEndTag(name);
	}
};

class SVGClipPath;

class SVGGroup : public XMLOBJECT {
	SVGPath *path;
	SVGText *text;
	SVGImage *image;
	GDDocument *doc;
	ObjectGroup *newGroup;
	ClipGroup *clipGroup;
	SVGGroup *group;
	SVGClipPath *clipPath;
	SVGDefines *defines;
	gboolean isDummy;
public:
	SVGGroup(XMLOBJECT *p, const char *tag, PARSERDATA parser, const char **attrs, GDDocument *doc, ClipGroup* parentClip);

	void handleBeginTag(const char *name, const char **attrs);
	void handleEndTag(const char *name);
};

SVGPattern::SVGPattern(XMLOBJECT *p, const char *tag, PARSERDATA parser, const char **attrs, GDDocument *document) :
			XMLOBJECT(p, tag, parser, attrs) {
	doc = document;
	pattContent = NULL;
	createPattern = false;
	docOrigin = document->nullPoint;
	document->nullPoint.x = 0; // set document nullPoint to disable translation of pattern content
	document->nullPoint.y = 0;
	width = 0;
	height = 0;
	origFactor = 1.0;
	gchar *id = getAttributeValue(attrs, "id");
	idString = id;
	//g_printerr("SVGPattern::SVGPattern id=%s\n", id);
	gchar *link = getAttributeValue(attrs, "xlink:href");
	if (link && *link == '#') {
		ResourceTransform pattTrans;
		string parent = (link + 1);
		pattTrans.resource = document->tmpPatterns[parent];
		if (pattTrans.resource) {
			gchar *transform = getAttributeValue(attrs, "patternTransform");
			if (transform) {
				sscanf(transform, "matrix(%lf,%lf,%lf,%lf,%lf,%lf)", 
					&pattTrans.matrix.m[0][0], 
					&pattTrans.matrix.m[1][0], 
					&pattTrans.matrix.m[0][1], 
					&pattTrans.matrix.m[1][1], 
					&pattTrans.matrix.m[0][2], 
					&pattTrans.matrix.m[1][2]);
				document->tmpResourceMap[idString] = pattTrans;
			}
		}
			
	} else {
		createPattern = true;
		pattContent = new ClipGroup(doc);
		//g_printerr("SVGPattern::SVGPattern created new ClipGroup %lX\n", (long) pattContent);
		gchar *name = getAttributeValue(attrs, "gestalter:name");
		if (name) {
			patternName = name;
		}
		gchar *w = getAttributeValue(attrs, "width");
		if (w) {
			width = atof(w);
		}
		gchar *h = getAttributeValue(attrs, "height");
		if (h) {
			height = atof(h);
		}
		gchar *of = getAttributeValue(attrs, "gestalter:origFactor");
		if (of) {
			origFactor = atof(of);
		}
		// set the clipping rectangle
		gdouble tlx = 0;
		gdouble tly = 0;
		gdouble brx = width;
		gdouble bry = height;
		BezierPath *path = new BezierPath(doc);
		BezierNode *n1 = new BezierNode(path);
		n1->setOnCurvePoint(tlx, tly);
		n1->setPreviousHandle(tlx, tly);
		n1->setNextHandle(tlx, tly);
		path->addNode(n1);
		BezierNode *n2 = new BezierNode(path);
		n2->setOnCurvePoint(brx, tly);
		n2->setPreviousHandle(brx, tly);
		n2->setNextHandle(brx, tly);
		path->addNode(n2);
		BezierNode *n3 = new BezierNode(path);
		n3->setOnCurvePoint(brx, bry);
		n3->setPreviousHandle(brx, bry);
		n3->setNextHandle(brx, bry);
		path->addNode(n3);
		BezierNode *n4 = new BezierNode(path);
		n4->setOnCurvePoint(tlx, bry);
		n4->setPreviousHandle(tlx, bry);
		n4->setNextHandle(tlx, bry);
		path->addNode(n4);
		n4->next = n1;
		n1->prev = n4; // close the path
		pattContent->addClipPart(path);
	}
}
void SVGPattern::handleBeginTag(const char *name, const char **attrs) {
	//g_printerr("SVGPattern::handleBeginTag name=%s\n", name);
	if (! strcmp(name, "path")) {
		path = new SVGPath(this, name, parser, attrs, doc);
		path->addObject.connect(SigC::slot(pattContent, &ClipGroup::addPart));
		path->init();
	}
	if (! strcmp(name, "text")) {
		text = new SVGText(this, name, parser, attrs, doc);
		text->addObject.connect(SigC::slot(pattContent, &ClipGroup::addPart));
		text->init();
	}
	if (! strcmp(name, "image")) {
		image = new SVGImage(this, name, parser, attrs, doc);
		image->addObject.connect(SigC::slot(pattContent, &ClipGroup::addPart));
		image->init();
	}
	if (! strcmp(name, "defs")) {
		defines = new SVGDefines(this, name, parser, attrs, doc);
		defines->init();
	}
	if (! strcmp(name, "g")) {
		group = new SVGGroup(this, name, parser, attrs, doc, NULL);
		group->init();
		group->addObject.connect(SigC::slot(pattContent, &ClipGroup::addPart));
	}
}
void SVGPattern::handleEndTag(const char *name) {
	//g_printerr("SVGPattern::handleEndTag name=%s\n", name);
	doc->nullPoint = docOrigin;
	if (createPattern) {
		//g_printerr("SVGPattern::handleEndTag id=%s name=%s patt=%lX\n", idString.c_str(), patternName.c_str(), (long) pattern);
		pattern = new Pattern(pattContent, width, height, origFactor);
		pattern->setName(patternName);
		doc->patterns[patternName] = pattern;
		doc->tmpPatterns[idString] = pattern;
	}
	XMLOBJECT::handleEndTag(name);
}

class SVGClipPath : public XMLOBJECT {
	SVGPath *path;
	SVGText *text;
	SVGImage *image;
	Layer *layer;
	GDDocument *doc;
	ObjectGroup *newGroup;
	ClipGroup *clipGroup;
	SVGGroup *group;
	SVGClipPath *clipPath;
	SVGDefines *defines;
public:
	SVGClipPath(XMLOBJECT *p, const char *tag, PARSERDATA parser, const char **attrs, GDDocument *doc, ClipGroup* parentClip) :
			XMLOBJECT(p, tag, parser, attrs) {
		this->doc = doc;
		this->clipGroup = parentClip;
	}
	void handleBeginTag(const char *name, const char **attrs) {
		if (! strcmp(name, "path")) {
			path = new SVGPath(this, name, parser, attrs, doc);
			path->addObject.connect(SigC::slot(clipGroup, &ClipGroup::addClipPart));
			path->init();
		}
		if (! strcmp(name, "text")) {
			text = new SVGText(this, name, parser, attrs, doc);
			text->addObject.connect(SigC::slot(newGroup, &ObjectGroup::addPart));
			text->init();
		}
		if (! strcmp(name, "image")) {
			image = new SVGImage(this, name, parser, attrs, doc);
			image->addObject.connect(SigC::slot(clipGroup, &ClipGroup::addClipPart));
			image->init();
		}
		if (! strcmp(name, "defs")) {
			defines = new SVGDefines(this, name, parser, attrs, doc);
			defines->init();
		}
		if (! strcmp(name, "g")) {
			group = new SVGGroup(this, name, parser, attrs, doc, clipGroup);
			group->addObject.connect(SigC::slot(clipGroup, &ClipGroup::addClipPart));
			group->init();
		}
	}
};

SVGGroup::SVGGroup(XMLOBJECT *p, const char *tag, PARSERDATA parser, const char **attrs, GDDocument *doc, ClipGroup* parentClip) :
			XMLOBJECT(p, tag, parser, attrs) {
	this->doc = doc;
	if (getAttributeValue(attrs, "clip-rule")) {
		//g_printerr("SVGGroup::SVGGroup is ClipGroup\n");
		clipGroup = new ClipGroup(doc);
		newGroup = clipGroup;
	} else if (parentClip) {
		//g_printerr("SVGGroup::SVGGroup is content group of ClipGroup\n");
		newGroup = parentClip;
		clipGroup = NULL;
	} else {
		//g_printerr("SVGGroup::SVGGroup is ObjectGroup\n");
		newGroup = new ObjectGroup(doc);
		clipGroup = NULL;
	}
	isDummy = false;
	if (parentClip) {
		isDummy = true;
	}
}

void SVGGroup::handleBeginTag(const char *name, const char **attrs) {
	//g_printerr("SVGInput: SVGGroup found subtag %s\n", name);
	if (! strcmp(name, "clipPath")) {
		clipPath = new SVGClipPath(this, name, parser, attrs, doc, clipGroup);
		clipPath->init();
	}
	if (! strcmp(name, "path")) {
		path = new SVGPath(this, name, parser, attrs, doc);
		path->addObject.connect(SigC::slot(newGroup, &ObjectGroup::addPart));
		path->init();
	}
	if (! strcmp(name, "text")) {
		text = new SVGText(this, name, parser, attrs, doc);
		text->addObject.connect(SigC::slot(newGroup, &ObjectGroup::addPart));
		text->init();
	}
	if (! strcmp(name, "image")) {
		image = new SVGImage(this, name, parser, attrs, doc);
		image->addObject.connect(SigC::slot(newGroup, &ObjectGroup::addPart));
		image->init();
	}
	if (! strcmp(name, "defs")) {
		defines = new SVGDefines(this, name, parser, attrs, doc);
		defines->init();
	}
	if (! strcmp(name, "g")) {
		group = new SVGGroup(this, name, parser, attrs, doc, clipGroup);
		group->init();
		group->addObject.connect(SigC::slot(newGroup, &ObjectGroup::addPart));
	}
}
void SVGGroup::handleEndTag(const char *name) {
	if (! isDummy) {
		addObject(newGroup);
	}
	// call out superclass method
	XMLOBJECT::handleEndTag(name);
}

class SVGLayer : public XMLOBJECT {
	SVGPath *path;
	SVGText *text;
	SVGImage *image;
	Layer *layer;
	GDDocument *doc;
	SVGGroup *group;
	SVGDefines *defines;
public:
	SVGLayer(XMLOBJECT *p, const char *tag, PARSERDATA parser, const char **attrs, GDDocument *doc) :
			XMLOBJECT(p, tag, parser, attrs) {
		this->doc = doc;
		layer = new Layer(doc);
		char *name = getAttributeValue(attrs, "gestalter:layer");
		if (name) {
			layer->setName(name);
		}
		char *selcol = getAttributeValue(attrs, "gestalter:selcolor");
		if (selcol) {
			if (*selcol == '#') {
				selcol++;
			}
			layer->setSelectionColor((strtol(selcol, (char **) NULL, 16) << 8) | 0xFF);
		}
		char *view = getAttributeValue(attrs, "gestalter:viewable");
		if (view && ! strncasecmp(view, "no", 2)) {
			layer->viewable = false;
		}
		char *edit = getAttributeValue(attrs, "gestalter:editable");
		if (edit && ! strncasecmp(edit, "no", 2)) {
			layer->editable = false;
		}
		doc->addLayer(layer);
		doc->activeLayer = layer;
	}
	void handleBeginTag(const char *name, const char **attrs) {
		//g_printerr("SVGInput: in Layer found subtag %s\n", name);
		if (! strcmp(name, "path")) {
			path = new SVGPath(this, name, parser, attrs, doc);
			path->addObject.connect(SigC::slot(layer, &Layer::addObject));
			path->init();
		}
		if (! strcmp(name, "text")) {
			text = new SVGText(this, name, parser, attrs, doc);
			text->addObject.connect(SigC::slot(layer, &Layer::addObject));
			text->init();
		}
		if (! strcmp(name, "image")) {
			image = new SVGImage(this, name, parser, attrs, doc);
			image->addObject.connect(SigC::slot(layer, &Layer::addObject));
			image->init();
		}
		if (! strcmp(name, "defs")) {
			defines = new SVGDefines(this, name, parser, attrs, doc);
			defines->init();
		}
		if (! strcmp(name, "g")) {
			group = new SVGGroup(this, name, parser, attrs, doc, NULL);
			group->addObject.connect(SigC::slot(layer, &Layer::addObject));
			group->init();
		}
	}
};

class SVGDocument : public XMLOBJECT {
	SVGLayer *layer;
	SVGDefines *defines;
	GDDocument *doc;
public:
	SVGDocument(XMLOBJECT *p, const char *tag, PARSERDATA parser, const char **attrs) :
			XMLOBJECT(p, tag, parser, attrs) {
		doc = new GDDocument(theApplication);
		// we don't want the automatic creation of a new layer
		delete doc->layers;
		doc->layers = NULL;
		doc->layerNumber = 0;
		char *pageX = getAttributeValue(attrs, "gestalter:pageX");
		if (pageX) {
			doc->nullPoint.x = atoi(pageX);
		}
		char *pageY = getAttributeValue(attrs, "gestalter:pageY");
		if (pageY) {
			doc->nullPoint.y = atoi(pageY);
		}
		theApplication->addDocument(doc);
	}
	void handleBeginTag(const char *name, const char **attrs) {
		if (! strcmp(name, "g")) {
			layer = new SVGLayer(this, name, parser, attrs, doc);
			layer->init();
		}
		if (! strcmp(name, "defs")) {
			defines = new SVGDefines(this, name, parser, attrs, doc);
			defines->init();
		}
	}
	void handleEndTag(const char *name) {
		// clean up temporary structures needed during document load
		doc->tmpGradients.clear();
		doc->tmpPatterns.clear();
		doc->tmpResourceMap.clear();
		// draw our new document
		doc->redraw();
		XMLOBJECT::handleEndTag(name);
	}
	void setDocFile(string documentfile) {
		doc->setFilePath(documentfile);
	}
};

class SVGRoot : public XMLOBJECT {
	
	SVGDocument *document;
	string docfile;
	
public:
	SVGRoot(XMLOBJECT *p, const char *tag, PARSERDATA parser, const char **attrs) :
			XMLOBJECT(p, tag, parser, attrs) {
	}
	void handleBeginTag(const char *name, const char **attrs) {
		if (! strcmp(name, "svg")) {
			document = new SVGDocument(this, name, parser, attrs);
			document->init();
			document->setDocFile(docfile);
		}
	}
	void setDocFile(string documentfile) {
		docfile = documentfile;
	}
};

void SVGInput::open(string filePath) {
	//g_printerr("Called SVGInput::open(%s)\n", filePath.c_str());
	XMLPARSER *xmlParser = new XMLPARSER();
	SVGRoot *svgr = new SVGRoot(NULL, "svg", NULL, NULL);
	svgr->setDocFile(filePath);
	xmlParser->parse(svgr, filePath);
}
