/* $Id: SelectObjectTool.cc,v 1.9 2004/03/29 16:06:50 pklaus Exp $
 *
 * SelectObjectTool.cc
 *
 * Gestalter - The free vector drawing application
 * Copyright (c) 2002 Philipp Klaus <gestalter@linotux.ch>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include <math.h>
#include "../pixmaps/obj_select.xpm"
#include "../cursors/select_object.xbm"
#include "../cursors/select_object_mask.xbm"
#include "../cursors/move.xbm"
#include "../cursors/move_mask.xbm"
#include "../cursors/move_magnet.xbm"
#include "../cursors/move_magnet_mask.xbm"
#include "../cursors/move_copy.xbm"
#include "../cursors/move_copy_mask.xbm"
#include "../cursors/move_copy_magnet.xbm"
#include "../cursors/move_copy_magnet_mask.xbm"
#include "config.h"
#include "SelectObjectTool.hh"
#include "GDDocument.hh"
#include "CursorFactory.hh"
#include "BezierNode.hh"
#include "BezierPath.hh"
#include "SelectionItemList.hh"
#include "DocWindow.hh"
#include "GDApplication.hh"
#include "TransformAction.hh"

extern TransformationMatrix m;

SelectObjectTool::SelectObjectTool() {
	downButton = -1;
	pixmapData = obj_select_xpm;
	toolName = "SelectObjectTool";
	moving = false;
	nodeCaught = false;

	//cursors
	CursorFactory::createCursorFromData(selectObjectCursor,
					select_object_bits, select_object_mask_bits,
					select_object_width, select_object_height,
					select_object_x_hot, select_object_y_hot);
	CursorFactory::createCursorFromData(moveCursor,
					move_bits, move_mask_bits,
					move_width, move_height,
					move_x_hot, move_y_hot);
	CursorFactory::createCursorFromData(moveMagnetCursor,
					move_magnet_bits, move_magnet_mask_bits,
					move_magnet_width, move_magnet_height,
					move_magnet_x_hot, move_magnet_y_hot);
	CursorFactory::createCursorFromData(moveCopyCursor,
					move_copy_bits, move_copy_mask_bits,
					move_copy_width, move_copy_height,
					move_copy_x_hot, move_copy_y_hot);
	CursorFactory::createCursorFromData(moveCopyMagnetCursor,
					move_copy_magnet_bits, move_copy_magnet_mask_bits,
					move_copy_magnet_width, move_copy_magnet_height,
					move_copy_magnet_x_hot, move_copy_magnet_y_hot);
}

void SelectObjectTool::activate() {
	Tool::activate();
}

void SelectObjectTool::deactivate() {
	Tool::deactivate();
}

gint SelectObjectTool::buttonDown(GdkEventButton *ev, DocWindow *docwin) {

	GDDocument	*doc;
	
	downButton = ev->button;
	movingNode = NULL;
	selectNode = NULL;
	
	if (downButton == 1) {
		doc = docwin->doc;
		shiftSelect = false;
		sx = ev->x / docwin->osBuffer.scaleFactor + docwin->osBuffer.origin.x;
		sy = ev->y / docwin->osBuffer.scaleFactor + docwin->osBuffer.origin.y;
		SelectableItem *bnode = doc->checkPoints(
					sx, sy, 
					doc->application->preferences.catchingRange / docwin->osBuffer.scaleFactor,
					true);
		if (bnode) {
			// node directly hit
			bnode->getCoordinates(sx, sy);
		} else {
			// no point hit, check curves
			bnode = doc->checkCurves(
							sx, sy, 
							doc->application->preferences.catchingRange / docwin->osBuffer.scaleFactor);
		}
		if (! bnode) {
			// no point and no curve hit, check areas
			bnode = doc->checkArea(sx, sy);
		}
		if (ev->state & GDK_SHIFT_MASK) {
			shiftSelect = true;
		}
		if (bnode != NULL) {
			if (! (shiftSelect || bnode->selected())) {
				doc->deSelectAll();
			}
			// save the node for eventually later selecting it
			bnode->selectObject(shiftSelect);
			// an object should be moved
			movingNode = bnode;
			selection = new SelectionItemList();
			doc->getSelection(*selection);
			// draw selection a first time to make xor operations correct
			doc->context.set_function(GDK_INVERT);
			selection->startTransformation(docwin, doc->context);
		} else {
			if (! (ev->state & GDK_SHIFT_MASK)) {
				// deselect everything
				doc->deSelectAll();
			}
		}
		startx = (gint) ev->x;
		starty = (gint) ev->y;
		oldx = startx;
		oldy = starty;
		moving = true;
		setCursor(ev->state, docwin);
	} else {
		Tool::buttonDown(ev, docwin);
	}
	return 0;
}

gint SelectObjectTool::motion(GdkEventMotion *ev, DocWindow *docwin) {
	
	
	if (ev->state & GDK_BUTTON1_MASK) {
		GDDocument *doc = docwin->doc;
		doc->context.set_function(GDK_INVERT);
		if (movingNode) {
			// move the selected node
			gdouble dx = ev->x / docwin->osBuffer.scaleFactor + docwin->osBuffer.origin.x;
			gdouble dy = ev->y / docwin->osBuffer.scaleFactor + docwin->osBuffer.origin.y;
			nodeCaught = magnet(dx, dy, docwin);
			setCursor(ev->state, docwin);
			if (ev->state & GDK_SHIFT_MASK) {
				// single axis/ equal axis movement
				gdouble deltax = fabs(dx - sx);
				gdouble deltay = fabs(dy - sy);
				if (deltax > 0 && deltay > 0 && deltax / deltay > 0.5 && deltax / deltay < 2) {
					gdouble delta = deltax > deltay ? deltax : deltay;
					dx = sx + copysign(delta, dx - sx);
					dy = sy + copysign(delta, dy - sy);
				} else if (deltax > deltay) {
					dy = sy;
				} else {
					dx = sx;
				}
			}
			dx -= sx;
			dy -= sy;
			m.reset();
			m.setTranslation(dx, dy);
			selection->transformAndDraw(docwin, doc->context, m);
		} else {
			// set the cursor
			Gdk_Window drawable = docwin->DocWinDrawingArea->get_window();
			setCursor(ev->state, docwin);
			// draw the elastic rectangle
			drawRectangle(drawable, doc->context, startx, starty, oldx, oldy);
			drawRectangle(drawable, doc->context, startx, starty, (int) ev->x, (int) ev->y);
		}
		oldx = (gint) ev->x;
		oldy = (gint) ev->y;
	} else if (ev->state & (GDK_BUTTON2_MASK | GDK_BUTTON3_MASK)) {
		Tool::motion(ev, docwin);
	} else {
		// set the cursor
		setCursor(ev->state, docwin);
	}
	
  return 0;
}

gint SelectObjectTool::buttonUp(GdkEventButton *ev, DocWindow *docwin) {
	if (downButton == 1) {
		GDDocument *doc = docwin->doc;
		if (movingNode) {
			if (ev->state & GDK_MOD1_MASK) {
				TransformationMatrix lastTransform = selection->getLastTransformation();
				if (! lastTransform.isIdentityMatrix()) {
					TransformAction *newAction = new TransformAction(lastTransform, true);
					doc->setAction(newAction);
					// insert a copy of the selection
					//g_printerr("SelectPointTool::butonUp insert a copy\n");
					doc->deSelectAll();
					selection->insertMoving(docwin->doc);
				}
			} else {
				TransformationMatrix lastTransform = selection->getLastTransformation();
				if (! lastTransform.isIdentityMatrix()) {
					TransformAction *newAction = new TransformAction(lastTransform);
					doc->setAction(newAction);
					// apply the movement to the original
					//g_printerr("SelectPointTool::butonUp apply to original\n");
					selection->applyToOriginal();
				}
			}
			selection->endTransformation(docwin, doc->context);
			delete selection;
			selection = NULL;
		} else {
			gdouble l,t,r,b;
			if (startx > oldx) {
				l = oldx / docwin->osBuffer.scaleFactor + docwin->osBuffer.origin.x;
				r = startx / docwin->osBuffer.scaleFactor + docwin->osBuffer.origin.x;
			} else {
				l = startx / docwin->osBuffer.scaleFactor + docwin->osBuffer.origin.x;
				r = oldx / docwin->osBuffer.scaleFactor + docwin->osBuffer.origin.x;
			}
			if (starty > oldy) {
				t = oldy / docwin->osBuffer.scaleFactor + docwin->osBuffer.origin.y;
				b = starty / docwin->osBuffer.scaleFactor + docwin->osBuffer.origin.y;
			} else {
				t = starty / docwin->osBuffer.scaleFactor + docwin->osBuffer.origin.y;
				b = oldy / docwin->osBuffer.scaleFactor + docwin->osBuffer.origin.y;
			}
			int sS = 0;
			if (shiftSelect) sS = 128;
			doc->selectRectangle(l, t, r, b, 2 | sS);
		}
		doc->context.set_function(GDK_SET);
		doc->redraw();
		moving = false;
		setCursor(ev->state, docwin);
	}
	downButton = 0;
	return 0;
}

void SelectObjectTool::setCursor(guint state, DocWindow *docwin) {
	if (moving) {
		if (state & GDK_MOD1_MASK) {
			if (nodeCaught) {
				docwin->DocWinDrawingArea->get_window().set_cursor(moveCopyMagnetCursor);
			} else {
				docwin->DocWinDrawingArea->get_window().set_cursor(moveCopyCursor);
			}
		} else {
			if (nodeCaught) {
				docwin->DocWinDrawingArea->get_window().set_cursor(moveMagnetCursor);
			} else {
				docwin->DocWinDrawingArea->get_window().set_cursor(moveCursor);
			}
		}
	} else {
		docwin->DocWinDrawingArea->get_window().set_cursor(selectObjectCursor);
	}
}
