/* $Id: TextObject.cc,v 1.18 2004/03/29 16:06:50 pklaus Exp $
 *
 * TextObject.cc
 *
 * Gestalter - The free vector drawing application
 * Copyright (c) 2002 Philipp Klaus <gestalter@linotux.ch>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include "TextObject.hh"
#include "GDDocument.hh"
#include "BezierNode.hh"
#include "FontPool.hh"
#include "GDStyle.hh"

TextObject::TextObject(GDDocument *document) : GObject(document) {
	drawStyle = NULL;
	fillStyle = NULL;
	lastBuffer = NULL;
	currBuffer = NULL;
	editing = true;
}

guint TextObject::getOutcode(position point,
					gdouble left, gdouble top, gdouble right, gdouble bottom) {
	guint code = 0;
	if (point.y > bottom) code |= 1;
	if (point.y < top) code |= 2;
	if (point.x > right) code |= 4;
	if (point.x < left) code |= 8;
	return code;
}

gboolean TextObject::checkRectangleLineIntersection(position start, position end,
					gdouble left, gdouble top, gdouble right, gdouble bottom) {
	/* implementation of Cohen-Sutterland
	   description in Computer Graphics, Principles and Practice,
	   Addison Wesley, 1997, page 112ff */
	
	guint scode = getOutcode(start, left, top, right, bottom);
	guint ecode = getOutcode(end, left, top, right, bottom);
	if (scode == 0 || ecode == 0) {
		return true; // one of the endpoints is within the rectangle
	}
	if (scode & ecode) {
		return false; // both endpoints are in one outarea
	}
	for (int i = 0; i < 4; i++) {
		if (scode & (1 << i) == 0) {
			continue;
		}
		if (i == 0) {
			start.x = start.x + (end.x - start.x) * (bottom - start.y) / (end.y - start.y);
			start.y = bottom;
		} else if (i == 1) {
			start.x = start.x + (end.x - start.x) * (top - start.y) / (end.y - start.y);
			start.y = top;
		} else if (i == 2) {
			start.y = start.y + (end.y - start.y) * (right - start.x) / (end.x - start.x);
			start.x = right;
		} else if (i == 2) {
			start.y = start.y + (end.y - start.y) * (left - start.x) / (end.x - start.x);
			start.x = left;
		}
		scode = getOutcode(start, left, top, right, bottom);
		if (! scode) {
			return true;
		}
		if (scode & ecode) {
			return false;
		}
	}
	return false;
}

int TextObject::bufferCopyCharBitmap(guchar *buffer, guint buffsize, guint rowstride, FT_Glyph *glyph,
								gint x, gint y) {
	int error = FT_Glyph_To_Bitmap(glyph, ft_render_mode_normal, 0, 1);
	if (error) return error;
	FT_BitmapGlyph  bitmap = (FT_BitmapGlyph)(*glyph);
	//g_printerr("TextObject::copyCharBitmap buffer %lX\n", buffer);
	//g_printerr("TextObject::copyCharBitmap x: %d, y: %d, width: %d, height: %d\n", x, y, bitmap->bitmap.width, bitmap->bitmap.rows);
	//g_printerr("TextObject::copyCharBitmap left: %d top: %d rowstride: %d\n", bitmap->left, bitmap->top, rowstride);
	int i;
	if (x + bitmap->left + bitmap->bitmap.width < 0) return 1;
	for (i = 0; i < bitmap->bitmap.rows; i++) {
		guint xn = 0;
		if (x + bitmap->left < 0) xn = -(x + bitmap->left);
		guchar *sbase = bitmap->bitmap.buffer + i * bitmap->bitmap.width + xn;
		guchar *tbase = buffer + 
						(y - bitmap->top + i) * rowstride + 
						(x + bitmap->left + xn) * 3;
		guint drawwidth = x + bitmap->left + bitmap->bitmap.width;
		if (drawwidth > rowstride / 3) {
			drawwidth = rowstride / 3;
		}
		drawwidth -= x + bitmap->left;
		for (; xn < drawwidth; xn++) {
			if (tbase + 3 > buffer + buffsize) return 1;
			int val = 255 - *sbase;
			tbase[0] = (guchar) ((tbase[0] * val) >> 8);
			//tbase[0] = 0;
			tbase[1] = (guchar) ((tbase[1] * val) >> 8);
			//tbase[1] = 0;
			tbase[2] = (guchar) ((tbase[2] * val) >> 8);
			//tbase[2] = 0;
			//g_printerr("%02X%02X%02X ", R,G,B);
			sbase++;
			tbase += 3;
		}
		//g_printerr("\n");
	}
	//g_printerr("TextObject::copyCharBitmap finished\n");
	return 0;
}

int TextObject::copyCharBitmap(Gdk_Window window, Gdk_GC context, 
										FT_Glyph *glyph, gint x, gint y) {
	int error = FT_Glyph_To_Bitmap(glyph, ft_render_mode_normal, 0, 1);
	if (error) return error;
	FT_BitmapGlyph bitmap = (FT_BitmapGlyph) (*glyph);
	guchar *buffer = (guchar *) g_malloc(bitmap->bitmap.rows * bitmap->bitmap.width * 3);
	for (int i = 0; i < bitmap->bitmap.rows; i++) {
		guchar *sbase = bitmap->bitmap.buffer + i * bitmap->bitmap.width;
		guchar *tbase = buffer + i * bitmap->bitmap.width * 3;
		for (int j = 0; j < bitmap->bitmap.width; j++) {
			int val = 255 - *sbase;
			tbase[0] = val;
			tbase[1] = val;
			tbase[2] = val;
			sbase++;
			tbase += 3;
		}
	}
	//g_printerr("TextObject::copyCharBitmap: x: %d, y: %d\n", x + bitmap->left, y - bitmap->top);
	gdk_draw_rgb_image(window, context, x + bitmap->left, y - bitmap->top,
				bitmap->bitmap.width, bitmap->bitmap.rows,
				GDK_RGB_DITHER_NONE, buffer, bitmap->bitmap.width * 3);
	g_free(buffer);
	return 0;
			
}

int TextObject::selCopyCharBitmap(Gdk_Window window, Gdk_GC context, FT_Glyph *glyph, gint x, gint y) {
	int error = FT_Glyph_To_Bitmap(glyph, ft_render_mode_normal, 0, 1);
	if (error) return error;
	FT_BitmapGlyph bitmap = (FT_BitmapGlyph) (*glyph);
	if (bitmap->bitmap.width <= 0 || bitmap->bitmap.rows <= 0) {
		// characters without bits don't have to be drawn
		return 0;
	}
	guchar *buffer = (guchar *) g_malloc(bitmap->bitmap.rows * bitmap->bitmap.width * 3);
	for (int i = 0; i < bitmap->bitmap.rows; i++) {
		guchar *sbase = bitmap->bitmap.buffer + i * bitmap->bitmap.width;
		guchar *tbase = buffer + i * bitmap->bitmap.width * 3;
		for (int j = 0; j < bitmap->bitmap.width; j++) {
			int val = *sbase;
			tbase[0] = val;
			tbase[1] = val;
			tbase[2] = val;
			sbase++;
			tbase += 3;
		}
	}
	//g_printerr("TextObject::selCopyCharBitmap: x: %d, y: %d\n", x + bitmap->left, y - bitmap->top);
	gdk_draw_rgb_image(window, context, x + bitmap->left, y - bitmap->top,
				bitmap->bitmap.width, bitmap->bitmap.rows,
				GDK_RGB_DITHER_NONE, buffer, bitmap->bitmap.width * 3);
	g_free(buffer);
	return 0;
			
}

void TextObject::drawPoint(DocWindow *docwin, Gdk_GC& context, position point) {
	docwin->DocWinDrawingArea->get_window().draw_rectangle(context, (selected() ? 1 : 0), 
		ROUND(point.x) - BEZIER_HALFSIZE, ROUND(point.y) - BEZIER_HALFSIZE,
						BEZIER_SIZE, BEZIER_SIZE);
}	

GObject *TextObject::removeObject(GObject *object) {
	if (object == this) {
		if (next) next->prev = prev;
		if (prev) prev->next = next;
		return this;
	}
	return NULL;
}

GObject *TextObject::findObject(GObject *object) {
	if (object == this) {
		return this;
	}
	return NULL;
}

void TextObject::selectAll() {
	//g_printerr("TextObject::selectAll\n");
	selection = true;
	doc->selection->insertObject(this);
}

void TextObject::deSelectAll() {
	selection = false;
	editing = false;
	doc->selection->removeObject(this);
}

gint TextObject::getSelection(SelectionItemList& selectionList, gint mode) {
	if (selected()) {
		selectionList.insert(selectionList.end(), new SelectionItem(this, mode));
		return 1;
	}
	return 0;
}

void TextObject::transformByMatrix(TransformationMatrix& m, gboolean incFillPattern) {
	if (drawStyle && incFillPattern) drawStyle->transformByMatrix(m);
	if (fillStyle && incFillPattern) fillStyle->transformByMatrix(m);
}

GDStyle *TextObject::getStrokeStyle() {
	return drawStyle;
}

GDStyle *TextObject::getFillStyle() {
	return fillStyle;
}

gboolean TextObject::selected() {
	return selection;
}

TextObject::~TextObject() {
	if (drawStyle) {
		delete drawStyle;
	}
	if (fillStyle) {
		delete fillStyle;
	}
}
