/* $Id: TextRow.cc,v 1.29 2005/05/07 17:54:16 pklaus Exp $
 *
 * TextRow.cc
 *
 * Gestalter - The free vector drawing application
 * Copyright (c) 2002 Philipp Klaus <gestalter@linotux.ch>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include "TextRow.hh"
#include "GDFont.hh"
#include "GDCharacter.hh"
#include <ft2build.h>
#include FT_FREETYPE_H
#include <math.h>

TextRow::TextRow() {
	length = 0;
	descender = 2;
	ascender = 12;
	linespace = 16;
}

TextRow::TextRow(const TextRow& from) {
	chars = from.chars;
	for (guint i = 0; i < chars.size(); i++) {
		// g_printerr(" chars[%d] %lX from.chars %lX\n", i, chars[i], from.chars[i]);
		chars[i] = new GDCharacter(*(from.chars[i]));
	}
	defaultStyle = from.defaultStyle;
	calcLength();
}

void TextRow::insertChar(gint after, GDCharacter *ch) {
	//g_printerr("TextRow::insertChar after:%d ch %lX size: %d\n", after, &ch, size());
	vector<GDCharacter *>::iterator it = chars.begin();
	if (after == -1) {
		it = chars.end();
	} else {
		for(;after; after--) it++;
	}
	chars.insert(it, ch);
	calcLength();
}

GDCharacter *TextRow::removeChar(gint pos) {
	vector<GDCharacter *>::iterator it = chars.begin();
	if (pos == -1) {
		it = chars.end();
	} else {
		it += pos;
	}
	GDCharacter *c = *it;
	chars.erase(it);
	calcLength();
	return c;
}

void TextRow::removeChars(gint startpos, gint endpos) {
	vector<GDCharacter *>::iterator sit = chars.begin();
	if (startpos != -1) {
		sit += startpos;
	}
	vector<GDCharacter *>::iterator eit = chars.begin();
	if (endpos == -1) {
		eit = chars.end();
		eit--;
	} else {
		eit += endpos;
	}
	vector<GDCharacter *>::iterator it = sit;
	for (; it <= eit; it++) {
		//g_printerr("TextRow::removeChars delete GDChar %c (%d)\n", (*it)->unicode, (*it)->unicode);
		delete *it;
	}
	eit++;
	chars.erase(sit, eit);
	//g_printerr("String is now: ");
	//for (int i = 0; i < chars.size(); i++) {
	//	g_printerr("%c", chars[i]->unicode);
	//}
	//g_printerr("#\n");
}

GDCharacter *TextRow::getChar(guint pos) {
	if (pos < chars.size()) {
		return chars[pos];
	} else {
		return NULL;
	}
}

guint TextRow::size() {
	return chars.size();
}

void TextRow::calcLength() {
	//g_printerr("TextRow::calcLength\n");
	descender = 0;
	length = 0;
	ascender = 0;
	linespace = 0;
	GDFont *oldfont = NULL;
	double oldsize = -1;
	double oldhs = -1;
	FT_Face face = NULL;
	if (size() > 0) {
		gint gindex = -1;
		for (guint i = 0; i < size(); i++) {
			GDCharacter *ch = chars[i];
			ch->startX = length / 100;
			TextStyle chStyle = *(ch->getTextStyle());
			if (chStyle.getFont() != oldfont || chStyle.getPointSize() != oldsize || chStyle.getHorizontalScale() != oldhs) {
				oldfont = chStyle.getFont();
				oldsize = chStyle.getPointSize();
				oldhs = chStyle.getHorizontalScale();
				face = oldfont->getFace();
				if (! face) {
					g_printerr("No font found in character %d of line %lX\n", i, (long) this);
					continue;
				}
				int pixWidth = 0;
				if (oldhs != 1) {
					pixWidth = (int) (oldsize * 100 * oldhs);
				}
				FT_Set_Pixel_Sizes(face, pixWidth, (int) (oldsize * 100));
			}
			if (gindex == -1) {
				gindex = FT_Get_Char_Index(face, ch->getUnicode());
			}
			int error = FT_Load_Glyph(face, gindex, 0);
			if (error) {
				g_printerr("TextRow::calcLength ERROR cannot load glyph %d (charcode %d) from font\n", gindex, chars[i]->getUnicode());
			}
			length += ((double) face->glyph->advance.x) / 64.0;
			double asc = face->ascender * oldsize / 64.0 / 16;
			double desc = -(double)face->descender * oldsize / 64.0 / 16.0;
			if (chStyle.getBaseLineShift() > 0) {
				desc -= chStyle.getBaseLineShift();
				asc += chStyle.getBaseLineShift();
			}
			if (asc > ascender) {
				ascender = asc;
			}
			//g_printerr("normal desc: %f\n", desc);
			if (desc > descender) {
				descender = desc;
			}
			int oldindex = gindex;
			if (i < size() - 1) {
				gindex = FT_Get_Char_Index(face, chars[i+1]->getUnicode());
				FT_Vector kerning;
				error = FT_Get_Kerning(face, oldindex, gindex, ft_kerning_default, &kerning);
				if (error) {
					g_printerr("TextRow::calcLength ERROR cannot get kerning pair (%d - %d)\n", oldindex, gindex);
				}
				//g_printerr("Kerning: %d %d\n", kerning.x, kerning.y);
				length += ((double) kerning.x) / 64.0;
			}
			if (chStyle.getTracking()) {
				length += chStyle.getTracking() * oldsize / 10;
			}
			linespace = oldsize * 1.2;
			if (chStyle.getLineSpace() >= 0) {
				linespace = chStyle.getLineSpace();
			}
			ch->endX = length / 100;
			//g_printerr("TextRow::calcLength character %d: %c  Xrange: %f-%f\n", i, ch->unicode, ch->startX, ch->endX);
			//g_printerr("TextRow::calcLength character %d: %c  height: %f linespace: %f length: %f\n", i, ch->unicode, height, linespace, length);
		}
	} else {
		linespace = defaultStyle.getLineSpace();
		if (linespace < 0) {
			linespace = defaultStyle.getPointSize() * 1.2;
		}
	}
	length /= 100;
	//g_printerr("TextRow::calcLength length: %f descender: %f size: %d\n", length, descender, size());
}

void TextRow::writeXML(ostream& xmlout, gdouble x, gdouble y, GDDocument *doc) {
	if (size()) {
		TextStyle *ts = chars[0]->getTextStyle();
		GDStyle *fs = chars[0]->getFillStyle();
		GDStyle *ss = chars[0]->getStrokeStyle();
		fs->writeFillPreparationXML(xmlout, doc);
		xmlout << "<tspan x=\"0\" y=\"" << y << "\"";
		ts->writeXML(xmlout);
		fs->writeFillXML(xmlout, doc);
		ss->writeStrokeXML(xmlout, doc);
		xmlout << ">";
		guint i;
		for (i = 0; i < size(); i++) {
			GDCharacter *ch = chars[i];
			if (! (*ts == *(ch->getTextStyle())) || 
				! (*fs == *(ch->getFillStyle())) ||
				! (*ss == *(ch->getStrokeStyle()))) {
				ts = ch->getTextStyle();
				fs = ch->getFillStyle();
				ss = ch->getStrokeStyle();
				xmlout << "</tspan>";
				fs->writeFillPreparationXML(xmlout, doc);
				xmlout << "<tspan";
				ts->writeXML(xmlout);
				fs->writeFillXML(xmlout, doc);
				ss->writeStrokeXML(xmlout, doc);
				xmlout << ">";
			}
			xmlout << ch->convert2UTF8();
		}
			
		xmlout << "</tspan>\n";
	}
}

TextStyle *TextRow::writePS(ostream& psout, gdouble x, gdouble y, TextStyle *last, GDDocument *doc) {
	if (size()) {
		TextStyle ts = *(chars[0]->getTextStyle());
		GDStyle fs = *(chars[0]->getFillStyle());
		GDStyle ss = *(chars[0]->getStrokeStyle());
		int ts_char = 0;
		ts.writePS(psout);
		//psout << "/Helvetica findfont\n";
		//psout << "12 scalefont setfont\n";
		psout << x << " " << y << " m\n";
		TextStyle *primeStyle = getChar(0)->getTextStyle();
		gint justification = primeStyle->getJustification();
		if (justification == TEXT_STYLE_JUSTIFICATION_RIGHT || justification == TEXT_STYLE_JUSTIFICATION_CENTER) {
			// text is not left adjusted, so a previous movement is necessary
			string line;
			gdouble tracking = 0;
			for (guint i = 0; i < size(); i++) {
				GDCharacter *ch = chars[i];
				guint chcode = ch->getUnicode();
				if (chcode == '(' || chcode == ')' || chcode == '\\') {
					line.append(1, '\\');
				}
				line.append(1, (char) chcode);
				TextStyle *ts = ch->getTextStyle();
				if (ts->getTracking()) {
					tracking += ts->getTracking() * ts->getPointSize() / 1000;
				}
			}
			psout << tracking << " (" << line << ")";
			if (justification == TEXT_STYLE_JUSTIFICATION_RIGHT) {
				psout << " mrjust\n";
			} else {
				psout << " mcjust\n";
			}
		}
		psout << "(";
		guint i;
		for (i = 0; i < size(); i++) {
			GDCharacter *ch = chars[i];
			if (! (ts == *(ch->getTextStyle())) || ! (fs == *(ch->getFillStyle())) ||
				! (ss == *(ch->getStrokeStyle()))) {
				psout << ") ";
				showText(psout, fs, ss, doc);
				if (! (ts == *(ch->getTextStyle()))) {
					ch->getTextStyle()->writeDiffPS(ts, psout);
					ts = *(ch->getTextStyle());
					ts_char = i;
				}
				fs = *(ch->getFillStyle());
				ss = *(ch->getStrokeStyle());
				psout << "(";
			}
			guint chcode = ch->getUnicode();
			if (chcode == '(' || chcode == ')' || chcode == '\\') {
				psout << '\\';
			}
			psout << ( (char) chcode );
			if (ts.getTracking()) {
				psout << ") ";
				showText(psout, fs, ss, doc);
				psout << "\n " << (ts.getTracking() * ts.getPointSize() / 1000) << " 0 rm\n(";
			}
		}
		psout << ") ";
		showText(psout, fs, ss, doc);
		psout << "\n";
		return chars[ts_char]->getTextStyle();
	} else {
		return last;
	}
}

void TextRow::showText(ostream& psout, GDStyle& fs, GDStyle& ss, GDDocument *doc) {
	gboolean filled = fs.drawType != GDSTYLE_DRAWTYPE_TRANSPARENT;
	gboolean stroke = ss.drawType != GDSTYLE_DRAWTYPE_TRANSPARENT;
	gboolean shaded = fs.drawType == GDSTYLE_DRAWTYPE_GRADIENT;
	if (stroke) {
		ss.writePSStrokeStyle(psout, doc, false);
		psout << " ";
		ss.writeColorPS(psout, false);
		psout << " ";
	}
	if (filled) {
		if (shaded) {
			fs.writeGradientPS(psout, doc);
		} else {
			fs.writeColorPS(psout, false);
		}
		psout << " ";
	}
	psout << (filled ? "f" : "") << (shaded ? "g" : "") << (stroke ? "s" : "") << "draw\n";
}

void TextRow::transformByMatrix(TransformationMatrix& m, gboolean incFillPattern) {
	for (vector<GDCharacter *>::iterator it = chars.begin(); it != chars.end(); it++) {
		(*it)->transformByMatrix(m, incFillPattern);
	}
}

void TextRow::setFillMatrix(TransformationMatrix& m) {
	for (vector<GDCharacter *>::iterator it = chars.begin(); it != chars.end(); it++) {
		(*it)->setFillMatrix(m);
	}
}

void TextRow::changeDocument(GDDocument *document) {
	for (vector<GDCharacter *>::iterator it = chars.begin(); it != chars.end(); it++) {
		(*it)->changeDocument(document);
	}
}

TextRow::~TextRow() {
	for (vector<GDCharacter *>::iterator it = chars.begin(); it != chars.end(); it++) {
		delete *it;
	}
}
