/* $Id: TextStyle.cc,v 1.11 2005/05/07 17:39:33 pklaus Exp $
 *
 * TextStyle.cc
 *
 * Gestalter - The free vector drawing application
 * Copyright (c) 2002 Philipp Klaus <gestalter@linotux.ch>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include "TextStyle.hh"
#include "GDFont.hh"

TextStyle::TextStyle() {
	pointSize = 12;
	lineSpace = -1;
	autoLineSpace = true;
	baseLineShift = 0;
	tracking = 0;
	horizontalScale = 1;
	font = NULL;
	valueMask = 0xFFFFFFFF;
	justification = TEXT_STYLE_JUSTIFICATION_LEFT;
}

TextStyle::TextStyle(const TextStyle& from) {
	//g_printerr("TextStyle::TextStyle(const &)\n");
	pointSize = from.pointSize;
	lineSpace = from.lineSpace;
	baseLineShift = from.baseLineShift;
	tracking = from.tracking;
	horizontalScale = from.horizontalScale;
	autoLineSpace = from.autoLineSpace;
	font = from.font;
	if (font) {
		//g_printerr("TextStyle::TextStyle font->incUsageCount\n");
		font->incUsageCount();
	}
	valueMask = from.valueMask;
	justification = from.justification;
}
TextStyle TextStyle::operator=(const TextStyle& from) {
	//g_printerr("TextStyle::operator=(const &) just: %d -> %d\n", justification, from.justification);
	pointSize = from.pointSize;
	lineSpace = from.lineSpace;
	baseLineShift = from.baseLineShift;
	tracking = from.tracking;
	horizontalScale = from.horizontalScale;
	autoLineSpace = from.autoLineSpace;
	if (font) {
		font->decUsageCount();
	}
	font = from.font;
	if (font) {
		//g_printerr("TextStyle::operator= font->incUsageCount\n");
		font->incUsageCount();
	}
	valueMask = from.valueMask;
	justification = from.justification;
	return *this;
}

int TextStyle::operator==(const TextStyle& from) {
	if (pointSize != from.pointSize) return 0;
	if (lineSpace != from.lineSpace) return 0;
	if (baseLineShift != from.baseLineShift) return 0;
	if (tracking != from.tracking) return 0;
	if (horizontalScale != from.horizontalScale) return 0;
	if (autoLineSpace != from.autoLineSpace) return 0;
	if (font != from.font) return 0;
	if (justification != from.justification) return 0;
	return 1;
}

TextStyle TextStyle::operator&(const TextStyle& b) {
	TextStyle a = *this;
	//g_printerr("TextStyle::operator&: check valueMask: %lX\n", a.valueMask);
	if (a.pointSize != b.pointSize) a.valueMask &= ~TEXT_STYLE_POINT_SIZE_MASK;
	if (a.lineSpace != b.lineSpace) a.valueMask &= ~TEXT_STYLE_LINE_SPACE_MASK;
	if (a.autoLineSpace != b.autoLineSpace) a.valueMask &= ~TEXT_STYLE_AUTO_LINE_SPACE_MASK;
	if (a.baseLineShift != b.baseLineShift) a.valueMask &= ~TEXT_STYLE_BASE_LINE_SHIFT_MASK;
	if (a.tracking != b.tracking) a.valueMask &= ~TEXT_STYLE_TRACKING_MASK;
	if (a.horizontalScale != b.horizontalScale) a.valueMask &= ~TEXT_STYLE_HORIZONTAL_SCALE_MASK;
	if (a.font != b.font) a.valueMask &= ~TEXT_STYLE_FONT_MASK;
	if (a.justification != b.justification) a.valueMask &= ~TEXT_STYLE_JUSTIFICATION_MASK;
	//g_printerr("TextStyle::operator&: valueMask: %lX\n", a.valueMask);
	return a;
}

void TextStyle::operator&=(const TextStyle& ts) {
	//g_printerr("TextStyle::operator&=: vM: %lX  ts.vM: %lX\n", valueMask, ts.valueMask);
	valueMask &= ts.valueMask;
	valueMask &= (*this & ts).valueMask;
}

void TextStyle::operator|=(const TextStyle& ts) {
	if (ts.valueMask & TEXT_STYLE_AUTO_LINE_SPACE_MASK) {
		autoLineSpace = ts.autoLineSpace;
	}
	if (ts.valueMask & TEXT_STYLE_POINT_SIZE_MASK) {
		pointSize = ts.pointSize;
		if (autoLineSpace) {
			lineSpace = pointSize * 1.2;
		}
	}
	if (ts.valueMask & TEXT_STYLE_LINE_SPACE_MASK) {
		lineSpace = ts.lineSpace;
	}
	if (ts.valueMask & TEXT_STYLE_BASE_LINE_SHIFT_MASK) {
		baseLineShift = ts.baseLineShift;
	}
	if (ts.valueMask & TEXT_STYLE_TRACKING_MASK) {
		tracking = ts.tracking;
	}
	if (ts.valueMask & TEXT_STYLE_HORIZONTAL_SCALE_MASK) {
		horizontalScale = ts.horizontalScale;
	}
	if (ts.valueMask & TEXT_STYLE_FONT_MASK) {
		if (font) {
			font->decUsageCount();
		}
		font = ts.font;
		if (font) {
			font->incUsageCount();
		}
	}
	if (ts.valueMask & TEXT_STYLE_JUSTIFICATION_MASK) {
		justification = ts.justification;
	}
}

void TextStyle::setFont(GDFont *newFont) {
	if (font) {
		//g_printerr("TextStyle::setFont font->decUsageCount\n");
		font->decUsageCount();
	}
	font = newFont;
	if (font) {
		//g_printerr("TextStyle::setFont font->incUsageCount\n");
		font->incUsageCount();
	} else {
		g_printerr("TextStyle::setFont newFont is NULL!!\n");
	}
}

void TextStyle::setPointSize(double value) {
	pointSize = value;
	if (autoLineSpace) {
		lineSpace = pointSize * 1.2;
		//g_printerr("TextStyle::setPointSize: %s (%f: %f)\n", (autoLineSpace ? "t" : "f"), pointSize, lineSpace);
	}
}

void TextStyle::setLineSpace(double value) {
	lineSpace = value;
}

void TextStyle::setAutoLineSpace(gboolean value) {
	autoLineSpace = value;
	if (autoLineSpace) {
		lineSpace = pointSize * 1.2;
	}
	//g_printerr("TextStyle::setAutoLineSpace: %s (%f: %f)\n", (autoLineSpace ? "t" : "f"), pointSize, lineSpace);
}

void TextStyle::setBaseLineShift(double value) {
	baseLineShift = value;
}

void TextStyle::setTracking(double value) {
	tracking = value;
}

void TextStyle::setHorizontalScale(double value) {
	horizontalScale = value;
}

void TextStyle::setJustification(guint value) {
	justification = value;
}

void TextStyle::writeXML(ostream& xmlout) {
	xmlout << " font-family=\"" << font->getFamily() << "\"";
	xmlout << " font-size=\"" << pointSize << "\"";
	if (*(font->getSlant()) == 'r') {
		xmlout << " font-style=\"normal\"";
	} else {
		xmlout << " font-style=\"italic\"";
	}
	xmlout << " font-weight=\"" << font->getWeight() << "\"";
	if (autoLineSpace) {
		xmlout << " gestalter:linespace=\"" << lineSpace << "\"";
	} else {
		xmlout << " gestalter:linespace=\"auto\"";
	}
	xmlout << " gestalter:baselineshift=\"" << baseLineShift << "\"";
	xmlout << " gestalter:tracking=\"" << tracking << "\"";
	xmlout << " gestalter:horizontalscale=\"" << horizontalScale << "\"";
	xmlout << " gestalter:font=\"" << font->getName() << "\"";
	xmlout << " text-anchor=\"";
	if (justification == TEXT_STYLE_JUSTIFICATION_LEFT) {
		xmlout << "start";
	} else if (justification == TEXT_STYLE_JUSTIFICATION_RIGHT) {
		xmlout << "end";
	} else if (justification == TEXT_STYLE_JUSTIFICATION_CENTER) {
		xmlout << "middle";
	} else {
		xmlout << "inherit";
	}
	xmlout << "\"";
}

void TextStyle::writePS(ostream& psout) {
	psout << "/" << font->getPSName() << "-isoenc " << pointSize << " ffsf\n";
	if (baseLineShift) {
		psout << "0 " << baseLineShift << " rm\n";
	}
	if (horizontalScale != 1.0) {
		psout << horizontalScale << " 1 sl\n";
	}
}

void TextStyle::writeDiffPS(TextStyle& old, ostream& psout) {
	if (baseLineShift != old.baseLineShift) {
		psout << "0 " << (baseLineShift - old.baseLineShift) << " rm\n";
	}
	if (horizontalScale != old.horizontalScale) {
		psout << (horizontalScale / old.horizontalScale) << " 1 sl\n";
	}
}

TextStyle::~TextStyle() {
	if (font) {
		//g_printerr("TextStyle::~TextStyle font->decUsageCount\n");
		font->decUsageCount();
	}
}
