/* $Id: Undo.cc,v 1.10 2005/05/07 17:39:33 pklaus Exp $
 *
 * Undo.cc
 *
 * Gestalter - The free vector drawing application
 * Copyright (c) 2002 Philipp Klaus <gestalter@linotux.ch>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "Undo.hh"
#include "GDDocument.hh"
#include "Preferences.hh"
#include "Layer.hh"
#include "GDApplication.hh"
#include "GDAction.hh"
#include "TransformAction.hh"
#include "gestalter.h"

Undo::Undo(GDDocument *document) {
	doc = document;
	actionConnection = doc->setAction.connect(SigC::slot(this, &Undo::newAction));
	doc->application->prefsChanged.connect(SigC::slot(this, &Undo::onPrefsChanged));
	start = 0;
	end = 0;
	current = 0;
	size = doc->application->preferences.undoStackSize;
	undoStack.resize(size, NULL);
	lastAction = NULL;
}

gint Undo::newAction(GDAction *action) {
	//g_printerr("Undo::newAction: %s\n", action->action.c_str());
	if (lastAction) {
		delete lastAction;
	}
	lastAction = action;
	if (undoStack[current]) {
		delete undoStack[current];
	}
	undoStack[current] = new UndoDoc(doc);
	current = (current + 1) % size;
	if (current == start) {
		start = (start + 1) % size;
	}
	end = current;
	doc->updateWindowTitle();
	return 0;
}

void Undo::onPrefsChanged(Preferences *newPrefs) {
	resizeStack(newPrefs->undoStackSize);
}

void Undo::resizeStack(guint newSize) {
	if (size == newSize) return;
	guint e = end;
	guint c = (current + size - start) % size;
	vector<UndoDoc *> newStack;
	newStack.resize(newSize, NULL);
	if (newSize < size) {
		e = current;
	}
	guint os = (e + size - start) % size;
	if (os > newSize) {
		c -= os - newSize;
		os = newSize;
	}
	start = 0;
	end = os;
	current = c;
	for (; os; os--) {
		if (e < 0) {
			e += size;
		}
		newStack[os - 1] = undoStack[e - 1];
		e--;
	}
	undoStack = newStack;
	size = newSize;
}

gint Undo::undo() {
	if (current != start) {
		current = (current + doc->application->preferences.undoStackSize - 1) % size;
		if (undoStack[current]) {
			UndoDoc *currentDoc = new UndoDoc(doc);
			undoStack[current]->installInDoc(doc);
			delete undoStack[current];
			undoStack[current] = currentDoc;
			doc->redraw();
			doc->updateWindowTitle();
			return 1;
		}
	}
	return 0;
}

gint Undo::redo() {
	if (current != end && undoStack[current]) {
		undoStack[current]->installInDoc(doc);
		current = (current + 1) % size;
		doc->redraw();
		doc->updateWindowTitle();
		return 1;
	}
	return 0;
}

gboolean Undo::undoPossible() {
	if (current != start) {
		return true;
	}
	return false;
}

gboolean Undo::redoPossible() {
	if (current != end) {
		return true;
	}
	return false;
}

gboolean Undo::repeatTransformPossible() {
	//g_printerr("Undo::repeatTransformPossible: %s\n", (lastAction ? lastAction->action.c_str() : ""));
	if (lastAction && ! STRINGOBJ_COMPARE(lastAction->action, 0, 9, "Transform")) {
		//g_printerr("TRUE\n");
		return true;
	}
	//g_printerr("FALSE\n");
	return false;
}

void Undo::repeatTransform() {
	if (!lastAction || STRINGOBJ_COMPARE(lastAction->action, 0, 9, "Transform")) {
		// only transforms can be repeated
		return;
	}
	TransformAction *tAction = (TransformAction *) lastAction;
	lastAction = NULL;
	SelectionItemList selection;
	TransformAction *newAction = new TransformAction(tAction->tMatrix, tAction->doCopy);
	doc->setAction(newAction);
	doc->getSelection(selection, SELECTIONITEM_MODE_COPY_OBJECT);
	selection.transformByMatrix(tAction->tMatrix);
	if (STRINGOBJ_COMPARE(tAction->action,0, 13, "TransformCopy")) {
		selection.applyToOriginal();
	} else {
		doc->deSelectAll();
		selection.insertMoving(doc);
	}
	delete tAction;
}

void Undo::setSaved() {
	if (current == start) {
		return;
	}
	for (vector<UndoDoc *>::iterator it = undoStack.begin(); it != undoStack.end(); it++) {
		if (*it) {
			(*it)->saved = false;
		}
	}
	guint cur = (current + size - 1) % size;
	if (undoStack[cur]) {
		undoStack[cur]->saved = true;
	}
}

gboolean Undo::getSaved() {
	if (current == start) {
		return true;
	}
	guint cur = (current + size - 1) % size;
	UndoDoc *currDoc = undoStack[cur];
	if (currDoc) {
		return currDoc->saved;
	} else {
		return false;
	}
}

UndoDoc::UndoDoc(GDDocument *doc) {
	//g_printerr("UndoDoc::UndoDoc this=%lX\n", this);
	layers = NULL;
	saved = false;
	Layer *layerp = NULL;
	for (Layer *doclayers = doc->layers; doclayers; doclayers = doclayers->next) {
		Layer *newLayer = doclayers->clone();
		if (layerp) {
			layerp->next = newLayer;
			newLayer->prev = layerp;
			layerp = newLayer;
		} else {
			layers = newLayer;
			layerp = newLayer;
			newLayer->prev = NULL;
		}
	}
	if (layerp) {
		layerp->next = NULL;
	}
}

void UndoDoc::installInDoc(GDDocument *doc) {
	//g_printerr("UndoDoc::installInDoc this=%lX\n", this);
	Layer *layer = doc->layers;
	while (layer) {
		Layer *next = layer->next;
		delete layer;
		layer = next;
	}
	Layer *layerp = NULL;
	for (Layer *doclayers = layers; doclayers; doclayers = doclayers->next) {
		Layer *newLayer = doclayers->clone();
		if (layerp) {
			layerp->next = newLayer;
			newLayer->prev = layerp;
			layerp = newLayer;
		} else {
			doc->layers = newLayer;
			layerp = newLayer;
			newLayer->prev = NULL;
		}
	}
	if (layerp) {
		layerp->next = NULL;
	}
	doc->activeLayer = doc->layers;
	doc->recalcSelection();
}	

UndoDoc::~UndoDoc() {
	//g_printerr("UndoDoc::~UndoDoc called this=%lX\n", this);
	Layer *layer = layers;
	while (layer) {
		Layer *next = layer->next;
		delete layer;
		layer = next;
	}
}
